
#include "qcalc.h"
#include "qwui.h"

/// \file qcalcListEvents.cc
///  Implementation file for the class QcalcListView. It contains mainly the overridden function that handles events which are: \n
///  \arg \c keyPressEvent                 \n
///  \arg \c contentsMouseReleaseEvent     \n
///  \arg \c dragMoveEvent                 \n
///  \arg \c contentsMouseDoubleClickEvent \n \n
///  In addition, this file includes two helper functions which are: \n
///  \arg \c setSelectedList()      \n \n
///  The first helper function adds the selected item(s) within the local list of selected item and emits a signal signifying a change \n
///  has occurred within the current list view.  This signal will be used to refresh the content of the join tab with the ioining options of the \n
///  item(s) that has(have) been selected.\n 
///  \author Nora Bounaira
///  \version 2.0          \n
///  \date    2004

using namespace QSAS;

extern qwuiWindow* qsasWindow;

/// \var  extern vector<QguListItem *> dragList
/// \brief Global variable that contains the list of items being dragged
///
extern vector<QguListItem *> dragList;

/// \var  extern QguListView * dragFromList
/// \brief Global variable that defines the list view from which the items are being dragged
///
extern QguListView * dragFromList;



/// \fn calcListView::keyPressEvent( QKeyEvent* e )
/// \brief Event handler for event e (key at the keyboard being pressed) to receive the key press events that are meaningful for the \n
/// calculator list view.  The events that are filtred are: \n
/// \arg \c Key_Up \n
/// \arg \c Key_Down \n
/// \arg \c Key_Home \n
/// \arg \c Key_End \n
/// \arg \c Key_PageDown \n
/// \arg \c Key_PageUp \n \n
/// Those events are processed within this hanlder in order to refresh the content of the join tab when the user traverses the list view \n
/// using one of the keys listed above.
/// \param     e of QKeyEvent * type
/// \exception None
/// \return    None
///
void QcalcListView::keyPressEvent( QKeyEvent* e )
{
	QcalcListVirtualItem*  selected_item  = NULL;
	setFocus();

	if (    ( e->key() == Qt::Key_Up       )
	     || ( e->key() == Qt::Key_Down     )
		|| ( e->key() == Qt::Key_Home     )
		|| ( e->key() == Qt::Key_End      )
		|| ( e->key() == Qt::Key_PageDown )
		|| ( e->key() == Qt::Key_PageUp   )    )
	{
		Q_LIST_VIEW::keyPressEvent( e );

		selected_item = ( QcalcListVirtualItem * )currentItem();

		if ( selected_item == NULL )
		{
			clearSelectedList();
		}
		else
		{
			if ( selected_item->getEntryType() == ListTypeNull )
			{
				clearSelectedList();
			}
			else
			{
				// single click on item not selected
				setSelectedList( selected_item );
				refreshSelectedList();
				this->owner->tabWidget->showPage(selected_item->m_pcalcTab);
			}
		}
	}
	else
	{
		// else pass on to list view
		e->ignore(); // let selection change
	}

}

void QcalcListView::contentsMouseReleaseEvent( QMouseEvent *event )
{
	// called if mouse pressed and released (i.e. not held down and moved) and
	// after contentsMousePressEvent
   if(QuSlotDebug())
  	printf("QcalcListView::contentsMouseReleaseEvent %ld\n", this);

	#ifdef QT_VERSION
        #if QT_VERSION <0x030000
		if(this->hasFocus()) {
			if( bSingleSelection && event->state() != Qt::ControlButton && event->state() != Qt::ShiftButton)
			{
				// not doing multiple selection, so show tab on mouse up
				QcalcListVirtualItem * selected_item = 
				dynamic_cast<QcalcListVirtualItem *>(itemAt(event->pos() - QPoint(0,this->contentsY()) ) );
				if(selected_item != NULL ) 
				{
					this->owner->tabWidget->showPage(selected_item->m_pcalcTab);
				}
	 		}
	
			// now pass this on to QguListView to handle selection feedback if drop in this list
			QguListView::contentsMouseReleaseEvent(event);
		}
		#else
		if(this->hasMouse()) {
			if( bSingleSelection && event->state() != Qt::ControlButton && event->state() != Qt::ShiftButton)
			{
				// not doing multiple selection, so show tab on mouse up
				QcalcListVirtualItem * selected_item = 
				dynamic_cast<QcalcListVirtualItem *>(itemAt(event->pos() - QPoint(0,this->contentsY()) ) );
				if(selected_item != NULL ) 
				{
					this->owner->tabWidget->showPage(selected_item->m_pcalcTab);
				}
	 		}
	
			// now pass this on to QguListView to handle selection feedback if drop in this list
			QguListView::contentsMouseReleaseEvent(event);
		}
		#endif // end of #if
	#endif // end of #ifdef

}

void QcalcListView::contentsMousePressEvent( QMouseEvent *event )
{
	// called if mouse pressed 
	
   if(QuSlotDebug())
  	printf("QcalcListView::contentsMousePressEvent %ld\n", this);
	
	//  pass this on to QguListView to handle selection feedback
	QguListView::contentsMousePressEvent(event);

}


/// \fn calcListView::contentsMouseMoveEvent( QMouseEvent *event )
/// \brief Event handler for event event \n
/// The text to be dragged is either the context if the target tab is the join or the alias if the target tab is a calculation
/// \param     e of QMouseEvent* type
/// \exception None
/// \return    None
///

void QcalcListView::contentsMouseMoveEvent( QMouseEvent *event )
{
	// called if mouse held down and moved and
	// after contentsMousePressEvent
 //  if(QuSlotDebug())
 // 	printf("QcalcListView::contentsMouseMoveEvent %ld, mousePressed flag = %ld\n", this, mousePressed);

	// check left mouse button pressed down
	if ( mousePressed )
	{
		// check drag item selected

		if ( dragItem )
		{
			// if mouse has moved more than 3 pixels in the x or/and y direction
			// from the starting position (the position of the mousePressEvent)
			// then assume a drag operation
			if (   ( abs( event->pos().x() - dragPos.x() ) >= 3  )
			    || ( abs( event->pos().y() - dragPos.y() ) >= 3  ) )
			{

				// start drag
				QcalcListVirtualItem *qItem = dynamic_cast< QcalcListVirtualItem* >( dragItem );

				if ( qItem )
				{

#if QT_VERSION < 0x040000
					Q_TEXT_DRAG* d = NULL;
					if ( owner )
					{
						d = new Q_TEXT_DRAG( qItem->text(Alias_).latin1(), this );
						if ( d )
						{
							d->drag();
						}
					}
#else
					QDrag* d;
					if(owner)
					{
						d=new QDrag(this);
						QMimeData * mimeData=new QMimeData;
						mimeData->setText(qItem->text(Alias_));
						d->setMimeData(mimeData);
#if QT_VERSION < 0x040300
						d->start();
#else
						d->exec();
#endif
					}
#endif
			
					// reset mousePressed flag
					mousePressed = false;

					// tidy up after drag i.e unselect items (may not be desired)
					// if already selected
					if ( dragFromList )
					{
						dragFromList->refreshSelectedList();
					}
					
				}
			}
		}
	}
	Q_LIST_VIEW::contentsMouseMoveEvent( event ); 
}

/// \fn calcListView::dragMoveEvent( QDragMoveEvent *event )
/// \brief Event handler for event e (key at the keyboard being pressed) to receive the key press events that are meaningful for the \n
/// calculator list view.  The events that are filtred are: \n
/// \arg \c Key_Up \n
/// \arg \c Key_Down \n
/// \arg \c Key_Home \n
/// \arg \c Key_End \n
/// \arg \c Key_PageDown \n
/// \arg \c Key_PageUp \n \n
/// Those events are processed within this hanlder in order to refresh the content of the join tab when the user traverses the list view \n
/// using one the keys listed above.
/// \param     e of QKeyEvent * type
/// \exception None
/// \return    None
///
void QcalcListView::dragMoveEvent( QDragMoveEvent *event )
{
	if ( QuSlotDebug() )
		std::cout << "QcalcListView:::dragMoveEvent " << dynamic_cast<QguListItem*>( currentItem() ) << endl;
	
	QguListItem*  item;
	QWidget*      window;

	window = this->parentWidget();
	emit raiseWindow();
	if ( window->isActiveWindow() )
	{
#if QT_VERSION < 0x040000
		Q_SCROLL_VIEW::dragMoveEvent( event );
		event->acceptAction( QTextDrag::canDecode(event) );
#else
		if (event->mimeData()->hasText()) event->acceptProposedAction();
#endif
		
		item = ( QguListItem* )itemAt( event->pos() - QPoint(23,23 ) );
		if ( item != 0)
		{
			setCurrentItem( item );
		}

		QPoint scrollPoint = event->pos() + QPoint( this->contentsX(), this->contentsY() );
		ensureVisible( scrollPoint.x() , scrollPoint.y() - 20, 25, 25 );
	}

	this->refreshSelectedList();
}


/// \fn calcListView::dropEvent( QDropEvent *event )
/// \brief Event This event handler is called when the drag is dropped on the list view. A same object can be dropped more than once. \n
/// calculator list view.  \n
/// \param     event of QDropEvent* type
/// \exception None
/// \return    None
///
void QcalcListView::dropEvent( QDropEvent *event )
{
	QString      txt;
	QguListItem* newItem = NULL;

	if ( QuSlotDebug() )
		std::cout << "QcalcListView::dropEvent " << dynamic_cast<QguListItem*>( currentItem() ) << endl;

	clearSelectedList();
	
	QguListItem *dropAtItem = ( QguListItem * )currentItem();

	if ( dragList.size() == 0 ) return;

#if QT_VERSION < 0x040000
	if ( QTextDrag::decode (event, txt ) )
#else
	if (event->mimeData()->hasText())
#endif
	{
#if QT_VERSION < 0x040000
		// special case of single click on item

		//  copy drag text to clipboard to allow repeat paste
		QApplication::clipboard()->setData( new QTextDrag(txt) );
#else
		txt=event->mimeData()->text();
		QMimeData * clipData=new QMimeData;
		clipData->setText(txt);
		QApplication::clipboard()->setMimeData(clipData);
#endif

		// prevent drop onto item in dragList or item with ancestor on list
		// Note dragList and localList contain same items here
		if ( dragList[ 0 ]->listView() == this )
		{
			if ( isOnSelectedList(dropAtItem) || findAncestorOnList(dropAtItem))
			{
				refreshSelectedList();
				return;
			}
		}

		insertListItems( dropAtItem );

		clearSelectedList();
		dragList.clear();
		validateWLitems();
		emit listModified();

	}// end if on decode event 
}


/// \fn calcListView::contentsMouseDoubleClickEvent( QMouseEvent *event )
/// \brief Event handler for event e (key at the keyboard being pressed) to receive the key press events that are meaningful for the \n
/// calculator list view.  The events that are filtred are: \n
/// \arg \c Key_Up \n
/// \arg \c Key_Down \n
/// \arg \c Key_Home \n
/// \arg \c Key_End \n
/// \arg \c Key_PageDown \n
/// \arg \c Key_PageUp \n \n
/// Those events are processed within this hanlder in order to refresh the content of the join tab when the user traverses the list view \n
/// using one the keys listed above.
/// \param     e of QKeyEvent * type
/// \exception None
/// \return    None
///
void QcalcListView::contentsMouseDoubleClickEvent( QMouseEvent *event )
{
  // disable mouse down on right and middle button
	// forces launch to be from left button and protects
	// from two mouse down events causing drag problems

	if ( event->button() == Qt::RightButton ) return;
	if ( event->button() == Qt::MidButton )   return;

	if ( event->state() == Qt::ControlButton || event->state() == Qt::ShiftButton )
	{
		return;
	}

	// clean up localList as double click only applies to single item
	clearSelectedList();

	QguListItem *selected_item = ( QguListItem * )itemAt( event->pos() - QPoint( 0, this->contentsY() ) );
	if ( selected_item == NULL )
	{
		// Make sure that the item is not edited if double click on remaining of the window
		return;
	}

	QguListItem *atItem = ( QguListItem * )currentItem() ;
	if ( atItem  == NULL ) return;

	if ( atItem->getEntryType() == ListTypeGroup )
	{
		// open/close group view
		Q_LIST_VIEW::contentsMouseDoubleClickEvent(event);
	}
	else
	{
		// view linked data object on WL
		// find item on WL from WLcontext and set it to be current item
		QcalcListVirtualItem*   qItem = dynamic_cast< QcalcListVirtualItem* >( atItem );
		if ( qItem )
		{
			if ( qItem->m_eObjType == WORKING_LIST_OBJECT )
			{
				QguListItem* wli = qsasWindow->WorkingList->getItemByName( atItem->WLcontext );
				qsasWindow->WorkingList->setCurrentItem( wli );
				// tell WL to browse current item
				emit viewWL_ObjRequest();
			}
			else
			{
				QcalcListVirtualItem* qVirt = dynamic_cast< QcalcListVirtualItem* >( qItem );
				
				if ( qVirt->m_qdObjResult.is_nil() )
				{
					this->reNameItemAlias( qItem );
				}
				else
				{
					QdObject_var      qdNilObj;
                   			DataBrowseWindow* QBROWSER_widget = QbuLaunchBrowse( qdNilObj,
				                                      this,
							              qVirt->text(Obj_Name_),
							              qVirt->m_qdObjResult,
					 	                      qVirt->getWLcontext() );
					if ( QBROWSER_widget->name_field )
					{
						QBROWSER_widget->name_field->lineBox->setReadOnly( true );
					}

					if ( QBROWSER_widget->browser )
					{
						if ( QBROWSER_widget->browser->activeEditor->valueField )
						{
							QBROWSER_widget->browser->activeEditor->valueField->lineBox->setReadOnly( true );
						}
						if ( QBROWSER_widget->browser->activeEditor->timeField )
						{
							QBROWSER_widget->browser->activeEditor->timeField->setReadOnly( true );
						}
						if ( QBROWSER_widget->editMenu )
						{
							QBROWSER_widget->editMenu->setEnabled( false );
						}
						if ( QBROWSER_widget->browser->parentsBtn )
						{
							QBROWSER_widget->browser->parentsBtn->setEnabled( false );
						}
						if ( QBROWSER_widget->browser->xrefList )
						{
							disconnect(    QBROWSER_widget->browser->xrefList,
						             	   SIGNAL( doubleClicked( Q_LIST_VIEW_ITEM* ) ),
         									   QBROWSER_widget->browser,
									 	   SLOT( xrefEdit( Q_LIST_VIEW_ITEM* ) ) );
						}
					}
				}
			}
		}
		else
		{
			// open/close group view
			Q_LIST_VIEW::contentsMouseDoubleClickEvent(event);
		}
	}
}


/// \fn calcListView::setSelectedList(QguListItem *item)
///
void QcalcListView::setSelectedList(QguListItem *item)
{
  this->QguListView::setSelectedList(item);
  emit SelectionUpdated();
}


