/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	Q4SC_Normal.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS 4 SC Determination of Boundary Normal and Velocity wrapper
*
* PURPOSE:      Calculate the normal and boundary velocity based on the
*               crossing of the boundary (assumed planar) by 4 spacecraft.
*               Based on the methd in Ch 10.4.3 of the ISSI book.
*
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*   INPUT   ANY_DATA_OBJECT    dr1  pos of s/c1 as const vec or time ser
    INPUT   ANY_DATA_OBJECT    dr2 ,
    INPUT   ANY_DATA_OBJECT    dr3 ,
    INPUT   ANY_DATA_OBJECT    dr4 ,
    INPUT   TIME_INTERVAL      tidr, centre time used if dr's are time ser
    INPUT   ANY_DATA_OBJECT    ti1 , event time as s/c 1 in secs or Time Int Obj
    INPUT   ANY_DATA_OBJECT    ti2 ,
    INPUT   ANY_DATA_OBJECT    ti3 ,
    INPUT   ANY_DATA_OBJECT    ti4 ,
    INPUT   ANY_DATA_OBJECT    taus , string array with 4 elements, the 4 crossing durations

    Out:

    OUTPUT  MATRIX             n_cva, n_cta,   
                                     normal vectors with same frame as dr's
                                     n_x forced positive
    OUTPUT  FLOAT              Vn_cva, Vn_cta,
                                      velocity along normal

* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: Hacked from Boundary Normal routine
*
* HISTORY: November 2003 Steve Schwartz 
*          January 2004 AJA added error traps
*          28 April, 2004 SEH : added CTA part
*          01 Jun, 2004 SEH : added combined CTA/CVA 
*          25 Aug, 2006 AJA : added to QSAS distribution
*
// Doxygen info :
//
/* -*- mode: C++ ; c-file-style: "stroustrup" */
/***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include "qplug_if.h"
#include "qrtn.h"

#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include "Xrefs.h"

#include <qstring.h>
#include <qdatetime.h>

//const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus CVA_CTA_Calculator(QplugArgList *);
QplugReturnStatus Q4SCPosition(QrVector *vecp,
                              QdObject_var pos_in,
			      QdTimeInterval_var time_int);
QplugReturnStatus Q4SCConstPosition(QrVector *vecp, 
                                    QdRMatrix_var pos_in, 
				    QdTimeInterval_var time_int);
QplugReturnStatus Q4SCCentrePosition(QrVector *vecp, 
                                    QdRMatrixSeq_var pos_in_seq, 
				    QdTimeInterval_var time_int);

void QrVector_diag_print(QrVector *vecp);
QdRMatrix_var str2Vec(QString s);
QdRMatrix_var CTA_calc(double R[4][3], double dt[4], QdString_var taustring, 
		       QdRMatrix_var vn_cta, QdRMatrix_var d_cta, 
		       QdRMatrix_var vn_com, QdRMatrix_var d_com, QdRMatrix_var n_com, 
		       QdRScalar_var v_cva, QdRMatrix_var d_cva, QdRMatrix_var n_cva, 
		       QdRScalarSeq_var v_cva_curve,  QdRScalarSeq_var v_cta_curve, QdRScalarSeq_var v_com_curve, 
		       QdTimeInterval_var tint, QdRMatrix_var tau);
double vecsum(QdRMatrix_var V);

 
#define MAT_INV_EPS 10*DBL_EPSILON

typedef enum MAT_INV_RET_E{
MAT_INV_OK,
MAT_INV_SINGULAR,
MAT_INV_UNKNOWN
} MAT_INV_RET;

MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]);
double determinant(double mat[3][3]);
void multiply_mat(double a[3][3],double b[3][3],double prod[3][3]);
void multiply_mat_vec(double a[3][3],double b[3],double prod[3]);
void print_mat(double a[3][3]);
 
  
/**********************************************************************/
QplugReturnStatus CVA_CTA_Calculator (  QplugArgList *call_list)
/*!
  \brief Perform a 4SC boundary normal calculation
  \param call_list - pointer to a list of input objects
*/
/**********************************************************************/
{
// define some C variables

int i,j,sign_nx;
double dt[4];
double  A[3][3], A_INV[3][3], B[3];
double  XX[3],n[3];
double drvec[4][3], signedmodX;
MAT_INV_RET ret_mat_inv;

// call-list arguments
 
QdObject_var         drin[4];        //  0-3
QdTimeInterval_var   tidr;           //  4
QdRScalar_var        ti_scalvar[4];  //  5-8
QdTimeInterval_var   ti_tivar[4];    //  5-8 alt
QdRMatrix_var        n_var;          //  10
QdRScalar_var        Vn_var;         //  12

// define Qdos data objects

QrVector  dr[4];

QplugReturnStatus ret_status;

if(QuSlotDebug()){
   printf("Q4SC Plugin Diagnostics On; undefine in .cc file and recompile\n");
   fflush(stdout);
}

// Ensure no input slots are empty

 if(call_list->size() != 10 ){
    QplugAppendTextDisplay("At least one input slot is empty");
    return QPLUG_FAILURE;
 }
 
// ...4 items are the s/c positions

for (i=0; i<4; i++){
  drin[i] = (*call_list)[i];
  if (drin[i].is_nil()){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Error getting pos dr[%d] object\n",i+1);
    QplugAppendTextDisplay(pos_err_string);
  }
	if(QuSlotDebug()){
		printf("Q4SC got input s/c %d position\n",i);
		fflush(stdout);
	}
}

tidr = get_timeinterval((*call_list)[4]);
if( tidr.is_nil() ) {
  QplugAppendTextDisplay ("Q4SC>> Time Interval object for tidr invalid.\n");
  return QPLUG_FAILURE;
}

if(QuSlotDebug()){
   printf("Q4SC got input time interval\n",i);
   fflush(stdout);
}


for (i=0; i<4; i++){
  ret_status = Q4SCPosition(&(dr[i]),drin[i],tidr);
  if (ret_status != QPLUG_SUCCESS){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Q4SCPosition failed for sc[%d]\n",i);
    QplugAppendTextDisplay(pos_err_string);
    return QPLUG_FAILURE;
  }
if(QuSlotDebug()){
     printf("Q4SC calculated s/c %d position...\n",i);
	 fflush(stdout);
}
  drvec[i][0] = *(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"x")));
  drvec[i][1] = *(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"y")));
  drvec[i][2] = *(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"z")));
	if(QuSlotDebug()){
		printf("and filled drvec = (%f, %f,%f)\n",drvec[i][0],drvec[i][1],drvec[i][2]);
		fflush(stdout);
	}
}

// Whew! dr[i]'s now have a sequence with 1 element, the position
// of the crossing at s/c i with one timetag, and drvecs are C-arrays 

for (i=0; i<4; i++){
  ti_scalvar[i] = QdRScalar_var::narrow((*call_list)[i+5]);
  ti_tivar[i]   = get_timeinterval( ((*call_list)[i+5]));
  if (!(ti_scalvar[i].is_nil())){
    dt[i] = (double) (*(ti_scalvar[i]));
  } else if (!(ti_tivar[i].is_nil())) {
    dt[i] = (double) Seconds ( ti_tivar[i]->duration() );
  } else {
    char time_err_string[200];
    sprintf(time_err_string,
         "Q4SC ERROR>> ti[%d] neither scalar nor time interval\n",i+1);
    QplugAppendTextDisplay(time_err_string);
    return QPLUG_FAILURE;
  }
	if(QuSlotDebug()){
		printf("Q4SC calculated s/c %d dt as %g\n",i,dt[i]);
		fflush(stdout);
	}
}
// dt[i] now contains times in secs for all 4 s/c

// Now fill dr matrix and right hand sides; s/c 4 is the reference
// [it doesn't matter which s/c is the ref; there are only 3 indep pairs]

for(i=0; i<3; i++){
  for(j=0; j<3; j++){
    A[i][j] = drvec[i][j] - drvec[3][j];
    printf(" %10.2f",drvec[i][j]);
	fflush(stdout);
  }
  printf("\n");
  fflush(stdout);
  B[i]      = dt[i] - dt[3];
}

if(QuSlotDebug()){
   printf("Q4SC A matrix is\n");
   fflush(stdout);
   print_mat(A);
}
ret_mat_inv = mat_inv(A,A_INV);
switch (ret_mat_inv) {
   case MAT_INV_OK:
     break;
   case MAT_INV_SINGULAR:
     QplugAppendTextDisplay("Q4SC ERROR>> dr matrix singular\n");
     return QPLUG_FAILURE;
     break;
   case MAT_INV_UNKNOWN: default:
     QplugAppendTextDisplay("Q4SC ERROR>> Matrix Inv unknown error\n");
     return QPLUG_FAILURE;
     break;
}
if(QuSlotDebug()){
   printf("Q4SC A inv matrix found successfully as\n");
   printf("Q4SC B       is (%g, %g, %g)\n", B[0],B[1],B[2]);
   print_mat(A_INV);
   fflush(stdout);
}




multiply_mat_vec(A_INV, B, XX);
sign_nx = (XX[0] < 0 ) ? -1 : 1;  
Vn_var = new QdRScalar();
signedmodX = (float)sign_nx*sqrt(XX[0]*XX[0] + XX[1]*XX[1] + XX[2]*XX[2]);
*Vn_var = 1.0/signedmodX;
n_var  = new QdRMatrix(Dim(3));
n_var->copy_xrefs_from( dr[3].ptr); // copy  frame, etc.
char crossingt_str[200];
char outstr[200];
sprintf(crossingt_str,"%s", (dr[3].t)[0].iso_srep().c_str());
sprintf(outstr,"Q4SC>> Crossing at %s\n\n",  crossingt_str);
QplugAppendTextDisplay(outstr);

// added by AJA Feb 2004
set_timetags((QdObject_var)n_var, dr[3].t); // copy  time tag

for (i=0; i<3; i++) {
  n[i]  = XX[i]/signedmodX;
 (* n_var)[i] = n[i];
}
   //  -- SEH, 28 April 2004 - CTA and COM calculations ------


   QdRMatrix_var n_cta  = new QdRMatrix(Dim(3), 0.00);
   QdRMatrix_var n_com  = new QdRMatrix(Dim(3), 0.00);
   QdRMatrix_var vn_cta = new QdRMatrix(Dim(4), 0.00);
   QdRMatrix_var vn_com = new QdRMatrix(Dim(4), 0.00);
   QdRMatrix_var d_cta  = new QdRMatrix(Dim(4), 0.00);
   QdRMatrix_var d_com  = new QdRMatrix(Dim(4), 0.00);
   QdRMatrix_var d_cva  = new QdRMatrix(Dim(4), 0.00);
   QdRMatrix_var tau    = new QdRMatrix(Dim(4), 0.00);
   QdRScalarSeq_var v_cta_curve =  new QdRScalarSeq(120);
   QdRScalarSeq_var v_cva_curve =  new QdRScalarSeq(120);
   QdRScalarSeq_var v_com_curve =  new QdRScalarSeq(120);
 
  QdObject_var tauS = (*call_list)[9];
   QdRMatrix_var tauVec =  QdRMatrix_var::narrow(tauS);   
   printf("Testing Tau \n");
   fflush(stdout);
   if (!tauVec.is_nil()) {
     printf("Tau = %g %g %g\n",(double)(*tauVec)[0],(double)(*tauVec)[1], (double)(*tauVec)[1]);
	 fflush(stdout);
   }
   QdString_var tau_string = QdString_var::narrow(tauS);
   if (!tau_string.is_nil()) n_cta = CTA_calc(drvec,dt,tau_string,
					      vn_cta, d_cta, 
					      vn_com, d_com, n_com, 
					      Vn_var, d_cva, n_var,
					      v_cva_curve,v_cta_curve,v_com_curve,
					      tidr,tau);
   printf("d_cta = %f\n",(double)(*d_cta)[0]);
   fflush(stdout);
   QuSetTxtAttr("FIELDNAM", "CTA_normal", n_cta);
   QuSetTxtAttr("SI_conversion", "1>()", n_cta);

   //  -- End of CTA and COM calculations ------


if(QuSlotDebug()){
   printf("Q4SC normal velocity is %g\n", (float)*Vn_var);

   printf("Q4SC B       is (%g, %g, %g)\n", B[0],B[1],B[2]);
   printf("Q4SC XX      is (%g, %g, %g)\n", XX[0],XX[1],XX[2]);
   printf("Q4SC norm    is %g\n",signedmodX);
   printf("Q4SC normal  is (%g, %g, %g)\n", n[0],n[1],n[2]);
   fflush(stdout);
}

// set the attributes for the output objects

QuSetTxtAttr("FIELDNAM", "CVA_normal", n_var);
QuSetTxtAttr("FIELDNAM", "Vn_CVA", Vn_var);

if(QuSlotDebug()){
   printf("Q4SC FIELDNAMs set\n");
   fflush(stdout);
}

QuSetTxtAttr("SI_conversion", "1>()", n_var);
char si_string[200];
sprintf(si_string, "%s s^-1",QuGetAttrText("SI_conversion",drin[3]));
//sprintf(si_string,"dummy SI insideQ4SC");
QuSetTxtAttr("SI_conversion",si_string,Vn_var);

if(QuSlotDebug()){
   printf("Q4SC SI_conversions set\n");
   fflush(stdout);
}

QuSetTxtAttr("UNITS", " ", n_var);
char units_str[200];
sprintf(units_str,"%s/s",QuGetAttrText("UNITS",drin[3]));
//sprintf(si_string,"dummy UNITS insideQ4SC");
QuSetTxtAttr("UNITS",units_str,Vn_var);

if(QuSlotDebug()){
   printf("Q4SC UNITS set\n");
   fflush(stdout);
}

QuSetTxtAttr("Crossing Time",crossingt_str,Vn_var);
QuSetTxtAttr("Crossing Time",crossingt_str,n_var);
// --- SEH CTA ---
QuSetTxtAttr("Crossing Time",crossingt_str,n_cta);

char frame_str[200];
sprintf(frame_str,"%s",QuGetAttrText("Frame",drin[3]));
QuSetTxtAttr("Frame",frame_str,n_var);
QuSetTxtAttr("Frame",frame_str,n_cta);

QuSetTxtAttr("UNITS",units_str,v_cta_curve);
QuSetTxtAttr("UNITS",units_str,v_cva_curve);
QuSetTxtAttr("UNITS",units_str,v_com_curve);

call_list->push_back( (QdObject_var) n_var);     
// --- SEH CTA ---
call_list->push_back( (QdObject_var) n_cta);     
call_list->push_back( (QdObject_var) Vn_var);     
call_list->push_back( (QdObject_var) v_cta_curve);     
call_list->push_back( (QdObject_var) v_cva_curve);     
call_list->push_back( (QdObject_var) v_com_curve);     

if(QuSlotDebug()){
   printf("Push back OK\n");
   fflush(stdout);
}

// --- SEH 28 April : modifications to display CTA results ---
// put some results to plug-in text window
  sprintf(outstr,"n_cva : [%10.4g  %10.4g  %10.4g] ", n[0], n[1], n[2]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_cva :  %10.1f %s\n", (float)*Vn_var, QuGetAttrText("UNITS",Vn_var));
  QplugAppendTextDisplay(outstr);
  double d_cva_ave=0.00;
  double v_cta_ave=0.00;
  double v_com_ave=0.00;
  double d_com_ave=0.00;
  
  for (i=0; i<4; i++) {
    d_cva_ave += 0.25 * (*d_cva)[i] ;
    v_cta_ave += 0.25 * (*vn_cta)[i];
    v_com_ave += 0.25 * (*vn_com)[i];
    d_com_ave += 0.25 * (*d_com)[i];
  }
  sprintf(outstr,"d_cva  :  %10.1f  %10.1f  %10.1f  %10.1f  (%10.1f) km\n\n",
	  (double)(*d_cva)[0],(double)(*d_cva)[1],(double)(*d_cva)[2],(double)(*d_cva)[3],d_cva_ave);
  QplugAppendTextDisplay(outstr);
 

  sprintf(outstr,"n_cta  : [%10.4g  %10.4g  %10.4g]\n",(double)(*n_cta)[0],(double)(*n_cta)[1],(double)(*n_cta)[2]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_cta  :  %10.2f  %10.2f  %10.2f  %10.2f  (%10.2f) km/s\n",
	  (double)(*vn_cta)[0],(double)(*vn_cta)[1],(double)(*vn_cta)[2],(double)(*vn_cta)[3],v_cta_ave);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"d_cta  : %10.2f  km\n\n",(double)(*d_cta)[0]);
  QplugAppendTextDisplay(outstr);

  
  sprintf(outstr,"n_com  : [%10.4g  %10.4g  %10.4g]\n",(double)(*n_com)[0],(double)(*n_com)[1],(double)(*n_com)[2]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_com  :  %10.2f  %10.2f  %10.2f  %10.2f  (%10.2f) km/s\n",
	  (double)(*vn_com)[0],(double)(*vn_com)[1],(double)(*vn_com)[2],(double)(*vn_com)[3],v_com_ave);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"d_com  :  %10.1f  %10.1f  %10.1f  %10.1f  (%10.1f) km\n\n",
	  (double)(*d_com)[0],(double)(*d_com)[1],(double)(*d_com)[2],(double)(*d_com)[3],d_com_ave);
  QplugAppendTextDisplay(outstr);
   

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus Q4SCPosition(QrVector *vecp,
                              QdObject_var pos_in,
			      QdTimeInterval_var time_int){
/**********************************************************************/

long  check_rtn;
QdRMatrixSeq_var pos_in_seq;
QdRMatrix_var    pos_in_const;
QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vecp a sequence of 1 cartesian vector whose timetag
// is the centre of the time interval time_int, and whose frame is
// that of pos_in. 

pos_in_seq       = QdRMatrixSeq_var::narrow(pos_in);
pos_in_const     = QdRMatrix_var::narrow(pos_in);

if( !pos_in_seq.is_nil() ) {
  //  QplugAppendTextDisplay ("Q4SCCoords>> Vector Sequence. Interpolating ...\n");
  ret_status = Q4SCCentrePosition(vecp,pos_in_seq,time_int);
  if (ret_status != QPLUG_SUCCESS){
   QplugAppendTextDisplay("Q4SCCoords>> Failed to calculate crossing position\n");
   return QPLUG_FAILURE;
  }
} else if (!pos_in.is_nil()){
  ret_status = Q4SCConstPosition(vecp, pos_in_const, time_int);
  if (ret_status != QPLUG_SUCCESS){
    QplugAppendTextDisplay("Q4SC>> Failed to calc const. position\n");
    return QPLUG_FAILURE;
  }
} else {
  QplugAppendTextDisplay("Q4SC>> pos_in neither const nor sequence!?\n");
  return QPLUG_FAILURE;
}


return QPLUG_SUCCESS;
}

/**********************************************************************/
QplugReturnStatus  Q4SCConstPosition(QrVector *vecp, 
                                    QdRMatrix_var pos_in, 
				    QdTimeInterval_var time_int){
/**********************************************************************/

QplugReturnStatus ret_status;
const char * xyz_comp[] = {"x", "y", "z"};
int i;

// grab constant vector and fill vecp.

if (!QuIsThreeVector(pos_in)){
   QplugAppendTextDisplay("Q4SC>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

QdTimeSeq_var tt = new QdTimeSeq(1, 0.5*(time_int->start()+time_int->end()));
vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vecp->ptr)->copy_xrefs_from(pos_in);
QdRMatrixSeq_var pos_seq = new QdRMatrixSeq(1, (*pos_in) );
pos_seq->copy_xrefs_from(pos_in);
for (i=0; i<3; i++){
  QdRScalarSeq_var comp = 
     QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_seq,xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("Q4SC>> Const pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  (vecp->ptr)[0][i] = comp[0];
}
vecp->n_recs = 1;
vecp->t = tt;

vecp->in_frame = QrGetFrameFromAttr(pos_in);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Const pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

return QPLUG_SUCCESS;
}


/**********************************************************************/
QplugReturnStatus  Q4SCCentrePosition(QrVector *vecp, 
                                    QdRMatrixSeq_var pos_in_seq, 
				    QdTimeInterval_var time_int){
/**********************************************************************/

int i,nttin,ntrue,il,ih;
QplugReturnStatus ret_status;
QdTime tc;
double cl,ch, frac;
const char * xyz_comp[] = {"x", "y", "z"};

if (!QuIsThreeVector(pos_in_seq)){
   QplugAppendTextDisplay("Q4SC>> Sequence pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

// prepare vecp 

tc = 0.5*(time_int->start()+time_int->end());
QdTimeSeq_var tt = new QdTimeSeq(1,tc );
vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vecp->ptr)->copy_xrefs_from(pos_in_seq);
vecp->t = tt;
vecp->n_recs = 1;
vecp->in_frame = QrGetFrameFromAttr(pos_in_seq);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Sequence pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

// linearly interpolate crossing position from bracketting time tags
// If any problems, default to 1st vector in pos_in_seq
// There are probably much simpler ways to do this. Note, though, that
// Qrtn routines only give correct answers if given x,y,z representation

QdTimeSeq_var tt_in = get_timetags(pos_in_seq);
if (tt_in.is_nil()){
  QplugAppendTextDisplay(
     "Q4SC>> No timetags for input pos seq. Taking 1st pt\n");
     il = ih = 0;
} else {
  nttin = tt_in->size();
  if (nttin <= 0){
     QplugAppendTextDisplay("Q4SC>> Input seq 0 length\n");
     return QPLUG_FAILURE;
  } else if (nttin == 1){
     il = ih = 0;
  } else {
     SequenceMask msk(tt_in->get_mask( QdTimeSeq::LE( tc)));
     ntrue = msk.num_true();
     if (ntrue <= 0) il = ih = 0;
     else {
       il = ntrue -1;
       ih =  (ntrue < nttin) ?   il + 1 : il; 
     }  
  }
}

if(ih == il) frac = 0.0;   // safe
else frac = (tc-tt_in[il])/(tt_in[ih] - tt_in[il]); // calc lin interpol factor

/* printf("tl: %s\ntc: %s\nth: %s\nfrac = %g\n",
                tt_in[il].iso_srep().c_str(),
		tc.iso_srep().c_str(),
		tt_in[ih].iso_srep().c_str(),
		frac);
*/

for (i=0; i<3; i++){                          // fill crossing position
  QdRScalarSeq_var comp = 
     QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_in_seq,xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("Q4SC>> Centre Pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  if (ih==il) {
    vecp->ptr[0][i] = (double)comp[il];
  } else {
    cl = (double)comp[il];
    ch = (double)comp[ih];
    vecp->ptr[0][i] = cl + frac*(ch-cl);
  }
}

return QPLUG_SUCCESS;
}



/**************************************************************************/
void QrVector_diag_print(QrVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.4f, %9.4f, %9.4f) at %s\n",
     (double)vecp->ptr[0][0], (double)vecp->ptr[0][1], (double)vecp->ptr[0][2],
     vecp->t[0].iso_srep().c_str());
	 fflush(stdout);
  return;
}



MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]){

double det;


det = determinant(a);
if (fabs(det) < MAT_INV_EPS) return MAT_INV_SINGULAR;

inverse[0][0] =  ( a[1][1]*a[2][2] - a[1][2]*a[2][1] )/det;
inverse[0][1] = -( a[0][1]*a[2][2] - a[0][2]*a[2][1] )/det;
inverse[0][2] =  ( a[0][1]*a[1][2] - a[0][2]*a[1][1] )/det;
inverse[1][0] = -( a[1][0]*a[2][2] - a[1][2]*a[2][0] )/det;
inverse[1][1] =  ( a[0][0]*a[2][2] - a[0][2]*a[2][0] )/det;
inverse[1][2] = -( a[0][0]*a[1][2] - a[0][2]*a[1][0] )/det;
inverse[2][0] =  ( a[1][0]*a[2][1] - a[1][1]*a[2][0] )/det;
inverse[2][1] = -( a[0][0]*a[2][1] - a[0][1]*a[2][0] )/det;
inverse[2][2] =  ( a[0][0]*a[1][1] - a[0][1]*a[1][0] )/det;

return MAT_INV_OK;


}



double determinant(double m[3][3]){

double d;

d  = m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] +m [0][2]*m[1][0]*m[2][1];
d -= m[2][0]*m[1][1]*m[0][2] + m[2][1]*m[1][2]*m[0][0] + m[2][2]*m[1][0]*m[0][1];

return d;
}

void multiply_mat(double a[3][3],double b[3][3], double prod[3][3]){
int i,j,k;

for (i=0; i<3; i++){
  for (j=0; j<3; j++){
    for (prod[i][j]=0.0,k=0; k<3; k++) prod[i][j]+=a[i][k]*b[k][j];
  }
}

return;

}

void multiply_mat_vec(double a[3][3],double b[3], double prod[3]){
int i,j;

for (i=0; i<3; i++){
    for (prod[i]=0.0,j=0; j<3; j++) prod[i]+=a[i][j]*b[j];
}


return;

}

void print_mat(double a[3][3]){

int i,j;

for (i=0; i<3; i++){
   for (j=0; j<3; j++) printf("%12.9f ", a[i][j]);
   printf("\n");
   fflush(stdout);
}
return;
}


// -------------------------------------------------------------------
/*!
  \brief Parses a text and converts text to a 4 element double array
*/
// -------------------------------------------------------------------
QdRMatrix_var str2Vec(QString s)
{
    QdRMatrix_var v = new QdRMatrix(Dim(4), 0.00);;
    QString str=s.simplifyWhiteSpace();
    str=str.replace(QChar('['),"").replace(QChar(']'),"").replace(QChar(',')," ");
    str=str.simplifyWhiteSpace();
    (*v)[0] = str.section(' ',0,0).toDouble();
    (*v)[1] = str.section(' ',1,1).toDouble();
    (*v)[2] = str.section(' ',2,2).toDouble();
    (*v)[3] = str.section(' ',3,3).toDouble();
    return(v);
}





// -------------------------------------------------------------------
/*!
  \brief CTA calculations
  \return a CTA normal and a velocity for each of the 4 SC
  \param A - 3 x 3 matrix with separation distances
  \param B - 3 x 1 vector with relative timing
  \param taustring - string with the 4 Tau values
  \param vn - 4 x 1 matrix with normal velocities for each of the 4 SC
  \param d_cta - 4 x 1 matrix with thicknesses for each of the 4 SC
  \param n_cva - CVA normal - used for combined method
  \param v_cta - velocity curve from CTA
  \param v_ctam - velocity curve from combined CTA, CVA
  \param tint - time interval (used to set timetags on velocity curves)

  Calculates a CTA (Constant Thickness Approach) boundary normal, normal
  velcoity and thickness. See Haaland et al, Annales Geophysicae, no 4, 2004
  for details.
*/
//
// SEH  May, 2004 
//
// -------------------------------------------------------------------
QdRMatrix_var CTA_calc(double dr[4][3], double dt[4], QdString_var taustring, 
		       QdRMatrix_var vn_cta, QdRMatrix_var d_cta, 
		       QdRMatrix_var vn_com, QdRMatrix_var d_com, QdRMatrix_var n_com, 
		       QdRScalar_var vn_cva, QdRMatrix_var d_cva, QdRMatrix_var n_cva, 
		       QdRScalarSeq_var v_cva_curve,QdRScalarSeq_var v_cta_curve, QdRScalarSeq_var v_com_curve, 
		       QdTimeInterval_var tint, QdRMatrix_var tau)
{
  int i,j;
  QdRMatrix_var n_cta =  new QdRMatrix(Dim(3), 0.00);
  QdRMatrix_var vn_instant =  new QdRMatrix(Dim(4), 0.00); 
  QString Qtau(taustring->c_str());
  double tc[4] = {dt[0],dt[1],dt[2],dt[3]};
  double A[3][3];
  double B[3];
  int SC_id[4] = {1,2,3,4};
  char outstr[200];

   tau =  str2Vec(Qtau);
if(QuSlotDebug()){
  printf(" ---------------------------------------------------------------------------\n");
  printf("SCID before : %d  %d  %d  %d\n",SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  printf("tc Before   : %10.2f  %10.2f  %10.2f  %10.2f\nR :\n", tc[0],tc[1],tc[2],tc[3]);
  for (i=0; i<3; i++) printf("%10.2f  %10.2f  %10.2f  %10.2f\n",dr[0][i],dr[1][i], dr[2][i], dr[3][i]);
  printf("Tau before  : %10.2f  %10.2f  %10.2f  %10.2f\n\n", (double)(*tau)[0], (double)(*tau)[1], (double)(*tau)[2],(double)(*tau)[3]);
  fflush(stdout);
}
 
  //  -- step 1 - sort in ascending order of crossing t0=0, t1 < t2 < t3 -----
 

  double tmp;
  double tmpa[3];
  long tmpID;
  for (i=1; i<4; i++) {
    for (j=3; j>=i; j--) {
      printf("Sort : j= %d\n",j);
	  fflush(stdout);
      if (tc[j-1] > tc[j]) {
	tmp = tc[j];
	tc[j] = tc[j-1];
	tc[j-1] = tmp;

       tmp=(double)(*tau)[j];
       (*tau)[j] = (*tau)[j-1];
       (*tau)[j-1] = tmp;
	
       tmpID =  SC_id[j];
       SC_id[j] =  SC_id[j-1];    
       SC_id[j-1] = tmpID;
       
       for (int k=0; k<3; k++) {
	 tmpa[k] = dr[j][k];
	 dr[j][k] = dr[j-1][k];
	 dr[j-1][k] = tmpa[k];
       }
      }
    }
  }


if(QuSlotDebug()){
  printf("SCID after : %d  %d  %d  %d\n",SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  printf("tc After   : %10.2f  %10.2f  %10.2f  %10.2f\nR:\n", tc[0],tc[1],tc[2],tc[3]);
  for (i=0; i<3; i++) printf("%10.2f  %10.2f  %10.2f  %10.2f\n",dr[0][i],dr[1][i], dr[2][i], dr[3][i]);
  printf("tau after  : %10.2f  %10.2f  %10.2f  %10.2f\n", (double)(*tau)[0], (double)(*tau)[1], (double)(*tau)[2],(double)(*tau)[3]);
  printf(" -----------------------------------------------------------------------------------\n\n");
  fflush(stdout);
}
 
  sprintf(outstr,"SC_id :  %10d    %10d    %10d    %10d\n", 
	  SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"tc     :   %10.2f  %10.2f  %10.2f  %10.2f\n", 
	  tc[0], tc[1], tc[2], tc[3]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"2Tau  :  %10.2f  %10.2f  %10.2f  %10.2f\n---------------------------------------------------------", 
	  (double)(*tau)[0], (double)(*tau)[1], (double)(*tau)[2], (double)(*tau)[3]);
  QplugAppendTextDisplay(outstr);
 

  QdRMatrix_var overtau = new QdRMatrix(Dim(4), 0.00);
  QdRMatrix_var Avec  = new QdRMatrix(Dim(4), 0.00);
  QdRMatrix_var TI  = new QdRMatrix(Dim(4,4), 1.00);
  QdRMatrix_var invTI  = new QdRMatrix(Dim(4,4), 0.00);
  QdRMatrix_var RdotM =  new QdRMatrix(Dim(3), 0.00);
  QdRMatrix_var R =  new QdRMatrix(Dim(3,3), 0.00);
  QdRMatrix_var invR =  new QdRMatrix(Dim(3,3), 0.00);
  QdRMatrix_var m =  new QdRMatrix(Dim(3), 0.00);
  QdRMatrix_var n =  new QdRMatrix(Dim(3), 0.00);

  // -- step 2 fix separation R matrix and invert this ----

  for (i=1; i<4; i++) 
    for (j=0; j<3; j++) (*R)[i-1][j] = dr[i][j] - dr[0][j];  // R = transposed of A
  
  for (i=4; i>=0; i--) tc[i] -= tc[0];   
  printf("tc After2   : %10.2f  %10.2f  %10.2f  %10.2f\nR:\n", tc[0],tc[1],tc[2],tc[3]);
  fflush(stdout);

  invR = QdRMatrix_var::narrow(invert(R));


  //  -s step 3 fill inn TI matrix and invert ------


  for (j=0; j<4; j++){
    (*TI)[1][j] = tc[j];
    (*TI)[2][j] = pow(tc[j],2.0) + pow((double)(*tau)[j],2.0)/3.0; 
    (*TI)[3][j] = pow(tc[j],3) + (tc[j]* pow((double)(*tau)[j],2)); 
  }
     
  invTI = QdRMatrix_var::narrow( invert(TI) );
  for (i=0; i<4; i++) (*overtau)[i] = 1.0/(2.0*(double)(*tau)[i]);


  Avec = QdRMatrix_var::narrow (overtau * invTI);

  // --- calculate R dot N -- se eq (4) in Haaland et al, Ang 2, 2004 ---

  for (i=0; i<3; i++){
    (*RdotM)[i] = (double)(*Avec)[0]*tc[i+1] + 
                  (double)(*Avec)[1]*(pow(tc[i+1],2)/2.0) + 
                  (double)(*Avec)[2]*(pow(tc[i+1],3)/3.0) + 
		  (double)(*Avec)[3]*(pow(tc[i+1],4)/4.0);
  }

  for (i=0; i<3; i++) 
    for (j=0; j<3; j++) 
      (*m)[i] += (double)(*invR)[i][j] * (double)(*RdotM)[j];

  double sign = ((*m)[0] < 0 ) ? -1 : 1;  

  double d = 1.0/vecsum(m);
  for (i=0; i<3; i++) 
    (*n)[i] = sign * (*m)[i] * d;

  (*vn_instant)[0] = d * (*Avec)[0];
  (*vn_cta)[0]  =  sign * ((*vn_instant)[0] + d*(*Avec)[2]*pow((*tau)[0],2)/3);
  for (j=1; j<4; j++) {
    (*vn_instant)[j] = d * ((*Avec)[0] + (*Avec)[1]*tc[j] + (*Avec)[2]*pow(tc[j],2) + (*Avec)[3]* pow(tc[j],3));
    (*vn_cta)[j]  = sign * ((*vn_instant)[j] + d*(*Avec)[2]*pow((*tau)[j],2)/3 + d*(*Avec)[3]*tc[j]*pow((*tau)[j],2));
  }


if(QuSlotDebug()){
  printf("TI = \n");
  fflush(stdout);
  for (j=0; j<4; j++){
    printf("%10.2f  %10.2f  %10.2f  %10.2f\n",
	   (double) (*TI)[0][j],
 	   (double) (*TI)[1][j],
 	   (double) (*TI)[2][j],
 	   (double) (*TI)[3][j]);

  }

  printf("invTI = \n");
  for (j=0; j<4; j++){
    printf("%10.4f  %10.4f  %10.4f  %10.4f\n",
	   (double) (*invTI)[0][j],
 	   (double) (*invTI)[1][j],
 	   (double) (*invTI)[2][j],
 	   (double) (*invTI)[3][j]);
  }

  printf("R = \n");
  for (j=0; j<3; j++){
    printf("         %10.2f  %10.2f  %10.2f\n",
	   (double) (*R)[0][j],
 	   (double) (*R)[1][j],
 	   (double) (*R)[2][j]);
  }

  printf("invR1R2 = \n");
  for (j=0; j<3; j++){
    printf("         %10.6f  %10.6f  %10.6f\n",
	   (double) (*invR)[0][j],
 	   (double) (*invR)[1][j],
 	   (double) (*invR)[2][j]);
  }


  printf("OverTau_CTA = \n");
  for (j=0; j<4; j++) printf("%10.4f\n",(double)(*overtau)[j]);
  printf("\n");

  printf("Avec_CTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",(double)(*Avec)[j]);
  printf("\n");

  printf("RdotM_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",(double)(*RdotM)[j]);
  printf("\n");

  printf("m_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",(double)(*m)[j]);
  printf("\n");

  printf("d_CTA = %10.4f\n",d);

  printf("n_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",(double)(*n)[j]);
  printf("\n");

  printf("V_CTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",(double)(*vn_instant)[j]);
  printf("\n");

  printf("VaveCTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",(double)(*vn_cta)[j]);
  printf("\n");
  fflush(stdout);
}

  for (i=0; i<4; i++) (*d_cta)[i] = d;

  // --- velocit curves - start a few steps before, and end a few steps later

  int Ntags = 120;
  double tstep = (fabs(tc[3])) / (double) (Ntags-20);
  double t_curvestart = -10*tstep;
  double t = t_curvestart;
  
  QdTimeSeq_var tt = new QdTimeSeq();
  tt->assign_regular(Ntags, tint->start()+Seconds(t), Seconds(tstep) );  

  for (i=0; i<Ntags; i++) {
    (*v_cta_curve)[i] =     sign * d * ((*Avec)[0] + (*Avec)[1]*t + (*Avec)[2]*pow(t,2) + (*Avec)[3]* pow(t,3));
    (*v_cva_curve)[i] =     (double) (*vn_cva);
    t+= tstep;
  }

  // check that d is really constant ---

//   double di = 0;
//   double ti[4] = {0.0, B[0], B[1], B[2]};
//   for (i=0; i<4; i++) {
//     di = 2*(*tau)[i] * ((*Avec)[0] + (*Avec)[1]*ti[i] + (*Avec)[2]*pow(ti[i],2) + (*Avec)[3]* pow(ti[i],3) + ((*Avec)[2]*pow((*tau)[i],2))/3.0 + (*Avec)[3]*ti[i]*pow((*tau)[i],2));
//     printf("i = %d, di = %f\n",i,di);
//   }

  printf("Sign = %f\n",sign);
  fflush(stdout);
  set_timetags((QdObject_var) v_cta_curve,tt);
  set_timetags((QdObject_var) v_cva_curve,tt);


  // -------------------------------------------------------------------------------------------------
  //
  //      combined CTA/CVA as proposed by BS, 26 May, 2004 
  //
  // -------------------------------------------------------------------------------------------------


  //  double theta = acos((*n_cva)[0] * (*n)[0] + (*n_cva)[1] * (*n)[1] +(*n_cva)[2] * (*n)[2]);
  
  for (i=0; i<3; i++) (*n_com)[i] = ((double)(*n)[i] + (double)(*n_cva)[i]) / 2.0;
  double normlen=0.00;
    for (i=0; i<3; i++) normlen += (*n_com)[i] * (*n_com)[i];
    for (i=0; i<3; i++) (*n_com)[i] /= sqrt(normlen);

  double k[4] = {0.0,  0.0,  0.0,  0.0};
  double aveTau=0.0, aveU=0.00, u[4];
  double ti[4] = {tc[0], tc[1], tc[2], tc[3]};

  for (i=0; i<4; i++) {
    u[i] = (*tau)[i] *  (1.0 
	       - (2.0 * ti[i]*(1.0/ti[1] + 1.0/ti[2] + 1.0/ti[3]))
	       + (3.0 * (pow(ti[i],2) + pow((*tau)[i],2)/3.0) * (1.0/(ti[1]*ti[2]) + 1.0/(ti[1]*ti[3]) + 1.0/(ti[2]*ti[3]))) 
	       - (4.0 * (pow(ti[i],3) + ti[i]*pow((*tau)[i],2)) / (ti[1]*ti[2]*ti[3])));
    aveU += 0.25 * u[i];
    aveTau += 0.25 * (*tau)[i];   
    printf("i = %d,  ui= %f\n",i,u[i]);
	fflush(stdout);
  }    

  double ave_Tau3     = 0.00;
  double ave_t_Tau    = 0.00;
  double ave_t_Tau3   = 0.00;
  double ave_t2_Tau   = 0.00;
  double ave_t3_Tau   = 0.00;
  double ave_U_Tau    = 0.00;
  double ave_U_Tau3   = 0.00;
  double ave_U_t_Tau  = 0.00;
  double ave_U_t_Tau3 = 0.00;
  double ave_U_t2_Tau = 0.00;
  double ave_U_t3_Tau = 0.00;

  for (i=0; i<4; i++) {
    ave_Tau3      += 0.25 * (pow((*tau)[i],3.0));
    ave_t_Tau     += 0.25 * (ti[i] * (*tau)[i]);
    ave_t2_Tau    += 0.25 * (pow(ti[i],2) * (*tau)[i]);
    ave_t3_Tau    += 0.25 * (pow(ti[i],3) * (*tau)[i]);
    ave_t_Tau3    += 0.25 * (ti[i] * pow((*tau)[i],3));
    ave_U_Tau     += 0.25 * (u[i] * (*tau)[i]);
    ave_U_Tau3    += 0.25 * (u[i] * pow((*tau)[i],3));
    ave_U_t_Tau   += 0.25 * (u[i] * ti[i] * (*tau)[i]);
    ave_U_t_Tau3  += 0.25 * (u[i] * ti[i] * pow((*tau)[i],3));
    ave_U_t2_Tau  += 0.25 * (u[i] * pow(ti[i],2) * (*tau)[i]);
    ave_U_t3_Tau  += 0.25 * (u[i] * pow(ti[i],3) * (*tau)[i]);
  }
  
  printf(" ave_Tau3     = %f\n", ave_Tau3 );
  printf(" ave_t_Tau    = %f\n", ave_t_Tau );
  printf(" ave_t_Tau3   = %f\n", ave_t_Tau3 );
  printf(" ave_U_Tau    = %f\n", ave_U_Tau );
  printf(" ave_U_Tau3   = %f\n", ave_U_Tau3 );
  printf(" ave_U_t_Tau  = %f\n", ave_U_t_Tau );
  printf(" ave_U_t2_Tau = %f\n", ave_U_t2_Tau );
  printf(" ave_U_t3_Tau = %f\n", ave_U_t3_Tau );

  k[0] = 1.0/aveU * (ave_U_Tau - aveU * aveTau);
  k[1] = 1.0/aveU * (ave_U_t_Tau - aveU * ave_t_Tau);
  k[2] = 1.0/aveU * (ave_U_t2_Tau - aveU * ave_t2_Tau + ave_U_Tau3/3.0 - (aveU * ave_Tau3)/3.0);
  k[3] = 1.0/aveU * (ave_U_t3_Tau - aveU * ave_t3_Tau + ave_U_t_Tau3 - aveU * ave_t_Tau3);

  printf("Avetau = %f\n",aveTau);
  printf("AveU = %f\n",aveU);
  printf("k = \n");
  for (j=0; j<4; j++) printf("%15.10f;\n",k[j]);
  printf("\n");
  fflush(stdout);

  QdRMatrix_var MTI  = new QdRMatrix(Dim(4,4), 0.00);
  QdRMatrix_var invMTI  = new QdRMatrix(Dim(4,4), 0.00);
  QdRMatrix_var RdotN =  new QdRMatrix(Dim(4), 0.00);


  for (j=0; j<3; j++) {
    (*MTI)[0][j] = ti[j+1];
    (*MTI)[1][j] = pow(ti[j+1],2) / 2.0;
    (*MTI)[2][j] = pow(ti[j+1],3) / 3.0;
    (*MTI)[3][j] = pow(ti[j+1],4) / 4.0;
  }

  for (i=0; i<4; i++) (*MTI)[i][3] = k[i];

  invMTI = QdRMatrix_var::narrow ( invert(MTI) );


  // --- calculate 4 elem R dot N --------

  for (j=0; j<3; j++){
    (*RdotN)[j] = (*R)[j][0]* (*n_com)[0] +  (*R)[j][1]* (*n_com)[1] + (*R)[j][2]*(*n_com)[2];
  }
   
  printf("Rdot n = [%g  %g  %g]\n",(double)(*RdotN)[0], (double)(*RdotN)[1],(double)(*RdotN)[2]) ;
fflush(stdout);
  Avec  = new QdRMatrix(Dim(4), 0.00); 
  Avec = QdRMatrix_var::narrow( RdotN * invMTI);


  for (i=0; i<4; i++) {
    (*d_com)[i] = sign * 2 * (*tau)[i] * ((*Avec)[0] + (*Avec)[1]*ti[i] + (*Avec)[2]*pow(ti[i],2) + (*Avec)[3]* pow(ti[i],3) + ((*Avec)[2]*pow((*tau)[i],2))/3.0 + (*Avec)[3]*ti[i]*pow((*tau)[i],2));
    printf("i = %d, di = %f\n", i, (double)(*d_com)[i]);
	fflush(stdout);
    (*d_cva)[i] = sign * 2.0*(*tau)[i] * (double) (*vn_cva);
  }

  // --- The 4 V_com velocitiese ----
  (*vn_instant)[0] = (*Avec)[0];
  (*vn_com)[0]  =  ((*vn_instant)[0] + (*Avec)[2]*pow((*tau)[0],2)/3);
  for (j=1; j<4; j++) {
    (*vn_instant)[j] = ((*Avec)[0] + (*Avec)[1]*tc[j-1] + (*Avec)[2]*pow(tc[j-1],2) + (*Avec)[3]* pow(tc[j-1],3));
    (*vn_com)[j]  = ((*vn_instant)[j] + (*Avec)[2]*pow((*tau)[j],2)/3 + (*Avec)[3]*tc[j-1]*pow((*tau)[j],2));
  }


if(QuSlotDebug()){
  printf("MTI = \n");
  for (j=0; j<4; j++){
    printf("%14.4f  %14.4f  %14.4f  %14.4f\n",
	   (double) (*MTI)[0][j],
 	   (double) (*MTI)[1][j],
 	   (double) (*MTI)[2][j],
 	   (double) (*MTI)[3][j]);
  }

  printf("invMTI = \n");
  for (j=0; j<4; j++){
    printf("%14.12f  %14.12f  %14.12f  %14.12f\n",
	   (double) (*invMTI)[0][j],
 	   (double) (*invMTI)[1][j],
 	   (double) (*invMTI)[2][j],
 	   (double) (*invMTI)[3][j]);
  }
  printf("AvecCOM = \n");
  for (j=0; j<4; j++) printf("%14.10f;\n",(double)(*Avec)[j]);
  printf("\n");

  printf("n_combined = \n");
  for (j=0; j<3; j++) printf("%14.10f\n",(double)(*n_com)[j]);
  printf("\n");
  fflush(stdout);
}
  
  // --- V_com curve ----
  t = t_curvestart;
  for (i=0; i<Ntags; i++) {
    (*v_com_curve)[i] =    ((*Avec)[0] + (*Avec)[1]*t + (*Avec)[2]*pow(t,2) + (*Avec)[3]* pow(t,3));
    t+= tstep;
  }
  set_timetags((QdObject_var) v_com_curve,tt);
  printf("at end : tau 0 = %f\n",(double)(*tau)[0]);
  fflush(stdout);

  // ---- output to /tmp/qsas_cvacta.txt ----
  FILE *f2;


    QdTimeSeq_var tt_out = get_timetags(v_cta_curve);
    QdTime timetag=(*tt_out)[0];

    QplugAppendTextDisplay("Log file written to /tmp/qsas_cva_cta.txt");
    f2=fopen("/tmp/qsas_cva_cta.txt","w"); 
    if(f2 == NULL) f2 = fopen("/dev/stdout", "w");

    QDateTime today = QDateTime::currentDateTime();
    QString logTxt;
    logTxt += "#QSAS CVA_CTA_COM run : "+today.toString() + "\n#\n";
    sprintf(outstr,"#SC_id   %10d    %10d    %10d    %10d\n",  SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
    logTxt += QString(outstr);
    sprintf(outstr,"#tc      %10.2f  %10.2f  %10.2f  %10.2f\n", tc[0], tc[1], tc[2], tc[3]);
    logTxt += QString(outstr);
    sprintf(outstr,"#tstep      %10.2f\n", tstep);
    logTxt += QString(outstr);
    sprintf(outstr,"#2Tau    %10.2f  %10.2f  %10.2f  %10.2f\n", (double)(*tau)[0], (double)(*tau)[1], (double)(*tau)[2], (double)(*tau)[3]);
    logTxt += QString(outstr);
    fprintf(f2,logTxt.latin1());
    fprintf(f2,"#Time = %s\n",timetag.iso_srep().c_str());
    fprintf(f2,"# \n");
    for (i=0; i<Ntags; i++)  {
      timetag = (*tt_out)[i];
      fprintf(f2, "%s  %10.4f  %10.4f  %10.4f \n", timetag.iso_srep().c_str(),
	     (double)(*v_cva_curve)[i],(double)(*v_cta_curve)[i],(double)(*v_com_curve)[i]);
    }
    fclose(f2);
  

  return (n);

}


// -------------------------------------------------------------------
/*!
  \brief returns the vector sum (length) of a vector
  \returns length of V
  \param V - 3 component vector
*/
// -------------------------------------------------------------------
double vecsum(QdRMatrix_var V)
{
  double sum=0.0;
  
  for (int i=0; i<3; i++) sum += (*V)[i] * (*V)[i];

  return(sqrt(sum));
}


int main()
{
}
