/* PGSpectro.h
*
* Header file for PGSpectro spectrogram, PGWheel, and 1D cut routines
*
* History
*
* July 2000  First draft  SJS
* Sept 2000  Tidy up
* Oct 2001   Add 1D cut routines
* Jan 2002   Fixed line thicknesses in XY Trace to make them ints
*
* Overview of use:
*
*    Allocate PGSpectro or PGWheel structure
*    PGSpectro_set_defaults or PGWheel_set_defaults sets:
*          numeric values to safe defaults
*          rainbow colours
*          scalings to linear
*          strings to " "
*          calls PGSpectro_domain_set_defaults or PGWheel_domain_set_defaults:
*             world coord limits to 1,10 in xy or 0,10 radius
*             ndc coord limits to nearly full page
*    Fill rest of PGSpectro or PGWheel structure
*          data arrays
*          string labels
*          wedge label
*    Open pgplot device
*    Call PGSpectro_plot or PGWheel_plot
*    Close pgplot device
*
*/

#ifndef _PGSPECTRO_
#define _PGSPECTRO_

#ifdef __cplusplus
extern "C"{
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NCOLOURS_MAX 256
#define LOWEST_COLOR_INDEX 16
#define PGSPECTRO_DEFAULT_CHAR_SIZE 0.7

#define DEG2RAD  0.0174532925
#define SKIP_LOG_VALUE -1.e31

typedef enum Yes_or_No_e{
  YES,
  NO
} Yes_or_No;

typedef enum PGSpectro_linlog_e{
  PGSPECTRO_LIN,
  PGSPECTRO_LOG
} PGSpectro_linlog;

typedef enum PGSpectro_col_e{
  PGSPECTRO_RAINBOW,
  PGSPECTRO_GRAY_WHITE_LOW,
  PGSPECTRO_GRAY_WHITE_HIGH
} PGSpectro_col;

typedef enum PGSpectro_ret_e{
  PGSPECTRO_OK,
  PGSPECTRO_NO_DEVICE,
  PGSPECTRO_MALLOC_FAILURE,
  PGSPECTRO_QJAS_INCONGRUENT,
  PGSPECTRO_SIZE_ERROR,
  PGSPECTRO_TRACE_ERROR
} PGSpectro_ret;


typedef enum QPGLine_colour_e{   /* set enum to match cpgsci arg */
  QPGWHITE = 0,
  QPGBLACK,
  QPGRED,
  QPGGREEN,
  QPGBLUE,
  QPGCYAN,
  QPGMAGENTA,
  QPGYELLOW,
  QPGORANGE,
  QPGYELLOWGREEN
} QPGLine_colour;

typedef enum QPGLinestyle_e{  /* set enum to match cpgsls arg */
  QPGSOLID = 1,
  QPGDASHED,
  QPGDASHDOT,
  QPGDOT,
  QPGDASHDOTDOTDOT
} QPGLinestyle;


typedef enum QPGSymbol_e{     /* set enum to match cpgpt arg */
  QPGPOINT          =  1,
  QPGPLUS           =  2,
  QPGSTAR           =  3,
  QPGCIRCLE         =  4,
  QPGCROSS          =  5,
  QPGSQUARE         =  6,
  QPGTRIANGLE       =  7,
  QPGDIAMOND        = 11,
  QPGFILLEDTRIANGLE = 13,
  QPGFILLEDSQUARE   = 16,
  QPGFILLEDCIRCLE   = 17,
  QPGFILLEDSTAR     = 18
} QPGSymbol;


typedef struct PGSpectro_domain_s{
  float xworldmin, xworldmax, yworldmin, yworldmax;
  float xndcmin, xndcmax, yndcmin, yndcmax;
} PGSpectro_domain;

typedef struct PGSpectro_wedge_s{
  float disp;   /* displacement from r edge in char height */
  float width;  /* width in char height */
  char * label; /* label for wedge */
} PGSpectro_wedge;

typedef struct PGSpectro_s{

  double *z;                   /* pntr to array z[nx][ny] floats in C-order, */
                              /* that is z = &z[0][0]                       */
//  double *xbinlow, *xbinhigh;  /* pntr to arrs of size = nx for bin bndaries */
  double *ybinlow, *ybinhigh;  /* ditto for y                                */
  int nx, ny;

  double zscalemin, zscalemax;        /* min max values for color table      */
  double zmissing;                    /* value to be treated as missing if   */
  Yes_or_No skip_missing;            /* skip_missing set to YES             */
  char *xlabel, *ylabel, *title;     /* plot labels;set to " " for no label */
  char *subtitle;                    /* smaller, underneath main title      */
  PGSpectro_col color_table;         /* see enum; rainbow, b->w and w->b    */
  PGSpectro_domain domain;           /* world and ndc coords                */
  PGSpectro_wedge wedge;             /* params for color bar key            */
  PGSpectro_linlog xlinlog,ylinlog,zlinlog; /*see enum; scalings for x, y, z*/
  Yes_or_No xnums, ynums;                   /* YES, NO  numeric tick labels */
  Yes_or_No ylabel_vert;                    /* YES for upright text         */
  Yes_or_No draw_wedge;                     /* NO to suppress color bar     */
/* Time x axes only: */
  Yes_or_No timex;    /* YES, NO for x <=> time; if YES LOG ignored */
                      /* Times (incl xworlds) in secs of day from datestart */
  char *datestart, *dateend;         /* date strings, e.g., 1984-13-07      */
  double char_size;                   /* pgplot char size for labels, .. */
} PGSpectro;

// typedef struct PGWheel_domain_s{
//   double rmin, rmax;
//   double xndcmin, xndcmax, yndcmin, yndcmax;
// } PGWheel_domain;
// 
// 
// typedef struct PGWheel_s{
// 
//   double *z;                   /* pointer to array z[nr][nth] floats in C-order,  */
//                               /* that is z = &z[0][0] */
//                               /************** NOTE **************************/
//                               /* ALL ANGLES in DEGREES !!!!!!!!!!!!!!!!!!!!!*/
//   double *rbinlow, *rbinhigh;  /* pntr to arrays of size = nr for bin bndries*/
//   double *thbinlow, *thbinhigh; /* ditto for theta */
//   int nr, nth;
// 
//   float *rcircles;            /* r values for ncircles reference circles    */
//   int ncircles;
// 
//   double zscalemin, zscalemax;      /* min max values for color table        */
//   double zmissing;                  /* value to be treated as missing if     */
//   Yes_or_No skip_missing;          /* skip_missing set to YES               */
//   char *xlabel, *ylabel, *title;   /* plot labels; set to " " for no label  */
//   char *subtitle, *subsubtitle;    /* smaller under main title              */
//   PGSpectro_col color_table;       /* see enum; rainbow, b->w and w->b      */
//   PGWheel_domain domain;           /* world & ndc coords; plot makes square */
//   PGSpectro_wedge wedge;           /* params for color bar key              */
//   PGSpectro_linlog rlinlog,zlinlog;/* see enum; scalings for r, z           */
//   Yes_or_No rnums;                 /* YES, NO for numeric tick labels       */
//   float title_just;                /* 0.0 - 1.0 for left to right just      */
// } PGWheel;
// 
// typedef struct PGXYtrace_s{
// 
// /* Note: tick lengths, title displacements, etc. are measured in multiples
// *  of the active character height, which here is taken from the appropriate
// *  context (title, labels, ...) 
// *
// *  Character heights are in multiples of the default ( ~ 1/40 H of view surf)
// *
// *  Line thicknesses are ints in range 1-200                                */
// 
// /* data and lines */
// 
//   double *x, *y;                    /* arrays to be plotted: plots y(x) */
//   int   npts;
//   QPGLine_colour colour;
//   QPGLinestyle   linestyle;
//   int line_width;                  /* 0 - 200 lines and drawn symbols*/
//   QPGSymbol symbol;
//   Yes_or_No YoNline, YoNsymbol;    /* choose line or symbol or both */
//   int symbol_size;                 /* units of char height */
// 
// /* layout */
// 
//   PGSpectro_domain plot_domain, legend_domain;  /* legend worlds ignored */
// 
// /* titles and labels - only applies if box drawn */
//   char *subtitle, *subsubtitle;    /* smaller under main title              */
//   char *xlabel, *ylabel, *title;   /* plot labels; set to " " for no label  */
//                                    /* or YoNbox = NO for nothing but data   */
//   Yes_or_No Xtextlab_bot, Ytextlab_left; /* only if num labs too but with   */
//                                          /* separate control of position    */
//   double title_just;                       /* 0.0 - 1.0 for l to r just      */
//   double title_disp, subtitle_disp, subsubtitle_disp; /* in char heights     */
//   double label_char_size, title_char_size; /* in units of 1/40 view surf H   */
//   int text_line_width;                  /* 1-200=pen-thick drawing chars  */
// 
// /* axes and box */
// 
//   Yes_or_No YoNbox;                 /* box. If no box, labels ignored      */
//   Yes_or_No YoNXnumlab, YoNYnumlab; /* YES, NO for numeric tick labels     */
//   Yes_or_No Ynumlab_left, Xnumlab_bot;          
//   PGSpectro_linlog xlinlog,ylinlog; /* see enum; scalings for x, y         */
//   double tick_length;                /* tick len=char height                */
//   int box_thickness; 		    /* thick<->line                        */
// 
// /* Time x axes only: decreasing time axes allowed */
//   Yes_or_No timex;    /* YES, NO for x <=> time; if YES LOG ignored */
//                       /* Times (incl xworlds) in secs of day from datestart */
//   char *datestart, *dateend;         /* date strings, e.g., 1984-13-07      */
//   
// /* legend */
//   double legend_text_char_size;     /* units of 1/40 view surf height        */
//   Yes_or_No YoNlegend,YoNlegendbox;/* legend? legendbox?  [for multiple]    */
//   double legend_just;               /* 0.0 - 1.0 for bottom to top           */
//   double legend_line_length_frac;   /* 0.0 - 1.0 for fraction of legend width*/
//                                    /*     occupied by line sample           */
//   char *legend_text;               /* letter thick = text_line_width: above */
// 
// } PGXYtrace;
// 
// /* prototypes: */
// 
// void set_pgp_rainbow(PGSpectro_col color_table);
// 
// void spectro_wedge(PGSpectro_wedge wedge, float zscalemin,
//            float zscalemax, PGSpectro_linlog lin_log);
// 
// void PGSpectro_set_defaults(PGSpectro *spectro);
// 
// void PGSpectro_domain_set_defaults (PGSpectro_domain *domainp);
// 
// PGSpectro_ret PGSpectro_Plot(PGSpectro *spectro);
// 
// void PGWheel_set_defaults(PGWheel *spectro);
// 
// void PGWheel_domain_set_defaults (PGWheel_domain *domainp);
// 
// PGSpectro_ret PGWheel_Plot(PGWheel *wheel);
// 
// void arc(float rlo, float rhi,  float thlo, float thhi);
// 
// PGSpectro_ret PGXYtrace_plot(PGXYtrace *t);
// 
// PGSpectro_ret PGXYtrace_legend(PGXYtrace *t);
// 
// PGSpectro_ret PGXYtrace_set_defaults(PGXYtrace *trace);
// 
// double SafeLog10(double x);

#ifdef __cplusplus
}
#endif


#endif  /* end #ifndef _PGSPECTRO_ */
