// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : EDI filter routine
// 
// This plugin takes an EDI vector (V or E) as inputs and
// filters it according to the settings in the given filter
// specifications file
//
// This file 'EDI_filter.cc' contains the QSAS-wrapper + minor
// error checking. Reading inputs from the filter settings
// file is handled by code in 'readfilterfile.c'. The actual
// filtering/sorting code can be found in 'DataOk.c'
//
// Inputs :
// 	edi_vector     	- vector time series with either E or V from EDI
//      edi_status      - the EDI status matrix
//      edi_chisqr      - EDI reduced ChiSquare results
//
// Outputs
//	edi_filtered    - filtered version of above
//
//
// SEH, 23 Oct, 2001
//   27. Oct 01 SEH : RedChiSqr is now floating point number
//   17. Nov 01 SEH : Output object named after input objects 
//   27. Nov 01 SEH : corrected naming in *metstr
//   29. Nov 01 SEH : Frame to vector>gse_xyz
//   02. Mar 02 SEH : New method RMT : expanded *,ethstr
//                    (see also DataOk + defs in EDI_filter.h)
//   08. Apr 02 SEH : Added support for TriErr, AngErr, Triout as
//                    signed bytes
//   28. Nov 03 SEH : QSAS 2 version; Changed header + 'external'
//   31. Oct 06 SEH : replaced 'Qu_is3Vector()' with cast and .is_nil() check
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "EDI_filter.h"


// comment out for QSAS 1.xx 
extern "C" QplugReturnStatus  EDI_filter  ( QplugArgList *);

// ----- Prototypes ----

 int DataOk(sbytes stat, sbytes settings);
 sbytes readfilterfile(const char *filename);
using namespace QSAS;

 QplugReturnStatus  EDI_filter  ( QplugArgList *call_list )
 {

 QdRMatrixSeq_var  edi_vector;
 QdRMatrixSeq_var  edi_temp;
 QdRMatrixSeq_var  edi_filtered;
 QdRMatrixSeq_var  edi_status;
 QdRScalarSeq_var  edi_chisqr;
 QdString_var      filename_input;
 char       txtbuf[200];
 sbytes     stat,settings;
 vec        EDI;
 long       n_vecs,n_stat,n_chis,n_left=0,i,j;
 int        REMOVE=0, ChiCheck=1;
 double     chisqr,fillval = -1.0E31;
 char       *str_out,*str_add,str_tot;
 const char *filename;      
                        
// --- get EDI vector -----

 edi_vector = QdRMatrixSeq_var::narrow((*call_list)[0]);
 QdTimeSeq_var tt_in = get_timetags(edi_vector);


 if ( edi_vector.is_nil() ) {	
   QplugAppendTextDisplay ( "ERROR - invalid EDI vector. Input not a Vector Time Series!\n");
   return QPLUG_FAILURE; 
 }        
 
 

 n_vecs = edi_vector->sequence_size();
 QdTimeSeq_var tt_temp = new QdTimeSeq(n_vecs);


// --- get Status matrix - a n x 7 matrix of bytes -----

 edi_status = QdRMatrixSeq_var::narrow((*call_list)[1]);
 if (strcmp(QuGetAttrText("FIELDNAM", edi_status),"EDI status") != 0) {
   QplugAppendTextDisplay ( "ERROR - invalid Status words! Input is not EDI status word \n");
   return QPLUG_FAILURE; 
 }

 n_stat = edi_status->sequence_size();
 printf("Testing ChiSqr ...\n");
 fflush(stdout);

// --- get Reduced ChiSqr scalar TS -----

 edi_chisqr = QdRScalarSeq_var::narrow((*call_list)[2]);
 if (!edi_chisqr.is_nil() && strcmp(QuGetAttrText("FIELDNAM", edi_chisqr),"EDI Reduced Chi-squared") == 0) {
   n_chis = edi_chisqr->sequence_size();
   printf("ChiSqr exists...\n");
   fflush(stdout);
 } 
 else {
   QplugAppendTextDisplay ( "WARNING - invalid or missing ChiSqr. Ignoring this filter \n");
   ChiCheck = 0;
   n_chis = n_vecs;
   printf("ChiSqr does not exists...\n");
   fflush(stdout);
 }

 if (n_vecs != n_stat && n_vecs != n_chis) {
   QplugAppendTextDisplay ( "ERROR - different number of records in VEC, Status and ChiSqr !\n");
   return QPLUG_FAILURE; 
 }        


filename_input = QdString_var::narrow((*call_list)[3]);
if( filename_input.is_nil() ) {
  QplugAppendTextDisplay("ERROR - no filter file specified\n");
  return QPLUG_FAILURE;
}

filename = filename_input->c_str();

// ---- Read filter definition file ----

 settings.quality = GOOD_CAUT_BAD;
 settings.method  = TRI_PMT_SMT;
 settings.ambigu  = YES_NO ;
 settings.angerr  = 255;
 settings.trierr  = 200;
 settings.triout  = 100;
 settings.redchi  = 1.0E6;

 settings = readfilterfile(filename);
 {
   printf(" Quality       :%d \n",settings.quality);
   printf(" Method        :%d \n",settings.method);
   printf(" Ambigu        :%d \n",settings.ambigu);
   printf(" TriErr        :%d \n",settings.trierr);
   printf(" AngErr        :%d \n",settings.angerr);
   printf(" TriOut        :%d \n",settings.triout);
   printf(" Chisqr        :%5.2f \n",settings.redchi);
   printf(" Filterfile OK :%d \n",settings.OK);
   fflush(stdout);
 }

 if (!settings.OK) {
   QplugAppendTextDisplay("\nWARNING >> Filter file not found - using defaults\n");
 }
 else {
   sprintf(txtbuf,"\nFilter file   : %s\n",filename);
   QplugAppendTextDisplay(txtbuf);
 }


/* ----------------------------------------  
     declare a temporary object to store the 
     filtered recs. Since we don't know the
     final lenght of the we have to allow
     place for all records
-------------------------------------------*/


 edi_temp = new QdRMatrixSeq(n_vecs,RMatrix(Dim(3)));


// ---- Scan through the records -------

 for(i=0; i<n_vecs; i++) {  

   stat.quality = (int) (edi_status[i][0]) &   3;
   stat.method  = (int) (edi_status[i][3]) &  96;
   stat.ambigu  = (int) (edi_status[i][3]) & 128;
   stat.trierr  = (int) (edi_status[i][5]) & 255;                // update 08 Apr 2002
   stat.angerr  = (int) (edi_status[i][6]) & 255;                //        "
   stat.triout  = (int) (edi_status[i][4]) & 255;                //        "
   stat.redchi  = -1.0;

   if (ChiCheck) {
     chisqr = (double) (edi_chisqr[i]);
     if (chisqr >= 0.0 ) stat.redchi  = (float) (edi_chisqr[i]);
   }

   if (VERBOSE)
   {
	printf("Rec : %5d  ",i); 
	fflush(stdout);
   }
   REMOVE = DataOk(stat,settings);

   if (!REMOVE) {
     for (j=0; j<3; j++) edi_temp[n_left][j] = (float) (edi_vector[i][j]);
     tt_temp[n_left] = tt_in[i];
     n_left++;
   } 
 }

 sprintf(txtbuf,"Number of vectors input  : %d \n",n_vecs);
 QplugAppendTextDisplay ( txtbuf);
 sprintf(txtbuf,"Number of vectors output   : %d \n\n",n_left);
 QplugAppendTextDisplay ( txtbuf);

 if (n_left == 0) {	
   QplugAppendTextDisplay ( "ERROR >>  no vectors fit filters !\n");
   return QPLUG_FAILURE; 
 }        


 // ---- transfer the n_left records to output object ----


 QdTimeSeq_var tt_out = new QdTimeSeq(n_left);
 edi_filtered = new QdRMatrixSeq(n_left,RMatrix(Dim(3)));
 
 for (i=0; i<n_left; i++) {
   tt_out[i] = tt_temp[i];
   for (j=0; j<3; j++) edi_filtered[i][j] = (float) (edi_temp[i][j]);
 }


 // ----- generate name of output object (SEH 17 Nov. 2001) ----------

 char str1[127]="";
 char *qsetstr[7]  = {"BAD","CAUTBAD","CAUT","ALL","GOODBAD","GOODCAUT","GOOD" };
 char *methstr[15] = {"SMT","PMTSMT","PMT","TRISMTPMT","TRISMT","TRIPMT","TRI",
                      "SMTRMT","PMTSMTRMT","PMTRMT","ALL","TRISMTRMT",
		      "TRIPMTRMT","TRIRMT","RMT"};
 char *ambistr[3]  = {"ALL","NO","YES" };

 strncat(str1,QuGetAttrText("ObjectName",edi_vector),1);
 strcat(str1,"_edi_C");
 strncat(str1,QuGetAttrText("SC_id",edi_vector),1);
 strcat(str1,"_");
 strncat(str1,(tt_out[0].iso_srep()).c_str(),10); 	
 strcat(str1,"_");
 if (settings.quality < 7) strcat(str1,qsetstr[settings.quality]);
 strcat(str1,"_");
 if (settings.method < 15) strcat(str1,methstr[settings.method]);
 strcat(str1,"_");
 if (settings.ambigu < 3) strcat(str1,ambistr[settings.ambigu]);

 char *slotname = GetSlotText(0);
 if (strcmp(slotname,"") == 0)
   SetSlotText(0,str1);



 // ----- transfer timetags, copy attributes and push back ------

 set_timetags(edi_filtered,tt_out);

 QuSetTxtAttr("Frame","vector>gse_xyz",(QdObject_var) edi_filtered);
 QuSetTxtAttr("SI_conversion",QuGetAttrText("SI_conversion", edi_vector),(QdObject_var) edi_filtered);
 QuSetTxtAttr("UNITS",QuGetAttrText("UNITS", edi_vector),(QdObject_var) edi_filtered);
 QuSetTxtAttr("LABLAXIS",QuGetAttrText("LABLAXIS", edi_vector),(QdObject_var) edi_filtered);
 QuSetTxtAttr("SC_id",QuGetAttrText("SC_id", edi_vector),(QdObject_var) edi_filtered);

 call_list->push_back((QdObject_var) edi_filtered);

 QplugAppendTextDisplay ("EDI_filter>> data object created - exiting successfully\n");

 return QPLUG_SUCCESS;
}




int DataOk(sbytes  stat, sbytes settings)
/* -----------------------------------------------------------------
  
  returns 1 if the EDI input data does not satisfy the
  given filter criteria.

  inputs
      stat - extracted status information from EDI cdf variable
      settings - filter options provided by the user

  SEH, 23 Oct 2001
       01 Nov 2001  (reduced ChiSqr now as float + break on Default
                     in switch statements)
       02 Mar 2002  (new method RMT - see mail from M.Foerster, 
                     2. Mar, 2002, 0923 UT)
--------------------------------------------------------------------*/

{

   int REMOVE=0;

   if (VERBOSE)
   {
	printf(" StatWord : %3d %3d %3d %3d %3d %3d %5.1f  ",stat.quality,stat.method,stat.ambigu,stat.trierr,stat.angerr,stat.triout,stat.redchi);
	fflush(stdout);
   }
   if (stat.quality == 3) {
     REMOVE = 1;
       if (VERBOSE)
	   {
		printf(" FillVal in QUALITY flag ");
		fflush(stdout);
	   }
   }

   if (!REMOVE) {
     switch (settings.quality) {
     case GOOD:
       if (stat.quality != 2) REMOVE = 1;
       if (VERBOSE) printf("Quality not GOOD ");
       break;
     case GOOD_CAUT:
       if (stat.quality == 0) REMOVE = 1;
       if (VERBOSE) printf("Quality not GOOD_CAUT ");
       break;
     case GOOD_BAD:
       if (stat.quality == 1) REMOVE = 1;
       if (VERBOSE) printf("Quality not GOOD_BAD ");
       break;
     case CAUT:
       if (stat.quality != 1) REMOVE = 1;
       if (VERBOSE) printf("Quality not CAUT ");
       break;
     case CAUT_BAD:
       if (stat.quality >= 2) REMOVE = 1;
       if (VERBOSE) printf("Quality not CAUT_BAD ");
       break;
     case BAD:
       if (stat.quality >= 1) REMOVE = 1;
       if (VERBOSE) printf("Quality not BAD ");
       break;
     default:
       REMOVE = 0;
       if (VERBOSE) printf("No Quality filtering \n");
       break;
     }
   }
   if(VERBOSE) fflush(stdout);
   // ----- check Method -----
   //
   // Updated 2. Mar 2002 SEH
   // 
   // stat.method can only attain one of the values
   // 0, 32, 64 or 128. We use the operatos != and == rather 
   // than >= and <= 
   // 

   if (!REMOVE) {
     switch (settings.method) {
     case TRI:
       if (stat.method != 0) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI ");
       break;
     case TRI_PMT:
       if (stat.method != 0 && stat.method !=32) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_PMT ");
       break;
     case TRI_SMT:
       if (stat.method != 0 && stat.method != 64) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_SMT ");
       break;
    case PMT:
       if(stat.method != 32) REMOVE = 1;
       if (VERBOSE) printf("Method not PMT ");
       break;
     case PMT_SMT:
       if (stat.method != 32 && stat.method != 64) REMOVE = 1;
       if (VERBOSE) printf("Method not PMT_SMT ");
       break;
     case SMT:
       if (stat.method != 64) REMOVE = 1;
       if (VERBOSE) printf("Method not SMT ");
       break;
     case TRI_PMT_SMT:       // remove if RMT 
       if (stat.method == 96) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_SMT ");
       break;
     case RMT:              // remove if not RMT 
       if (stat.method != 96) REMOVE = 1;
       if (VERBOSE) printf("Method not RMT ");
       break;
     case TRI_RMT:          // i.e remove if SMT or PMT        
       if (stat.method != 0 && stat.method != 96) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_RMT ");
       break;
     case TRI_PMT_RMT:      // i.e remove if SMT 
       if (stat.method == 64) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_PMT_RMT ");
       break;
     case TRI_SMT_RMT:       // i.e remove if PMT      
       if (stat.method == 32) REMOVE = 1;
       if (VERBOSE) printf("Method not TRI_PMT_RMT ");
       break;
     case PMT_RMT:           // remove if TRI or SMT   
       if (stat.method == 0 || stat.method == 64) REMOVE = 1;
       if (VERBOSE) printf("Method not PMT_RMT ");
       break;
     case PMT_SMT_RMT:       // allow all but TRI 
       if (stat.method == 0) REMOVE = 1;
       if (VERBOSE) printf("Method not PMT_SMT_RMT ");
       break;
     case SMT_RMT:           // i.e. remove if TRI or PMT 
       if (stat.method == 0  || stat.method == 32) REMOVE = 1;
       if (VERBOSE) printf("Method not SMT_RMT ");
       break;

     default:
       REMOVE = 0;
       if (VERBOSE) printf("No Method filtering \n");
       break;
     }
	     if(VERBOSE) fflush(stdout);
   }

   // ----- check ambiguity ------

   if (!REMOVE) {
     switch(settings.ambigu) {
     case YES:
       if (stat.ambigu != 128) REMOVE = 1;
       if (VERBOSE) printf("Ambig not YES ");
       break;
     case NO:
       if (stat.ambigu !=0) REMOVE = 1;
       if (VERBOSE) printf("Ambig not NO ");
       break;
     default:
       REMOVE = 0;
       if (VERBOSE) printf("No Ambiguity filtering \n");
       break;
     }
	     if(VERBOSE) fflush(stdout);
   }
			       

   // ----- check angle error ------

   if(!REMOVE) {
     if(stat.angerr > settings.angerr && stat.angerr != 255) {
       REMOVE = 1;
       if (VERBOSE) printf(" to high ANGERR ");
     }
   }


   // ----- check magnitude error ------

   if(!REMOVE) {
     if ((stat.trierr > settings.trierr) && (stat.trierr != 255)) {
       REMOVE = 1;
       if (VERBOSE) printf(" to high MAGERR ");
     }
   }


   // ----- check triangulation outliers (triangulation only) -----

   if (!REMOVE && (stat.method == 0) && (stat.triout != 255)) {
     if (stat.triout > settings.triout) {
       REMOVE = 1;
       if (VERBOSE) printf(" to many TRI_OUTs ");
     }
   }


   // ----- check reduced chi-sq (triangulation only) -----

   if (!REMOVE && (stat.method == 0)) {
     if (stat.redchi > settings.redchi) {
       REMOVE = 1;
       if (VERBOSE) printf(" Rchi above limit ");
     }
   }
  
   if (REMOVE && stat.quality !=3 ) {
     if (VERBOSE) printf(" - record  deleted \n");
   }
   else {
     if (VERBOSE) printf("\n");
   }
    if(VERBOSE) fflush(stdout);
   return (REMOVE);
}    


sbytes readfilterfile(const char *filename)
/* ------------------------------------------------------------------------

  reads filter settings from file

  SEH 23 Oct 2001

 ------------------------------------------------------------------------*/
{

  int lineno=0, comment=1;
  FILE* f1;
  char buf[500];
  char *svalue,*token;
  sbytes stat;

  // ---- default values ----

  stat.quality = GOOD_CAUT_BAD;
  stat.method  = TRI_PMT_SMT;
  stat.ambigu  = YES_NO ;
  stat.angerr  = 100;
  stat.trierr  = 100;
  stat.triout  = 100;
  stat.redchi  = 100.0;
  stat.OK      = 0;                                           // No filterfile yet -> OK =0 

  f1 = fopen(filename,"r");
  //  printf("Filter settings file : %s\n",filename);

  token = " ";                                                // possible separators in input file 

  if (f1 != NULL){
    while(fgets(buf,500,f1) != NULL) {
      comment = (buf[0] == '#' || buf[0] == '%' || buf[0] == ';' || buf[0] == '!');
      if (!comment) {
        for (svalue = strtok(buf,token); svalue != 0; svalue=strtok(0,token)) {
	  switch (lineno) {
	  case 0:
	    stat.quality = atoi(svalue);
	    break;
	  case 1:
	    stat.method = atoi(svalue);
	    break;
	  case 2:
	    stat.ambigu = atoi(svalue);
	    break;
	  case 3:
	    stat.angerr = atoi(svalue);
	    break;
	  case 4:
	    stat.triout = atoi(svalue);
	    break;
	  case 5:
	    stat.trierr = atoi(svalue);
	    break;
	  case 6:
	    stat.redchi = atof(svalue);
	    break;
	  }
	}
	lineno++;
      }
    }
    stat.OK = 1;                                             // Filter file found -> OK = 1
    fclose(f1);
  } 
  else {
    printf(" NO FILTER FILE - using defaults \n");
	fflush(stdout);
  }

  return(stat);

}



int main()
{
  // do nothing
}
