/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

/* PROJECT:     Cluster UK CDHF
* COMPONENT:   IRM Onboard Timetags Plugin 
* MODULE:      IRMTags
* LANGUAGE:	   C++
* FUNCTION:
* PURPOSE:     Shift IRM Timetags to Onboard UT 
* DESCRIPTION:
* ARGUMENTS: [pased via call_list]
*    IN:
*     ANY_TIME_SERIES      ts_in        input time series data object
*     VECTOR_TIME_SERIES   IRMPos       IRM Position in GEOC joined to ts_in
*
*    OUT:
*     ANY_TIME_SERIES      ts_out       same as ts_in but with timetags
*                                              shifted to UT Onboard         
* GLOBALS:
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
* FUNCTIONS_USED: 
* ALGORITHM:   pretty obvious
* HISTORY:	   Steve Schwartz  Nov 98
*              AJA March 2000
***********************************************************************/

#ifndef __cplusplus
#define __cplusplus
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "IRMTags.h"

using namespace QSAS;

#ifdef __cplusplus
  extern "C" {
#endif
#ifdef __cplusplus
  }
#endif

QplugReturnStatus  IRMTags  ( QplugArgList *call_list )
{

int     at_ok;
char    outstring[200];
int ndata;

// QDOS Objects:

QdObject_var    ts_in;
QdRMatrixSeq_var    irmpos;

QplugReturnStatus      IRMTags_return;   // used to return from sub-modules

QplugAppendTextDisplay("IRM Timetag shift to onboard UT. Nov 1998\n");

int nlist = call_list->size();        // are there enough args? 
if (nlist < 2){
   QplugAppendTextDisplay( "Not enough entries in argument list from GUI\n");
   return QPLUG_FAILURE;
}  //endif


//	...next, get the input data objects and fill arrays

 ts_in = (*call_list)[0];    // FIRST INPUT ARGUMENT = IRM Data Object

#ifdef VERBOSE
  QplugAppendTextDisplay("Got first argument\n");
#endif

      
irmpos = QdRMatrixSeq_var::narrow( (*call_list)[1] );      // SECOND INPUT ARGUMENT = IRM Position

if( irmpos.is_nil() || !QuIsThreeVector((QdObject_var)irmpos) ){
       QplugAppendTextDisplay("ERROR>>Second Argument not a vector TS\n");	
       return QPLUG_FAILURE;
} //endif
  
#ifdef VERBOSE
 QplugAppendTextDisplay("Got second argument\n");
#endif

 char * attribute_value = QuGetAttrText("Frame", (QdObject_var)irmpos);
 if (TestGeocFrame(attribute_value) != QPLUG_SUCCESS){ 
    sprintf(outstring,"IRMTags>> IRM Pos Frame given as %s, not geoc_xyz.\n Proceed with caution\n", attribute_value);           
    QplugAppendTextDisplay(outstring);
 }
 
 
#ifdef VERBOSE
      sprintf(outstring,"Frame = %s\n",attribute_value);
      QplugAppendTextDisplay(outstring);
#endif
 
//      ... test that DO and Pos objects are joined

QdTimeSeq_var ts_tt  = get_timetags(ts_in);
QdTimeSeq_var pos_tt = get_timetags( (QdObject_var)irmpos );
 
if(ts_tt.is_nil() || pos_tt.is_nil() ) {
  QplugAppendTextDisplay("ERROR>>IRM Data or position not a time series\n");	
  return QPLUG_FAILURE;
}
 
 ndata = ts_in->sequence_size();

// test joined

if ( !( (*ts_tt) ==  (*pos_tt) ) ) {
      QplugAppendTextDisplay (
          "IRMTags >> Data Object and Position not Joined\n");
      return QPLUG_FAILURE;  
}
#ifdef VERBOSE
  QplugAppendTextDisplay("IRMTags; DO and Pos joined checks\n");
#endif
 
 // Create and fill output data object; add Mod re shifted tags
// Note that copy operation automatically passes over all Xrefs (attributes)

 QdTimeSeq_var tt_out = (QdTimeSeq_var) Factory::clone(ts_tt);
 QdObject_var  ts_out = (QdObject_var) Factory::clone(ts_in);

IRMTags_return = IRMTags_generate (irmpos, tt_out, ndata);
if (IRMTags_return != QPLUG_SUCCESS) {
      QplugAppendTextDisplay (
          "IRMTags >> New tags failed to be generated!\n");
      return QPLUG_FAILURE;  
}

#ifdef VERBOSE
  QplugAppendTextDisplay("IRMTags New timetags generated.\n");
#endif


 set_timetags(ts_out, tt_out);


 // Add attribute describing what we've done
 
 QuSetTxtAttr("Analysis_Notes", "IRMTags used to shift Timetags to On-board UT", ts_out);

 call_list->push_back(ts_out);     
 

#ifdef VERBOSE
 QplugAppendTextDisplay(">>IRMTags  object done successfully\n");
#endif


return QPLUG_SUCCESS;
                                                 // cleanup and finish 
}  // end IRMTags



/***********************************************************************
***************************************************  IRMTags_generate
***********************************************************************/

QplugReturnStatus  IRMTags_generate (QdRMatrixSeq_var pos, QdTimeSeq_var outtag, int ndata)
{

int     i;
double   dr,temp;

for(i=0; i<ndata; i++){
  dr=0.0;
  temp= pos[i][0] - gndstn_x; temp =temp*temp;dr += temp;
  temp= pos[i][1] - gndstn_y; temp =temp*temp;dr += temp;
  temp= pos[i][2] - gndstn_z; temp =temp*temp;dr += temp;
  dr=sqrt(dr);
  outtag[i] -= Seconds(dr/c_re_s);
}   // end for
return QPLUG_SUCCESS;
}

/**********************************************************************
****************************************************** TestGeocFrame  
**********************************************************************/

/* Borrowed fragments from Tony's QRTN QrGetFrameFromAttribute routine */

QplugReturnStatus TestGeocFrame(char * attstring){
 
char *str;          /* need to copy string as strtok changes its arg */
char *ptr;
 
str = new char[strlen(attstring)+1];
strcpy(str,attstring);

  /* Frame is of format vector>frame_xyz" */
   
 ptr = strtok(str,">");
 if(ptr != NULL) ptr = strtok(NULL, "_");
 
 if ( (ptr != NULL)  && (strcmp(ptr, "geoc") == 0) ){
    delete[] str;
    return QPLUG_SUCCESS;
 } else {
    delete[] str;
    return QPLUG_FAILURE;
 }
}
