/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	LombSpectrum.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS LombSpectrum calculation
*
* PURPOSE:      PlugIn for calculation of the Lomb Spectrogram
*               using same method as  NR 'fasper' routine.
* 
* DESCRIPTION:  Plug In module checking input parameters.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*              QdCTimeSeries*   vector or scalar regular time series object
*              QdCTimeInterval intervall object for subsetting
*              hifac    scalar high frequency factor
*              ofac     scalar oversampling factor
*
*   OUT:
*         double *frequency       frequency tags		
*         double *power_mag       power spectral density of magnitude
*         vector series *power_vec       power spectral density of components
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED: Qfasper  avevar  spread rdft
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*             in contrast to NR we use C-indices and Oouras FFT code.
*
* HISTORY: created 1997 Markus Fraenz 
*          migrated to qdos 2 - AJ Allen Oct 2000
*          vector spectrum and subsetting added Markus Fraenz 18/01/2001
*          changed to C-indices and Oouras rdft routine. Markus Fraenz 11/07/2001
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

void Qfasper(double x[], double y[], unsigned long n, double ofac, double hifac,
	double wk1[], double wk2[], unsigned long nwk, unsigned long *nout,
	unsigned long *jmax, double *prob);
void avevar(double data[], unsigned long n, double *ave, double *var);
extern "C" void rdft(int n, int isgn, double *a, int *ip, double *w);
void spread(double y, double yy[], unsigned long n, double x, int m);

#define MOD(a,b) 	while(a >= b) a -= b;
#define MACC 4

static long lmaxarg1,lmaxarg2;
#define LMAX(a,b) (lmaxarg1=(a),lmaxarg2=(b),(lmaxarg1) > (lmaxarg2) ?\
        (lmaxarg1) : (lmaxarg2))

static long lminarg1,lminarg2;
#define LMIN(a,b) (lminarg1=(a),lminarg2=(b),(lminarg1) < (lminarg2) ?\
        (lminarg1) : (lminarg2))

static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

extern "C" QplugReturnStatus  LombSpectrum(QplugArgList *);

using namespace QSAS;    

QplugReturnStatus  LombSpectrum(  QplugArgList *call_list)
{


// define some C variables

long	i, j, vectorset=0;

unsigned long	Ndata, nwk, nfreq, nfreqt, nout,jmax;
double           hifac,ofac,prob;

double	hs;
char	str_out[150];

double	fmin,fmax,pmin,pmax;
double  *data,*wk1,*wk2,*ftags;
double *tags;


// define Qdos data objects

QdTimeInterval_var start_end;
QdObject_var ts_out;
QdTimeSeq_var tt_out;

QdRScalar_var      option_1;
QdRScalar_var      option_2;
QdRScalarSeq_var   ts_inr;
QdRScalarSeq_var   out_freq_series;
QdRScalarSeq_var   out_spmag_series;
QdObject_var  out_spvec_series;
QdRMatrixSeq_var ts_inm;



// ...first item is input Time Series

QdObject_var ts_in = (* call_list)[0];

//  check QSAS datatype

QdTimeSeq_var tt_in = get_timetags(ts_in);
if( tt_in.is_nil() ) {
  QplugAppendTextDisplay ("LombSpectrumFFT>> Failed to get time tags\n");
  return QPLUG_FAILURE;
}

//	..2nd item is option 'time_interval'

 QdObject_var dobj1 = (* call_list)[1];
 start_end = get_timeinterval(dobj1);
 

//	 option 'hifac'

QdObject_var dobj2 = (* call_list)[2];

option_1 = QdRScalar_var::narrow(dobj2);

if ( option_1.is_nil() ) { 
	QplugAppendTextDisplay ("LombSpectrum>> ERROR - hifac must be double.\n");
	return QPLUG_FAILURE; }

hifac = (double)(*option_1);

if (hifac <= 0.0) {
	QplugAppendTextDisplay ("LombSpectrum>> ERROR - hifac must be greater than 0.0\n");
	return QPLUG_FAILURE; }



//	 option 'ofac'

QdObject_var dobj3 = (* call_list)[3];

option_2 = QdRScalar_var::narrow(dobj3);

if ( option_2.is_nil()  ) {
	QplugAppendTextDisplay ("LombSpectrum>> ERROR - ofac must be double.\n");
	return QPLUG_FAILURE; }
	
ofac = (double)(*option_2);
if (ofac < 1.0) {
	QplugAppendTextDisplay ("LombSpectrum>> ERROR - ofac must be > 1.0. \n");
	return QPLUG_FAILURE; }

// now take the subset for the time interval chosen  ###SUBSETTING START

QplugAppendTextDisplay ("LombSpectrum>> Subsetting Time Series...\n");
SequenceMask msk(tt_in->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 
tt_out = new QdTimeSeq();
  
tt_in->get_subsequence(*tt_out, msk);

  
QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(ts_in);
if(!ss.is_nil() ){
    QdRScalarSeq_var  ss_out = new QdRScalarSeq();
    ss->get_subsequence(*ss_out, msk);
    ts_out = (QdObject_var) ss_out;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil() ){
      QplugAppendTextDisplay ("LombSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
    if(QuIsThreeVector(ms) ){
      QdRMatrixSeq_var  ms_out = new QdRMatrixSeq();
      ms->get_subsequence(*ms_out, msk);
      ts_out = (QdObject_var) ms_out;
    }
    else {
      QplugAppendTextDisplay ("LombSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }

}
 

if ( ts_out.is_nil() ){
    QplugAppendTextDisplay ("LombSpectrumFFT>> ERROR - subset failed:\n");
    return QPLUG_FAILURE;
} 
if ( ts_out->sequence_size() <1 ){
    QplugAppendTextDisplay ("LombSpectrumFFT>> Empty subset:\n");
    return QPLUG_FAILURE;
}

ts_in->copy_xrefs_to(ts_out);
set_timetags(ts_out, tt_out);


//	..extract the time tags from the data series object ###SUBSETTING END


Ndata = ts_out->sequence_size();

if (Ndata <= 1) {
	QplugAppendTextDisplay ("LombSpectrum>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

//	..extract the data from the data series object

if ( QuIsThreeVector(ts_out) ) {
  vectorset=1;	
  QplugAppendTextDisplay ("LombSpectrum>> Decomposing Vector Timeseries...\n");
  ts_inm = QdRMatrixSeq_var::narrow(ts_out);
  if ( ts_inm.is_nil() ) { return QPLUG_FAILURE; }
  ts_inr = QdRScalarSeq_var::narrow(QuGetVectorComponent(ts_out, "r(mag)")); 
}
else
 ts_inr = QdRScalarSeq_var::narrow(ts_out);


data  = new double[Ndata];
ftags = new double[Ndata];
tags  = new double[Ndata];


for(int i=0; i<Ndata; i++){
  data[i] = (double) (ts_inr[i]);
  tags[i] = cdfEpoch( tt_out[i] );
  
  // substract offset from tags and convert to double seconds
  ftags[i]=(double)(tags[i]-tags[0])/1000.0;
}


// calculate the output array sizes according to the error check in 'fasper'
// (this is to avoid 'nrerror' calls from 'fasper'.)

nfreqt= (unsigned long) (ofac*hifac*((double)(4*Ndata)));
nfreq=64;
while (nfreq <  nfreqt){
 nfreq <<= 1;
}
nwk=nfreq << 1;
wk1 = new double[nwk];
wk2 = new double[nwk];



QplugAppendTextDisplay ("LombSpectrum>> Calculating Spectrum.\n");

//	call the NR fasper power spectrum calculation
//      (notice that NR routines start indices with 1!!) 

Qfasper(ftags,data,Ndata,ofac,hifac,wk1,wk2,nwk,&nout,&jmax,&prob);


//	..some messages to the analysis interface window..

sprintf(str_out, "LombSpectrum>> %i data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"LombSpectrum>> %i frequencies\n", nout);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("LombSpectrum>> Values for Magnitude:\n");
sprintf(str_out,"LombSpectrum>> maximal power:%f, at freqency: %f, and significance: %e \n", 
            wk2[jmax],wk1[jmax],prob);
QplugAppendTextDisplay (str_out);


//	..create the output data series objects


out_freq_series  = new QdRScalarSeq( nout);
out_spmag_series = new QdRScalarSeq( nout);

for (int i=0; i<nout; i++){
  (out_freq_series[i]) = (double) wk1[i];
  (out_spmag_series[i]) = (double) wk2[i];
}


if(vectorset) { 
 QdRMatrixSeq_var spvec_series = new QdRMatrixSeq(nout,RMatrix(Dim(3)));

 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][0]);
 Qfasper(ftags,data,Ndata,ofac,hifac,wk1,wk2,nwk,&nout,&jmax,&prob);
 for (long i =0; i<nout; i++) spvec_series[i][0] = (double) wk2[i];
 QplugAppendTextDisplay ("LombSpectrum>> Values for Component 1:\n");
 sprintf(str_out,"LombSpectrum>> maximal power:%f, at freqency: %f, and significance: %e \n", 
            wk2[jmax],wk1[jmax],prob);
 QplugAppendTextDisplay (str_out);

 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][1]);
 Qfasper(ftags,data,Ndata,ofac,hifac,wk1,wk2,nwk,&nout,&jmax,&prob);
 for (long i =0; i<nout; i++) spvec_series[i][1] = (double) wk2[i];
 QplugAppendTextDisplay ("LombSpectrum>> Values for Component 2:\n");
 sprintf(str_out,"LombSpectrum>> maximal power:%f, at freqency: %f, and significance: %e \n", 
            wk2[jmax],wk1[jmax],prob);
 QplugAppendTextDisplay (str_out);

 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][2]);
 Qfasper(ftags,data,Ndata,ofac,hifac,wk1,wk2,nwk,&nout,&jmax,&prob);
 for (long i =0; i<nout; i++) spvec_series[i][2] = (double) wk2[i];
 QplugAppendTextDisplay ("LombSpectrum>> Values for Component 3:\n");
 sprintf(str_out,"LombSpectrum>> maximal power:%f, at freqency: %f, and significance: %e \n", 
            wk2[jmax],wk1[jmax],prob);
 QplugAppendTextDisplay (str_out);
 out_spvec_series = (QdObject_var) spvec_series;
}


//	...set some attributes

  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spmag_series);
  QuSetTxtAttr("LABLAXIS", "Lomb  Periodogram", (QdObject_var) out_spmag_series);
  QuSetTxtAttr("FIELDNAM", "Probability spectrum", (QdObject_var) out_spmag_series);
  QuSetFrame( (QdObject_var) out_spmag_series, "scalar");
  QuSetTxtAttr("UNITS", "normalized", (QdObject_var) out_spmag_series);
  out_spmag_series->set_xref("DEPEND_0", out_freq_series);
  
  if(vectorset) {
    QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("LABLAXIS", "Lomb Periodogram", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("FIELDNAM", "Probability spectrum", (QdObject_var) out_spvec_series);
//     QuSetFrame((QdObject_var) out_spvec_series, "vector", "na", "na");
    QuSetFrame( (QdObject_var) out_spvec_series, QuGetOrder(ts_in), QuGetFrame(ts_in), QuGetRep(ts_in)); 
    QuSetTxtAttr("UNITS", "normalized", (QdObject_var) out_spvec_series);
    out_spvec_series->set_xref("DEPEND_0", out_freq_series);
  }

  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_freq_series);
  QuSetTxtAttr("LABLAXIS", "frequency", (QdObject_var) out_freq_series);
  QuSetTxtAttr("FIELDNAM", "frequency", (QdObject_var) out_freq_series);
  QuSetFrame((QdObject_var) out_freq_series, "scalar");
  QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_freq_series);


  QplugAppendTextDisplay ("LombSpectrum>> finished...\n");


//	..add them to the list

  call_list->push_back( (QdObject_var) out_spmag_series);     
  call_list->push_back( (QdObject_var) out_spvec_series);     



//	..clean up and exit

delete[] data;
delete[] tags;
delete[] ftags;
delete[] wk1;
delete[] wk2;


return QPLUG_SUCCESS;


}


void Qfasper(double x[], double y[], unsigned long n, double ofac, double hifac,
	double wk1[], double wk2[], unsigned long nwk, unsigned long *nout,
	unsigned long *jmax, double *prob)
{
	unsigned long j,k,ndim,nfreq,nfreqt;
	double ave,ck,ckk,cterm,cwt,den,df,effm,expy,fac,fndim,hc2wt;
	double hs2wt,hypo,pmax,sterm,swt,var,xdif,xmax,xmin,*w_tak;
	int *ip_tak;

	*nout=(unsigned long) (0.5*ofac*hifac*n);
	nfreqt=(unsigned long) ((ofac*hifac)*n*MACC);
	nfreq=64;
	while (nfreq < nfreqt) nfreq <<= 1;
	ndim=nfreq << 1;
	if (ndim > nwk) {
	  printf("workspaces too small in fasper");
	  fflush(stdout);
	  return;
	}
	ip_tak = new int[(long) sqrt((double) ndim/2) +3];
        ip_tak[0]=0;
        w_tak = new double[ndim/2];


	avevar(y,n,&ave,&var);
	xmin=x[0];
	xmax=xmin;
	for (j=1;j<n;j++) {
		if (x[j] < xmin) xmin=x[j];
		if (x[j] > xmax) xmax=x[j];
	}
	xdif=xmax-xmin;
	for (j=0;j<ndim;j++) wk1[j]=wk2[j]=0.0;
	fac=ndim/(xdif*ofac);
	fndim=ndim;
	for (j=0;j<n;j++) {
		ck=(x[j]-xmin)*fac;
		MOD(ck,fndim)
		ckk=2.0*(ck++);
		MOD(ckk,fndim)
		++ckk;
		spread(y[j]-ave,wk1,ndim,ck,MACC);
		spread(1.0,wk2,ndim,ckk,MACC);
	}
	//realft(wk1-1,ndim,1);
	//realft(wk2-1,ndim,1);
	rdft((int)ndim,1,wk1,ip_tak,w_tak);
	rdft((int)ndim,1,wk2,ip_tak,w_tak);
	df=1.0/(xdif*ofac);
	pmax = -1.0;
	for (k=2,j=0;j<(*nout);j++,k+=2) {
		hypo=sqrt(wk2[k]*wk2[k]+wk2[k+1]*wk2[k+1]);
		hc2wt=0.5*wk2[k]/hypo;
		hs2wt=0.5*wk2[k+1]/hypo;
		cwt=sqrt(0.5+hc2wt);
		swt=SIGN(sqrt(0.5-hc2wt),hs2wt);
		den=0.5*n+hc2wt*wk2[k]+hs2wt*wk2[k+1];
		cterm=SQR(cwt*wk1[k]+swt*wk1[k+1])/den;
		sterm=SQR(cwt*wk1[k+1]-swt*wk1[k])/(n-den);
		wk1[j]=(j+1)*df;
		wk2[j]=(cterm+sterm)/(2.0*var);
		if (wk2[j] >= pmax) pmax=wk2[(*jmax=j)];
	}
	expy=exp(-pmax);
	effm=2.0*(*nout)/ofac;
	*prob=effm*expy;
	if (*prob > 0.01) *prob=1.0-pow(1.0-expy,effm);
        delete ip_tak;
        delete w_tak;

}

void avevar(double data[], unsigned long n, double *ave, double *var)
{
	unsigned long j;
	double s,ep;

	for (*ave=0.0,j=0;j<n;j++) *ave += data[j];
	*ave /= n;
	*var=ep=0.0;
	for (j=0;j<n;j++) {
		s=data[j]-(*ave);
		ep += s;
		*var += s*s;
	}
	*var=(*var-ep*ep/n)/(n-1);
}


	

void spread(double y, double yy[], unsigned long n, double x, int m)
{
	int ihi,ilo,ix,j,nden;
	static int nfac[11]={0,1,1,2,6,24,120,720,5040,40320,362880};
	double fac;

	if (m > 10) {
	  printf("factorial table too small in spread");
	  fflush(stdout);
	  return;
	}
	ix=(int)x;
	if (x == (double)ix) yy[ix-1] += y;
	else {
		ilo=LMIN(LMAX((long)(x-0.5*m+1.0),1),n-m+1);
		ihi=ilo+m-1;
		nden=nfac[m];
		fac=x-ilo;
		for (j=ilo+1;j<=ihi;j++) fac *= (x-j);
		yy[ihi-1] += y*fac/(nden*(x-ihi));
		for (j=ihi-1;j>=ilo;j--) {
			nden=(nden/(j+1-ilo))*(j-ihi);
			yy[j-1] += y*fac/(nden*(x-j));
		}
	}
}

