// ------------------------------------------------------------------
//
//  ConstrainedMinVarMFR.h
//  Qdos2 - Constrained Minimum Variance Algorithm for MFR. Based on 
//  the original MinVariance.h.
//
//  Janet Barnes
//  May 2000
//  Cloned & modified by SEH, May 2002
//  MFR version, SEH 28 May 2002
//
// ------------------------------------------------------------------

#ifndef QDOS_CONSTRAINEDMINVARMFR_H
#define QDOS_CONSTRAINEDMINVARMFR_H

#define VERBOSE              0

#include "Algorithm.h"
#include "Matrix.cc"

namespace QDOS{

class ConstrainedMinVarMFR
{
    Matrix<double> _eigvals;       // Matrix Dim(3) of eigenvalues
    Matrix<double> _eigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    Matrix<double> _eigqual;       // Matrix Dim(3) eigenvalue quality
    double _orgqual;               // orthogonality quality
    Matrix<double> _covariance;    // covariance matrix
    Matrix<double> P;              // projection matrix P (see ISSI book, p 195)
    Matrix<double> Q;              // MFR Q matrix for which eigenvals _ eigenvecs are found
    Matrix<double> temp;           // temporary storage matrix
    int i,j;
    
public:
    ConstrainedMinVarMFR(const MatrixSeq<double>& ms, int constrained, Matrix<double> Q):_eigvals(Matrix<double>(Dim(3))),\
        _eigvecs(Matrix<double>(Dim(3,3))),_covariance(Matrix<double>(Dim(3,3))),\
      _eigqual(Matrix<double>(Dim(3)))
    {
        if(ms.empty())
            return;
        if (!ms.is_homogeneous_seq())
            throw MatrixDomainError();
        if(ms.elt(0).rank()!=1)
            throw MatrixRankError();
        if(ms.elt(0).dimdata()!=Dim(3))
             throw MatrixLengthError();
             
        int sz=ms.size();
        
        // compute the variance matrix 
       
        ScalarSeq<double> x=ms.get_element_seq(vector<int>((size_t)1,0));
        ScalarSeq<double> y=ms.get_element_seq(vector<int>((size_t)1,1));
        ScalarSeq<double> z=ms.get_element_seq(vector<int>((size_t)1,2));
   
        double sum_x=x.sum();
        double sum_y=y.sum();
        double sum_z=z.sum();
 
        _covariance.resize(Dim(3,3));

        _covariance = Q;        
   

        _covariance[1][0] = _covariance[0][1];
        _covariance[2][0] = _covariance[0][2];
        _covariance[2][1] = _covariance[1][2];
        
	// 
	// Constrained MVA : find eigenvec of [P # C # P] * n = lamda * n
	// instead of [C] * n = lambda * n. 
	//
	//   0 - standard MVA
	//   1 - constrain so that Bn == 0 (see p 209 in blue ISSI book)
	//
	// SEH, 14 May 2002
	//      28 May 2002, MFR version
	//

	if (constrained == 1) {

          if (VERBOSE) printf("MonVar_MFR > Using constrain\n");

          temp.resize(Dim(3,3));
	  double sz2 = (double)(sum_x*sum_x + sum_y*sum_y + sum_z*sum_z);

	  P.resize(Dim(3,3));

	  P[0][0] = 1.0 - (sum_x*sum_x) / sz2;
	  P[0][1] = - (sum_x*sum_y) / sz2;
	  P[0][2] = - (sum_x*sum_z) / sz2;

	  P[1][0] = - (sum_y*sum_x) / sz2;
	  P[1][1] = 1.0 - (sum_y*sum_y) / sz2;
	  P[1][2] = - (sum_y*sum_z) / sz2;

	  P[2][0] = - (sum_z*sum_x) / sz2;
	  P[2][1] = - (sum_z*sum_y) / sz2;
	  P[2][2] = 1.0 - (sum_z*sum_z) / sz2;

	  temp = _covariance*P;
	  _covariance = P*temp;

  	// force symmetric covariance matrix again
	  _covariance[1][0] = _covariance[0][1];
	  _covariance[2][0] = _covariance[0][2];
	  _covariance[2][1] = _covariance[1][2];
	}

        // get eigenvectors and values  
        MatrixEig eig(_covariance);
        _eigvecs=eig.eigenvectors().seq_to_matrix();
        _eigvals=eig.eigenvalues().seq_to_matrix();


	// If constrained : eigval 0 == 0. Swap 1st and 2nd eigenvec

        if (constrained == 1) {
          _eigvals[0] = _eigvals[1];
          _eigvals[1] = _eigvals[2];
          _eigvals[2] = 0;                                // by definition
          temp = _eigvecs;
          for (i=0; i<3; i++) _eigvecs[0][i] = temp[1][i];
          for (i=0; i<3; i++) _eigvecs[1][i] = temp[2][i];
          for (i=0; i<3; i++) _eigvecs[2][i] = temp[0][i];
        }

        
        // check the positive sense of the eigenvector associated with
        // the minimum eigenvalue and if necessary reverse the signs
        // of all three components i.e. vint x vmax = vector perpendicular
        // to both vint and vmax. If vmin is perpendicular to vint 
        // and vmax (i.e. system is ortogonal), then (vinx x vmax).vmin=1
        // If < 0 then not a right-handed system, so reverse signs of vmin
        double rh_factor=(_eigvecs[1].cross(_eigvecs[2])).dot(_eigvecs[0]);
        if(rh_factor<0.)
            _eigvecs[0]*=-1;
        _orgqual=std::abs(std::abs(rh_factor)-1.0);
                
        // ensure that the sign of the average minimum variance component
        // is positive by computing average of sum of B-field data.min
        // eigenvector. If less than zero negate both vmin and vint (so
        // preserving right-handed system)
        double avg_minvar_comp=(ms*_eigvecs[0]).sum();
        if(avg_minvar_comp<0) 
        {
            _eigvecs[0]*=-1;
            _eigvecs[1]*=-1;
        }
        _eigqual=_eigvals/_eigvals[2];
            
    }



    // methods for returning members and calculating variance
    Matrix<double>& covariance() 
    { return _covariance; }
    
    Matrix<double>& eigenvalues() 
    { return _eigvals;}
     
    Matrix<double>& eigenvectors()
    { return _eigvecs;}   
    
    Matrix<double>& eigenquality() 
    { return _eigqual;}
    
    double& orthogonality_quality()
    { return _orgqual;}    // orthogonality quality     

        
};
    
}  // end namespace QDOS
#endif
