//  qgu.cc
//
// GUI Utility Widgets
//
// Jan 2002: Tony Allen
//

/// \file qgu.cc
///  Implementation file for the main GUI classes that are used within QSAS which are: \n
///  \arg \c QguListItem \n
///  \arg \c QguListView \n
///  \arg \c QguWLinput  \n
///  \arg \c QguWindow   \n \n
///  Those classes specialise respectively the QT3 built-in classes as listed below\n
///  \arg \c Q3ListViewItem \n
///  \arg \c Q3ListView     \n
///  \arg \c QLineEdit     \n
///  \arg \c QMainWindow   \n 
///  \author Tony Allen    \n
///  \version 2.0          \n
///  \date    2004

#include "qgu_qt4.h"
#include "qwui.h"

#include <stdio.h>

 qwuiWindow* qsasWindow;

/// \var  extern vector<QguListItem *> dragList
/// \brief Global variable that contains the list of items being dragged
///
 vector<QguListItem *> dragList;

/// \var  extern QguListView * dragFromList
/// \brief Global variable that defines the list view from which the items are being dragged
///
 QguListView * dragFromList;
//    QguListItem CLASS
//    -------------------


/// \fn QguListItem::QguListItem(QguListView *parent,  QString txt, QguListType type, QguListItem *after)
/// \brief Constructor of the QguListItem class
/// \param     parent of type QguListView * 
/// \param     txt    of type QString  
/// \param     type   of type QguListType  
/// \param     after  of type QguListItem *
/// \exception None
/// \return    None
///  

QguListItem::QguListItem(QguListView *parent,  QString txt, 
                         QguListType type, QguListItem *after)
 :  Q3ListViewItem(parent, after)
{

  entryType = type;
  Lock = false;

  if(type == ListTypeGroup) {
    setExpandable(TRUE);  
    this->setOpen(TRUE);
  }
  
  setText(Obj_Name_, txt);
   
  ((QguListView *)this->listView())->clearSelection();
  
  // ensure name is unique
    
  // don't take "cancel" as an answer
  while( nameAlreadyUsed(this->text(Obj_Name_)) ){ 
    QString newText = this->text(Obj_Name_);
    newText += "A";
    this->setText(Obj_Name_, newText);
  }

  WLcontext = nameContext();

}


QguListItem::QguListItem(QguListItem *parent,  QString txt,  
QguListType type, QguListItem *after)
 :  Q3ListViewItem(parent, after)
{

  entryType = type;
  Lock = false;

  if(type == ListTypeGroup) {
    setExpandable(TRUE);
    this->setOpen(TRUE);
  }

  setText(Obj_Name_, txt);
  
  ((QguListView *)this->listView())->clearSelection();
    
  // ensure name is unique
    
  // don't take "cancel" as an answer
  while( nameAlreadyUsed(this->text(Obj_Name_)) ){ 
    QString newText = this->text(Obj_Name_);
    newText += "A";
    this->setText(Obj_Name_, newText);
  }

  WLcontext = nameContext();
}

/// \fn void QguListItem::~QguListItem()
/// \brief Destructor of QguListItem class
/// \param     None
/// \exception None
/// \return    None
///  
// QguListItem::~QguListItem()
// {  
// }

/// \fn void QguListItem::setLocked(bool lock)
/// \brief Method that sets the lock private flag for the current list item
/// \param     lock  of bool type
/// \exception None
/// \return    None
///  
void QguListItem::setLocked(bool lock)
{ 
   Lock = lock;
} 

/// \fn bool QguListItem::isLocked()
/// \brief Method that returns the value of the lock private flag for the current list item
/// \param     None
/// \exception None
/// \return    The value of the flag Lock
///  
bool QguListItem::isLocked()
{ 
   return Lock;
}
 
/// \fn bool QguListItem::nameUsed(QString &name)
/// \brief Method thats checks if the provided name is included within this folder. Alays false if this is an entry.
/// \param     name of QString & type
/// \exception None
/// \return    Returns TRUE if the provided name has been used
///   
bool QguListItem::nameUsed(QString &name){
  // tests if a folder (this) already has an item of given name
  QguListItem * sibling;
  sibling = (QguListItem *)this->firstChild();
  
  while(sibling){
   if(sibling->text(Obj_Name_) == name) return true;
   QguListItem *next = (QguListItem *)sibling->nextSibling();
   sibling = next;
  }
  return false;
}


/// \fn bool  QguListItem::nameAlreadyUsed(QString name)
/// \brief Method that checks if the provided name has been already used within the list view starting at the current list item
/// \param     name of QString  type
/// \exception None
/// \return    Returns TRUE if the provided name has been already used
///  
bool  QguListItem::nameAlreadyUsed(QString name)
{
  // finds if a sibling of this already has this name
  // name does not have path attached
  // note, item already exists, so this->text(Obj_Name_) = name already
  // and must be discounted in search

  if (this->entryType == ListTypeNull) return false;

  QguListItem *nextchild;
  QguListItem *thischild;
    
  // find first child of parent (ie first sibling)
  if(this->parent() == NULL) // at root, so go to ListView
    thischild = (QguListItem *)(this->listView())->firstChild();
  else   // has an item as parent
    thischild = (QguListItem *)(this->parent())->firstChild();
      
  // go through all siblings (children ofthis->parent)
  while(thischild) {
      // name is text of this item, so exclude this from test
      // return true if any sibling has same name
      if (thischild->text(Obj_Name_) == name && thischild != this) return true;
      nextchild = (QguListItem *)thischild->nextSibling();
      thischild = nextchild;
    }
  return false;
}


/// \fn QguListType QguListItem::getEntryType()
/// \brief Method that returns the entry type of the current item
/// \param     None
/// \exception None
/// \return    Returns the type of the current list item as defined within QguListType enumeration.
///  
QguListType QguListItem::getEntryType()
{
   return entryType;
}


/// \fn bool QguListItem::validateWLitems(bool valid)
/// \brief recursive to test if this and its siblings and all their children still refer to valid data objects on the WL
/// \param     valid of bool type
/// \exception None
/// \return    Returns TRUE if the item is valid
///  

bool QguListItem::validateWLitems(bool valid)
{
  // recursive to test if this and its children still refer
  // to valid data objects on the WL
  // valid starts true, but if found false at any stage stays false
  QguListItem * nextItem = this;
  
  while(nextItem){
    if(nextItem->getEntryType() == ListTypeGroup) {
      QguListItem *firstItem = (QguListItem *)nextItem->firstChild();
      valid = firstItem->validateWLitems(valid);
    }
    else if(nextItem->getEntryType() == ListTypeEntry){
      valid = nextItem->validateLink(valid);
    }
    nextItem = (QguListItem *)nextItem->nextSibling();
  }
  return valid;
}


/// \fn bool QguListItem::validateLink(bool valid)
/// \brief Method that checks that the WL context of an item still refers to a valid item on the working list
/// \param     valid of bool type
/// \exception None
/// \return    Returns TRUE if this link is valid.
///
bool QguListItem::validateLink(bool valid)
{
  // checks that the WL context of an item still refers
  // to a valid item on the working list
  // folders don't change value of valid

  if (this->getEntryType() != ListTypeEntry) return valid;
  // if WL item is no longer there, set local item colour red

  QString ctxt = this->getWLcontext();

  if( qsasWindow->WorkingList->getItemByName(ctxt) != NULL) {
       setColour(valid,true);
  }
  else{
       valid = false;
       setColour(valid, false);
  }

  return valid;
}


/// \fn void QguListItem::setColour(bool valid, bool thisValid)
/// \brief Method that sets the colour of the current item depending upon its validity
/// \param     valid     of bool type
/// \param     thisValid of bool type
/// \exception None
/// \return    None
///  
void QguListItem::setColour(bool valid, bool thisValid){

  // valid is false once any item is found invalid
  // thisValid is false if this item is invalid
  // If this item invalid apend warning text
  // If any item is invalid set colour of list red (can't do individually)
  // If thisItem is valid ensure appended flag is removed

  if(QuSlotDebug())
    std::cout << "QguListItem::setColour (valid, thisValid) = (" << valid << ", " << thisValid << ")" <<endl;
	
  static QString flag ="?-Link-?";
  QString txt = this->text(Obj_Name_);
  int pos = txt.findRev(flag);

  if(thisValid){
    if(pos >= 0) {
      // was flagged, now fixed, so clean up
      txt.truncate(pos);
      this->setText(Obj_Name_, txt);
    }
  }
  else{
    if(pos < 0) {
        txt += flag;
	this->setText(Obj_Name_, txt);
    }
  }
 
  QPalette palette = QPalette(qApp->palette());
  QColorGroup active_g = QColorGroup(palette.active());
  QColorGroup inactive_g = QColorGroup(palette.inactive());
  if(valid){
   // use default application colours
  }
  else{
    active_g.setColor(QColorGroup::Highlight,active_g.link());
    active_g.setColor(QColorGroup::HighlightedText,active_g.dark());
    inactive_g.setColor(QColorGroup::Highlight, QColor("#ff8888"));
    inactive_g.setColor(QColorGroup::HighlightedText,inactive_g.mid());
  }
  palette.setActive(active_g);
  palette.setInactive(inactive_g);
  this->listView()->setPalette(palette);

}


/// \fn QString  QguListItem::nameContext()
/// \brief Method that returns the full name and path in the list view hierarchy.
/// \param     None
/// \exception None
/// \return    Returns the name context of the current list item
///  
QString  QguListItem::nameContext(){
  // name context is the full name and path in the list view hierarchy
  // folders are separated by /
  
  QString contxt = this->text(Obj_Name_);
  QguListItem *nextparent;
  
   // follow up hierarchy prepending folder names and separators
  
  QguListItem *thisparent = (QguListItem *) this->parent();
  
  while(thisparent){
    nextparent = (QguListItem *) thisparent->parent();
    contxt.prepend('/');
    contxt.prepend(thisparent->text(Obj_Name_));
    thisparent = nextparent;
  }
  return contxt;
}




/// \fn QString QguListItem::text(int txtCol)
/// \brief Method returnes text for an item that never contains "/" which can cause confusion with folders
/// \param     None
/// \exception None
/// \return    Returns the sanitised text.
///  
QString QguListItem::text(int txtCol){
// note: '}' is forbidden in object names since it is used as a dummy marker for '/'
	QString cleanText = Q3ListViewItem::text(txtCol);
	for(int k=0; k<cleanText.size(); k++)
	{
		if (cleanText[k] == QChar('/')) cleanText[k] = QChar('}');
	}
	return cleanText;
}


/// \fn QString QguListItem::setText(int txtCol, QString txt)
/// \brief Method sets correct context for an item replacing dummy "}" with "/" 
/// \param     txtCol of type int, column for text in listview
/// \param     txt of type QString 
/// \exception None
/// \return    None
///  
void QguListItem::setText(int txtCol, QString txt){
// note: '}' is forbidden in object names since it is used as a dummy marker for '/'
	QString cleanTxt = txt.toLatin1().data();
	for(int k=0; k<cleanTxt.size(); k++)
	{
		if (cleanTxt[k] == QChar('}')) cleanTxt[k] = QChar('/');
	}

	Q3ListViewItem::setText(txtCol, cleanTxt);
  
	return ;
}


/// \fn void QguListItem::removeIfEmpty()
/// \brief Method that removes a folder from the current list view if it is empty.
/// \param     None
/// \exception None
/// \return    None
///  
void QguListItem::removeIfEmpty(){
  
  // Removes folders safely if they are empty. 
  
  if (this->getEntryType() != ListTypeGroup) return;
  
  // if not empty  don't delete
  if (this->firstChild() ) return;
  
  // if locked root don't delete
  if ( !(this->parent()) && 
       ((QguListView *)this->listView())->rootIsLocked() ) return;

  // delete then tell its parent to do same test  - iterates up safely
  QguListItem * thisParent = (QguListItem *)this->parent(); 
  delete this;
  if(thisParent) thisParent->removeIfEmpty();

}


/// \fn QString  QguListItem::getWLcontext()
/// \brief Method that returns the working list context for the current item (name and path of actual data object on WL)
/// \param     None
/// \exception None
/// \return    Returns the working list context for the current item, or empty string if not available.
///  
QString  QguListItem::getWLcontext(){
  // safe call to get the WL context (name and path of item
  // on Working List this represents. Returns empty string
  // if there is no WL context set (should never happen)
  
  static QString WLstr;
  if(WLcontext.isNull()) WLstr = QString("");
  else WLstr = QString(WLcontext.toLatin1().data()); // force deep copy
  return WLstr;
}


/// \fn void  QguListItem::set_Selected(bool choice)
/// \brief Method that sets the current item to selected/unselected depending upon the value of the choice parameter
/// \param     choice of type bool* 
/// \exception None
/// \return    None
///  
void  QguListItem::set_Selected(bool choice){

  // allows us to disable QTs own multiple selection rules
  // while allowing multiple selections
  // changes visual feedback only

  Q3ListView * list = this->listView();
  if(!list) return;
 
  list->setMultiSelection(true); // ensure members of group get selected
  list->setSelected(this, choice);
  list->setMultiSelection(false); // clean up

  return;
}


/// \fn bool QguListItem::isDescendent(QguListItem * child)
/// \brief Method that checks if the child argument is descendent of this, or is this item itself.
/// \param     child of type QguListItem *
/// \exception None
/// \return    Returns true if child is descendent of this, or is this.
///  
bool QguListItem::isDescendent(QguListItem * child){

  // returns true if child is a descendent of this or this itself
  
  if (child == NULL) return false;
  
  if (child == this) return true;
  
  QguListItem *nextItem = (QguListItem *)this->firstChild();
  
  while(nextItem)
  {
    if (child == nextItem) return true;
    if (nextItem->isDescendent(child)) return true;
    nextItem = (QguListItem *)nextItem->nextSibling();
  }
  return false;
}


/// \fn QguListItem * QguListItem::safeParent()
/// \brief Method that returns the parent of this, or the first root level folder if there is no QguListItem parent
/// \param     None
/// \exception None
/// \return    Returns the parent of this, or the first root level folder if this has no parent
///  
QguListItem * QguListItem::safeParent(){
  // root items have no parent  
  // safeParent returns the parent folder if it exists 
  // or the first top root folder
  
  // used to ensure drop on entry is handled safely
  // when the entry is a root item (defensive coding only)
  
  QguListItem *folder;
  folder = (QguListItem*) this->parent();
  if (folder == 0) folder = (QguListItem*)this->listView()->firstChild();
  
  return folder;
}


/// \fn QguListItem * QguListItem::findName(QString name)
/// \brief Method to look up 'name' within the folder containing this item (which MUST be the first item in the folder)
/// \param     name of type QString
/// \exception None
/// \return    Returns the item if found or NULL if not found.
///  
QguListItem * QguListItem::findName(QString name)
{
  // This is ALWAYS called with the firstChild of a folder
  // it finds the item in that parent folder with text(Obj_Name_) = name
  // nextItem is NULL if the name is not found in the parent folder.
  
  QguListItem *nextItem;
    
  nextItem = (QguListItem *)this;

  // will return NULL if not found
     
   while(nextItem){
     if(nextItem->text(Obj_Name_) == name) return nextItem;
     nextItem = (QguListItem *)nextItem->nextSibling();
   }
   
   return nextItem;
}


/// \fn void QguListItem::setChildrenSelected(bool choice)
/// \brief Method that recursively sets an item and any children selected according to arguent 'choice'. For visual feedback only.
/// \param     choice of type bool 
/// \exception None
/// \return    None
///  
void QguListItem::setChildrenSelected(bool choice)
{  
  // recursively sets an item and any children selected
  // but does not change the dragList or localList
  // handles visual feedback only
  
  QguListItem *nextItem;
  this->set_Selected(choice);
  listView()->update();
  
    nextItem = (QguListItem *) this->firstChild();
    while(nextItem){
       nextItem->setChildrenSelected(choice);
       nextItem = (QguListItem *) nextItem->nextSibling();
    }
    
}


/// \fn void QguListItem::refreshWLcontext()
/// \brief Internal method, use QwlListView::refreshWLcontext() which calls QguListItem::refreshWLcontext safely.
/// \param     None
/// \exception None
/// \return    None
///  
void QguListItem::refreshWLcontext()
{
  // Called by QwlListView::refreshWLcontext() for first item of WL list view only
  // which allows a recursive algorithm here to refresh the entire list
  
  QguListItem *nextItem = (QguListItem *)this;
  
  // recursively descends folders
  // For any entry types it resets both the name
  // on the qdos object list and the WL context
  // This is needed to clean up in case something
  // has moved on the WL
  
    while (nextItem)
    {
       if (nextItem->getEntryType() == ListTypeEntry){

	 // update the qdos object list
         QuRenameObject( nextItem->getWLcontext(), nextItem->nameContext() );
	 // update the WL context string
	 nextItem->WLcontext = QString(nextItem->nameContext().toLatin1().data());
       }
       
      // recursion
      ((QguListItem *)nextItem->firstChild())->refreshWLcontext();
      nextItem = (QguListItem *)nextItem->nextSibling();
    }

    return;
}


/// \fn QdObject_var QguListItem::QguGetObject()
/// \brief Method that returns the QdObject_var pointer associated with this item.
/// \param     None
/// \exception None
/// \return    Returns the QdObject_var pointer. If object not found then obj.is_nil()
///  
QdObject_var QguListItem::QguGetObject(){
  
  // returns the qdos data object referred to by the WL item in WL context.
  // Works for item on any QguListView list

  QString name = this->getWLcontext();
  return QuGetObjByName(name);
  
}


/// \fn QguListItem * QguListItem::QguGetNext()
/// \brief Method that returns the next item on the list view starting at this item. first call should be for first item on list
/// \param     None
/// \exception None
/// \return    Returns the next item or NULL when no more items on list
///  
QguListItem * QguListItem::QguGetNext(){

  QguListItem * next;
  QguListItem * folder;
  QguListItem * child;


  // this recursive algorithm only returns listTypeEntry items or NULL
  // but may be called for a folder. 
  // First call must be for first child in list view. 
  // It traverses the whole list view with subsequent calls until
  // list is finished, when it returns NULL
  
  // It could probably be replaced with a listView iterator
  // and test on entryType
  
  
  
  if (this->getEntryType() == ListTypeGroup){
    next = (QguListItem *)this->firstChild();
    while(next){
      if(next->getEntryType() == ListTypeEntry) return next;
      child = next->QguGetNext(); // this will descend folders
      if(child == 0) next=(QguListItem *) next->nextSibling(); // empty folder(s)
      else next = child; // this will be a valid entry
    }
    return NULL;
  }
  
  next = (QguListItem *)this->nextSibling();
  
  // if at end of this folder, go back up and find parent's next sibling
  if(next == 0) {
    folder = (QguListItem *)this->parent();
    
    while(folder != 0){
      next = (QguListItem *)folder->nextSibling();
      if(next == 0){
        // at end of this folder too, back up tree
        folder = (QguListItem *)folder->parent();
      }
      else{
        if(next->getEntryType() == ListTypeEntry)
          return next;
	else // descend new folder
	  return next->QguGetNext();
      }
    }
    
    return NULL;
  }

  // if a folder descend
  if(next->getEntryType() == ListTypeGroup){
    next = next->QguGetNext();
    return next;
  }

  // if an entry return it
  if(next->getEntryType() == ListTypeEntry){
    return next;
  }
  
  return NULL;
}



//     QguListView  CLASS 
//----------------------------

/// \fn QguListView::QguListView(QWidget *parent, const char *name) 
/// \brief QguListView class constructor
/// \param parent The parent object of the list view. It is of type QWidget 
/// \param name   The name of this list view being instantiated. It is a const char*
/// \exception none
/// \return none
///

QguListView::QguListView(QWidget *parent, const char *name) 
   :  Q3ListView(parent,name),mousePressed(false),alreadySelected(false),
      dragItem(0)
{
      // initialise pointer for current working item

   setAcceptDrops(true);
   setFocusPolicy(WHEEL_FOCUS);
   setColumnText(Obj_Name_, QString(name));
   setDragAutoScroll(true);
   setSorting(-1);
   setRootLocked(false);
   
      // create cursor for shift selection
   QPixmap shift_cursor_icon( shift_cursor_xpm);
   ShiftCursor = new QCursor(shift_cursor_icon, 0, 0);
   connect(this, SIGNAL(currentChanged(Q3ListViewItem*)), this, SLOT(clearMsgBar()));
   connect(this,SIGNAL(rightButtonPressed(Q3ListViewItem*,const QPoint&,int)), this,SLOT(rightButtonPressEvent(Q3ListViewItem*,const QPoint&,int)));

   clearSelection();

}


/// \fn QguListView::~QguListView() 
/// \brief QguListView class destructor
/// \param none
/// \exception none
/// \return none
///
QguListView::~QguListView()
{  
}

void QguListView::emitListModified()
{ emit listModified();}

#ifdef QT_VERSION
#if QT_VERSION <0x030000
Q3ListViewItem* QguListView::findItem(const QString& str,int column)
{
  // this is missing from QT2
  // finds named item on list (anywhere, but used only for root
  // items on WL)
  
  Q3ListViewItemIterator it(this);
  while(it.current())
  {
    if(it.current()->text(column)==str)
      return it.current();
    ++it;
  }
  return 0;
}
#endif // end of #if
#endif // end of #ifdef


/// \fn QguListView::clearMsgBar() 
/// \brief Method to empty the list view message bar
/// \param none
/// \exception none
/// \return none
///
void QguListView::clearMsgBar()
{  
  QuMsgBar(" ", false);
}

/// \fn QguListView::listSize() 
/// \brief counts ALL ListTypeEntry items in list view
/// \param none
/// \exception none
/// \return none
///
int QguListView::listSize()
{  
   // counts all entries, folders are not counted
    int counter = 0;
    Q3ListViewItemIterator it( this );
    while ( it.current() ) {
        if ( ((QguListItem *)it.current())->getEntryType() == ListTypeEntry) counter++;
	   ++it;
    }
    return counter;
}


/// \fn void QguListView::::deleteListItems() 
/// \brief Method that deletes all selected items (on localList) from the list view. It warns if non-empty folder is selected.
/// \param parent None
/// \exception none
/// \return none
///
void QguListView::deleteListItems(){

 if(QuSlotDebug())
 {
	printf("QguListView::deleteListItems %ld\n", this);
	fflush(stdout);
 }
 
  // ensure that selectionChanged() is not emitted while deleting
  clearSelection();

  // this function is overloaded for WL QwlListView

 // first warn user if non-empty folder is selected or locked
 for(int i=0; i<localList.size(); i++)
  {
    QguListItem * mouseDownItem = localList[i];
    if(  mouseDownItem->isLocked() ) {
      QuMsgBar("Cannot delete locked item", true);
      return;
    }
       
    if(  mouseDownItem != NULL 
      && mouseDownItem->getEntryType() == ListTypeGroup){
    
      if( mouseDownItem->firstChild() != NULL ) {
        // launch dialogue
        switch( QMessageBox::information( this, "Object Lists",
                                        "Non-empty folder selected",
                                        "Delete", "Cancel",
                                           0,         1 ) ) 
        {
          case 0:
	    break;   // drop through and delete
          case 1:
            return;  // don't delete
        }
      }
    }
  } // end for over localList testing

  // delete entry type items first to avoid trying to delete
  // item in deleted folder
  
  for(int i=0; i<localList.size(); i++)
  {
    QguListItem * mouseDownItem = localList[i];
    if(  mouseDownItem != NULL 
      && mouseDownItem->getEntryType() == ListTypeEntry){
     
      QString str =  mouseDownItem->getWLcontext() ;     
      //QApplication::clipboard()->setData( new Q3TextDrag(str)  );
      QMimeData * clipData=new QMimeData;
      clipData->setText(str);
      QApplication::clipboard()->setMimeData(clipData);
      delete mouseDownItem;
    }
  } // end for over Selected List deleting entries
      
  // delete remaining empty folders
  for(int i=0; i<localList.size(); i++)
  {
    QguListItem * mouseDownItem = localList[i];
    if(  mouseDownItem != NULL 
      && mouseDownItem->getEntryType() == ListTypeGroup){
    
      if( mouseDownItem->firstChild() == NULL ) {
        // empty folder, so ok to delete
        delete mouseDownItem;
      }
    }
  } // end for over localList

  // ensure no deleted item gets selected 
  
  localList.clear();
  
  // only WL deletions need alert other list views
  // validateWLitems();
  
  emit listModified();
}


/// \fn void QguListView::renameSelectedItem() 
/// \brief Method that renames a selected item within the current list view. Launches dialog using QguListView::reNameItem()
/// \param None
/// \exception None
/// \return None
///
void QguListView::renameSelectedItem(){

 if(QuSlotDebug())
 {
	printf("QguListView::renameSelectedItem %ld\n", this);
	fflush(stdout);
 }
 
 // this is implemented only on the QT selection since
 // we can only handle one selected item
 
  QguListItem *selected_item;
  
  selected_item = (QguListItem *)this->selectedItem();
  if (selected_item == NULL) return;

  reNameItem(selected_item);
}


/// \fn void QguListView::reNameItem(QguListItem * item) 
/// \brief Method that launches dialog to re-name item. Ensures name is unique.
/// \param item of type QguListItem *
/// \exception None
/// \return None
///
 bool QguListView::reNameItem(QguListItem * item)
{
  // launch a dialog box to change an item's name
  // insist it is unique
 // change text of this item
  
  QString userNameChoice; 
  bool ok;

  // don't let user see substituting for '/' which is done when getting
  // text() to avoid confusion in the code with directories
  
  QString cleanTxt = QString(item->text(Obj_Name_).toLatin1().data());  // force deep copy
  int posn = cleanTxt.find("}");
  while (posn >= 0) {
    cleanTxt.replace(posn, 1, "/");
    posn = cleanTxt.find("}");
  }
       
  // launch getText box with current item
  
  userNameChoice = QuTextInputDialog("Rename", "Enter new name",  
                     cleanTxt, &ok, item->listView());
    
  if (ok && !(userNameChoice.isEmpty()) ){
    if(userNameChoice.find("}") >= 0) {
      // loop until user doesn't use '}'
       userNameChoice = QuTextInputDialog("Rename", " '}' is not allowed in name",
                         userNameChoice.toLatin1().data(), &ok, item->listView());
    }


    while(ok && item->nameAlreadyUsed(userNameChoice) ){
       userNameChoice = QuTextInputDialog("Rename", "Name used",  
                          item->text(Obj_Name_).toLatin1().data(), &ok, item->listView());    
    
    } 
    if (ok) {     
      item->setText( Obj_Name_, userNameChoice );
    }
    else return false;
  }
  else return false;
  
  // if a data object has been modified let the world know
  if(item->getEntryType() == ListTypeEntry) emit listModified();

  return true;
}


/// \fn void QguListView::selectAll() 
/// \brief Method that selects all the items within the current list view and puts them on localList
/// \param None
/// \exception None
/// \return None
///
void QguListView::selectAll(){

 if(QuSlotDebug())
 {
	printf("QguListView::selectAll %ld\n", this);
	fflush(stdout);
 }
 

  QguListItem *nextItem;
  
  clearSelectedList();

  nextItem = (QguListItem *) firstChild();
  while(nextItem){
   AddToSelectedList(nextItem);
   nextItem = (QguListItem *) nextItem->nextSibling();
  }
}


/// \fn void QguListView::makeGroup()
/// \brief Method that creates a new folder containing all items on localList (ie selected). Inserts folder in place of selected items or at root if nothing is selected.
/// \param None
/// \exception None
/// \return None
///
void QguListView::makeGroup(){

 if(QuSlotDebug())
 {
	printf("QguListView::makeGroup %ld\n", this);
	fflush(stdout);
 }
 
 
  // this function is overloaded for Wl QwlListView
    
  QguListItem *newFolder;
  bool ok;

  // make new group item in folder or parent of selected item
  
    // if nothing is selected put at root (if allowed)
    if(localList.size() == 0){  
      if( this->rootIsLocked() )
        // root is locked, ie Working List, so there must be a first child
        newFolder =  new QguListItem((QguListItem*)this->firstChild(), QString("Group"), ListTypeGroup, 0);
      else{ 
        newFolder =  new QguListItem(this, QString("Group"), ListTypeGroup, 0);
      } 
            
      clearSelection();
      newFolder->setSelected(true);
      reNameItem(newFolder);  // launch re-name text box  
      return;
    }
    
    // else something is selected, find where to put folder
    QguListItem *commonRoot = findCommonFolder();   
    if(commonRoot != NULL)
    {
      newFolder =  new QguListItem (commonRoot, QString("Group"), ListTypeGroup);    
    }
    else if( !(this->rootIsLocked()) ){
      newFolder =  new QguListItem (this, QString("Group"), ListTypeGroup);    
    }
    else return;  // cannot find where to put folder that is allowed
    
    clearSelection();
    
    ok = reNameItem(newFolder);  // launch re-name text box
    if(!ok) {
      delete newFolder;
      return;
    }
    // move selected items to this new folder
    
    for (int i=0; i<localList.size(); i++){
      
      // move item from selection list
      QguListItem * movingItem = (QguListItem *)localList[i];
      QguListItem * parentItem = (QguListItem *)movingItem->parent();
	
      // move from old folder or listView
      if (parentItem == 0) this->takeItem(movingItem);
      else                 parentItem->takeItem(movingItem);
	
      newFolder->insertItem(movingItem);
	
    }
    clearSelectedList();
    
}


/// \fn void QguListView::setRootLocked(bool lock)
/// \brief Method that sets the value of the private data member rootLock
/// \param lock which is of bool type
/// \exception None
/// \return None
///
void QguListView::setRootLocked(bool lock)
{ 
   rootLock = lock;
} 
 

/// \fn bool QguListView::rootIsLocked()
/// \brief Method that returns the value of the private data member rootLock
/// \param none
/// \exception None
/// \return bool - value of the rootLock data member
/// 
bool QguListView::rootIsLocked()
{ 
   return rootLock;
}
 

/// \fn void QguListView::contentsMouseDoubleClickEvent(QMouseEvent *event)
/// \brief Method that processes the mouse double-click event. It has been re-implemented from Q3ListView
/// \param event of type QMouseEvent * 
/// \exception None
/// \return None
/// 
 void QguListView::contentsMouseDoubleClickEvent(QMouseEvent *event)
{
  // disable mouse down on right and middle button
  // forces launch to be from left button and protects
  // from two mouse down events causing drag problems
  
   if( event->button() == Qt::RightButton) return;   
   if( event->button() == Qt::MidButton) return;
   
   if(event->state() == Qt::ControlButton || event->state() == Qt::ShiftButton)
         return;

  // clean up localList as double click only applies to single item
   clearSelectedList();
   
   QguListItem *atItem = (QguListItem *)currentItem() ;  
   if(atItem  == NULL) return;

   if(atItem->getEntryType() == ListTypeGroup) {
      // open/close group view
      Q3ListView::contentsMouseDoubleClickEvent(event);
   }  
   else{
       // view linked data object on WL
       // find item on WL from WLcontext and set it to be current item
     QguListItem * wli = qsasWindow->WorkingList->getItemByName(atItem->WLcontext);
     qsasWindow->WorkingList->setCurrentItem(wli);
     // tell WL to browse current item
     viewWL_ObjRequest();
   }
}


/// \fn void QguListView::reLinkItem()
/// \brief Method that relinks the selected item to a new object within the working list.
///
///  This is used to change the data object that is being referred to. Not meaningful in the WL listview.
/// \param none
/// \exception none
/// \return none
///
 void QguListView::reLinkItem()
{ 
   bool ok = FALSE;
   QguListItem *item;
   int oldLen;
   
   if(localList.size() == 0) return;
   
   // easy case of single item selected
   if(localList.size() == 1) {
     item = localList[0];
     if(item == 0 || item->getEntryType() != ListTypeEntry ) return;
     
     // ask user
     QString newContext = QuTextInputDialog( "Link to Working List", 
                                            "re-link to new Item?", 
				            item->getWLcontext().toLatin1().data(), 
					    &ok, this );
     if ( ok && !(newContext.isEmpty()) ){
       item->WLcontext = newContext;// user entered something and pressed ok
       // see if user input works (drag and drop should work here too
       validateWLitems();
     }
     return;
   }
   
   // relink list with wild card
   QString commonContext;
   for(int i= 0; i<localList.size(); i++){   
     item = localList[i];
     if(item == 0 || item->getEntryType() != ListTypeEntry ) continue;
     
     // find common stem in WL contexts 
     // this needs improving to common sub-strings and wild cards
     
     QString itemContext = item->getWLcontext();
     if(i==0){
       commonContext = QString(itemContext.toLatin1().data());
     }
     else{
       int d=0;
       commonContext.truncate(itemContext.length());
       while(  d < commonContext.length() 
            && commonContext[d] == itemContext[d] ){
         d++;
       }
       commonContext.truncate(d);
     }
   }
   oldLen = commonContext.length();
   commonContext += QString("***");
   
   QString newContext = QuTextInputDialog( "Link to Working List", 
                                           "re-link to new Item?", 
				            commonContext.toLatin1().data(), 
					    &ok, this );
   if ( ok && !(newContext.isEmpty()) ){
   
     int posn = newContext.findRev("***");
     newContext.truncate(posn);
    
     // fix those that match
     for(int j= 0; j<localList.size(); j++){   
       item = localList[j];
       if(item == 0 || item->getEntryType() != ListTypeEntry ) continue; 
       
       item->WLcontext.replace(0, oldLen, newContext);// user entered something and pressed ok
     }
     // tell list to test new WL links
     validateWLitems();
   }
   

} 
 

/// \fn void QguListView::clearSelection()
/// \brief Method that sets all the items to be not selected. Re-implemented from Q3ListView. 
/// \param none
/// \exception none
/// \return none
///
void QguListView::clearSelection()
{
  // call recursive setChildrenSelected(false) for each item in list view
  QguListItem *next = (QguListItem *)this->firstChild();
  while( next != NULL){
    next->setChildrenSelected(false);
    next = (QguListItem *)next->nextSibling();
  }
} 


/// \fn void QguListView::contentsMousePressEvent(QMouseEvent *event)
/// \brief Method to process mouse down events. Re-implemented from Q3ListView. Handles select, multi-select and drag
/// \param event of type QMouseEvent * 
/// \exception none
/// \return none
/// 
void QguListView::contentsMousePressEvent(QMouseEvent *event)
{
  // disable mouse down on right and middle button
  // forces drag to be from left button and protects
  // from two mouse down events causing drag crash
    
   if(QuSlotDebug())
  	printf("QguListView::contentsMousePressEvent %ld\n", this);
  
	QguListItem *selected_item = NULL;
	QguListItem *ancestor      = NULL;
	QguListItem *descendent    = NULL;
	static QguListItem *lastSelection = NULL;
   
	bSingleSelection = false;

	alreadySelected = false;
	mousePressed    = false;
	dragItem        = NULL;
   
	if ( localList.size() == 0 )
	{
	     lastSelection = NULL;
	}
	
	if( event->button() == Qt::MidButton) return;

   // find item under click, allowing for scrolling
   selected_item = (QguListItem *)itemAt(event->pos() - QPoint(0,this->contentsY()) );
   if (selected_item == NULL) {
     clearSelectedList();
     return;
   }
   
   // spacers are untouchable
   if(selected_item->getEntryType() == ListTypeNull) {
     clearSelectedList();
     return;
   } 
   
  
  // Right Button
  
   if( event->button() == Qt::RightButton) {
     clearSelectedList();
     setSelected(selected_item, true);
     AddToSelectedList(selected_item );
     refreshSelectedList();
     // force signal to be emitted because this fn is overloaded
     emit rightButtonPressed(selected_item,QCursor::pos(),0); 
     return;   
   }
  
  
   // Left Button
   
   // Shift Selection

   if( event->state() == Qt::ShiftButton) {
     if(QApplication::overrideCursor()!=ShiftCursor)
       QApplication::setOverrideCursor(*ShiftCursor);
     
     //  delete/add as appropriate
     if( isOnSelectedList(selected_item) )
     {
       clearSelectedList();
       addItemsInclusive(selected_item, lastSelection);
     }
     else if (( ancestor = findAncestorOnList(selected_item) ) != NULL)
     {
       clearSelectedList();
       addItemsInclusive(ancestor, lastSelection);
     }
     else if (( descendent = findDescendentOnList(selected_item) ) != NULL)
     {
       RemoveFromSelectedList(descendent);
       addItemsInclusive(selected_item, lastSelection);
     }
     else{
       addItemsInclusive(selected_item, lastSelection);
     }
     
     lastSelection = selected_item;
     refreshSelectedList();
     return;
     
   }
   
   // Control Selection
   
   else if(event->state() == Qt::ControlButton ) {
     
     if(QApplication::overrideCursor()!=ShiftCursor)
       QApplication::setOverrideCursor(*ShiftCursor);
       
     //  delete/add as appropriate
     if( isOnSelectedList(selected_item) )
     {
       RemoveFromSelectedList(selected_item);
     }
     else if (( ancestor = findAncestorOnList(selected_item) ) != NULL)
     {
       // item is in folder that was selected already, so remove
       RemoveFromSelectedList(ancestor);
     }
     else if (( descendent = findDescendentOnList(selected_item) ) != NULL)
     { 
       RemoveFromSelectedList(descendent);
       AddToSelectedList(selected_item);
     }
     else{
       AddToSelectedList(selected_item);
     }
      
     lastSelection = selected_item;
     refreshSelectedList();
     return;
   }

   // Single selection

   else  {
     if(QApplication::overrideCursor())
       QApplication::restoreOverrideCursor();
     lastSelection = selected_item;
     setCurrentItem(selected_item);
   }
   // When decoration hit - avoid confusion from resizing list
   // - also prevents drag from decoration

   int n_indent = selected_item->depth();
   if(++n_indent*(this->treeStepSize()) > event->pos().x() ) {
     // do open/close on item if group
     Q3ListView::contentsMousePressEvent(event);
     return;
   }

   dragFromList = this;  // to keep track of drag between lists

   // drag if mouse over a highlighted item try to drag selections
   if( findAncestorOnList(selected_item) || isOnSelectedList(selected_item)) {
     Q3ListView::setCurrentItem( selected_item );
     refreshSelectedList();
     
     // copy to dragList before drag
     setDragList();
     // store item selected for dragging
     dragItem=selected_item;
     // store start position of drag event
     dragPos=event->pos();
     // set flag to signal left mouse pressed 
     mousePressed=true;
     // set flag to signal item ALREADY selected 
     alreadySelected=true;
     
   }
   else{
     // single click on item not selected
     clearSelectedList();
     AddToSelectedList(selected_item);
     Q3ListView::setCurrentItem( selected_item );
     refreshSelectedList();
     // copy to dragList before drag
     setDragList();
     // store item selected for dragging
     dragItem=selected_item;
     // store start position of drag event
     dragPos=event->pos();
     // set flag to signal left mouse pressed
     mousePressed=true;
     // set single selection falg
	bSingleSelection = true;
   }
  
}


/// \fn void QguListView::contentsMouseReleaseEvent(QMouseEvent *event)
/// \brief Method that processes the mouse up event. It has been re-implemented from Q3ListView
/// \param event of type QMouseEvent * 
/// \exception none
/// \return none
/// 
void QguListView::contentsMouseReleaseEvent(QMouseEvent *event)
{
// called if mouse pressed and released (i.e. not held down and moved) and
// after contentsMousePressEvent

   if(QuSlotDebug())
  	printf("QguListView::contentsMouseReleaseEvent %ld\n", this);
  
  mousePressed=false;
  // dragFromList only set up if left button clicked
  if( dragFromList ) 
  {
    // if items alreadySelected on this dragFromList, then an unselection
    // has been made - so clear localList for this dragFromList listview
    if(dragFromList->alreadySelected)	  
      dragFromList->clearSelectedList();
    // either way...refresh localList for this dragFromList listview
    dragFromList->refreshSelectedList();	  
  }
}


/// \fn void QguListView::contentsMouseMoveEvent(QMouseEvent *event)
/// \brief Method that processes the mouse move event. It has been re-implemented from Q3ListView
/// \param event of type QMouseEvent * 
/// \exception none
/// \return none
/// 
void QguListView::contentsMouseMoveEvent(QMouseEvent *event)
{
// called if mouse held down and moved and
// after contentsMousePressEvent

 
  // check left mouse button pressed down 	
  if(mousePressed)
  {
	// check drag item selected	
    if(dragItem)
    {
      // if mouse has moved more than 3 pixels in the x or/and y direction
      // from the starting position (the position of the mousePressEvent)
      // then assume a drag operation
      if(abs(event->pos().x() - dragPos.x() ) >=3 ||
        abs(event->pos().y() - dragPos.y() ) >= 3)
      {
	 // start drag     
//          Q3TextDrag *d = new Q3TextDrag( dragItem->nameContext(), this );
//          d->drag();
	      QDrag *d = new QDrag(this);
	      QMimeData *mimeData = new QMimeData;
	      mimeData->setText(dragItem->nameContext());
	      d->setMimeData(mimeData);
#if QT_VERSION < 0x040300
	      d->start();
#else
	      d->exec();
#endif
	  
		 
		// reset mousePressed flag
		mousePressed=false;    
	
	// tidy up after drag i.e unselect items (may not be desired ) if already selected 
         if( dragFromList ) 
         {
           //if(dragFromList->alreadySelected)
           //  dragFromList->clearSelectedList();
           dragFromList->refreshSelectedList();	  
         }
      }
    }
  }
  Q3ListView::contentsMouseMoveEvent( event );
}
	

/// \fn void QguListView::setDragList()
/// \brief Method to copy the localList (selected items) to the dragList to use for drag and drop
/// \param none 
/// \exception none
/// \return none
/// 
void QguListView::setDragList()
{
  // copy localList to drag list - used for drag-n-drop and copy
  
  dragList.clear();
  for (int i=0; i<localList.size(); i++){
    dragList.push_back(localList[i]);
  }
}


/// \fn void QguListView::dragMoveEvent(QDragMoveEvent *event)
/// \brief Event handler that is called when a drag is in progress. Re-implemented from QWidget
/// \param event of type QDragMoveEvent * 
/// \exception none
/// \return none
/// 
void QguListView::dragMoveEvent(QDragMoveEvent *event)
{
  QguListItem *item;
  QWidget *window;
 

  window = this->parentWidget();
  emit raiseWindow();
  if(window->isActiveWindow() ){
         
    //event->acceptAction( Q3TextDrag::canDecode(event) );
	if (event->mimeData()->hasText())
		  event->acceptProposedAction();
    
    item = (QguListItem*)itemAt(event->pos()-QPoint(23,23));
    if(item != 0){
      setCurrentItem(item);
    }
    
   QPoint scrollPoint = event->pos() + QPoint(this->contentsX(),this->contentsY());
    ensureVisible(scrollPoint.x(),scrollPoint.y()-20,25,25);
  }
  this->refreshSelectedList();

}

void QguListView::dragEnterEvent(QDragEnterEvent *event)
{
    //event->accept(Q3TextDrag::canDecode(event));
	if (event->mimeData()->hasText())
		event->acceptProposedAction();
}

/// \fn QguListItem * QguListView::findCommonRoot()
/// \brief Method that returns the deepest folder that contains all the selected items (on localList)
/// 
/// Differs from findCommonFolder() only in that if only one item is selected it returns the selected item.
/// \param None
/// \exception none
/// \return Returns QgListItem* or NULL if there is no common root
/// 
QguListItem * QguListView::findCommonRoot()
{
  // will return NULL if there is no common root (except ListView)
  
   QguListItem *commonRoot = NULL;
   if(localList.size() == 0) return commonRoot;
   
   // start anywhere, and move outwards
   commonRoot = (QguListItem *)localList[0];
  
   while (commonRoot && !(isCommonAncestor(commonRoot))){
     commonRoot = (QguListItem *)commonRoot->parent();
   } 
   
   return commonRoot;
}


/// \fn QguListItem * QguListView::findCommonFolder()
/// \brief Method that returns the deepest folder that contains all the selected items (on localList)
/// 
/// Differs from findCommonRoot() only in that if only one item is selected it returns the folder holding the selected item.
/// \param None
/// \exception none
/// \return Returns QgListItem* or NULL if there is no common root
/// 
QguListItem * QguListView::findCommonFolder()
{
  // will return NULL if there is no common root (except ListView)
  
   QguListItem *commonFolder = NULL;
   if(localList.size() == 0) return commonFolder;
   
   // start anywhere, and move outwards
   commonFolder = (QguListItem *)localList[0]->parent();
  
   while (commonFolder && !(isCommonAncestor(commonFolder))){
     commonFolder = (QguListItem *)commonFolder->parent();
   } 
   
   return commonFolder;
}


/// \fn bool QguListView::isCommonAncestor(QguListItem *folder)
/// \brief Method that checks if folder is a common ancestor for all selected items (in localList)
/// \param folder of type QguListItem * 
/// \exception none
/// \return  True if folder is an ancestor of all selected items or FALSE if any selected item is not descended from folder
/// 
bool QguListView::isCommonAncestor(QguListItem *folder){

  for(int i=0; i<localList.size(); i++)
  {
    // search parents of every selected item to see
    // if the folder is an ancestor
    
    QguListItem *thisItem = localList[i];
    while(thisItem != folder)
    {
      thisItem = (QguListItem *)thisItem->parent();
      if (thisItem == NULL) return false; // not ancestor of Selected Item
    }
    // found folder is ancestor carry on with list
  }
  
  // found folder as ancestor of whole list
  return true; 
}


/// \fn void QguListView::clearSelectedList()
/// \brief Method that deselects items in this list view and triggers an update. localList is emptied.
/// \param None
/// \exception none
/// \return None
/// 
void QguListView::clearSelectedList()
{    
  // clear list and visual feedback for selections in this listView
  localList.clear();
  
  clearSelection();
}


/// \fn void QguListView::refreshSelectedList()
/// \brief Method that refreshes the list of selected items within the current list view based on contents of localList.
/// \param None
/// \exception none
/// \return None
/// 

void QguListView::refreshSelectedList()
{    
  // refreshes list view according to selections ln localList
  clearSelection();
  if(localList.size() == 0) return;

  for(int i=0; i<localList.size(); i++){
    // set item & descendents selected
    ((QguListItem *)localList[i])->set_Selected(true);
  }
  return;
}


/// \fn bool QguListView::isOnSelectedList(QguListItem *item)
/// \brief Method that checks if a given item is within the selected list of the current list view
/// \param item of type QguListItem * 
/// \exception none
/// \return Returns TRUE if the item belongs to the selected list of items (localList) and FALSE if it does not 
/// 
bool QguListView::isOnSelectedList(QguListItem *item)
{
 // returns true if item is already on the localList
 
  if (item == 0 ) return false;
  
  for(int i=0; i<localList.size(); i++){
    if (localList[i] == item) return true;
  }

  // item not on list
  return false;
}


/// \fn QguListItem * QguListView::findAncestorOnList(QguListItem *item)
/// \brief Method that locates an ancestor of the specified item on the list of selected items (localList)
/// \param item of type QguListItem * 
/// \exception none
/// \return Returns the ancestor if found or NULL if not found
/// 
QguListItem * QguListView::findAncestorOnList(QguListItem *item)
{
  // this goes back up the heirarchy until a parent is found that
  // is on the localList and returns this parent.
  // the item itself being on the list is handled separately
  
  QguListItem * nextItem;
  
  if (item == 0 ) return NULL;
  
  // item itself is NOT valid ancestor in this context
  if ( isOnSelectedList(item) ) return NULL;
  
  nextItem = (QguListItem *)item->parent();
  while(nextItem){
    if( isOnSelectedList(nextItem) ) return nextItem;
    nextItem = (QguListItem *)nextItem->parent();
  }

  // item ancestors not on list
  return NULL;
}


/// \fn QguListItem * QguListView::findDescendentOnList(QguListItem *item)
/// \brief Method that finds a descendent of a given item within the list of selected items (localList)
/// \param item of type QguListItem * 
/// \exception none
/// \return Returns the descendent if found or NULL if not found
/// 
QguListItem * QguListView::findDescendentOnList(QguListItem *item)
{
  // this steps down through the hierarchy until a child is
  // found on the localList (which is returned) otherwise returns 0
  
  QguListItem * nextItem;
  QguListItem * foundItem;
  
  if (item == 0 ) return NULL;
  
  // item itself is valid descendent in this context
  if ( isOnSelectedList(item) ) return item;
  
  nextItem = (QguListItem *)item->firstChild();
  while(nextItem){
    if( (foundItem = findDescendentOnList(nextItem)) != NULL ) return foundItem;
    nextItem = (QguListItem *)nextItem->nextSibling();
  }

  // item and descendents not on list
  return NULL;
}


/// \fn QguListItem * QguListView::findDescendentOnList(QguListView *lv)
/// \brief Method that finds a descendent of any item on a list view within the list of selected items (localList)
/// \param lv of QguListView * type
/// \exception none
/// \return Returns a descendent if found or NULL if not found
/// 
QguListItem * QguListView::findDescendentOnList(QguListView *lv)
{
  // same as above, but allows this to be called for a listView
  
  QguListItem * nextItem;
  QguListItem * foundItem;
  
  if (lv == 0 ) return NULL;
    
  nextItem = (QguListItem *)lv->firstChild();
  while(nextItem){
    if( (foundItem = findDescendentOnList(nextItem)) != NULL ) return foundItem;
    nextItem = (QguListItem *)nextItem->nextSibling();
  }

  // item and descendents not on list
  return NULL;
}


/// \fn void QguListView::setSelectedList(QguListItem *item)
/// \brief Method that sets the selected list (localList) to contain item and all its descendents
/// \param item of type QguListItem * 
/// \exception none
/// \return none
/// 
void QguListView::setSelectedList(QguListItem *item)
{
  // sets the localList to contain the item and all its descendents
  if(QuSlotDebug())
  {
	printf("QguListView::setSelectedList %ld %ld\n", this, item);
	fflush(stdout);
  }
  
  if (item == 0 ) return;
  
  localList.clear();
  clearSelection();
  
  if ( !(item->isLocked()) ) 
  {
    localList.push_back(item);
    item->set_Selected(true);
  }
  
  QguListItem *nextItem=dynamic_cast<QguListItem*>(item->firstChild());
  while(nextItem)
  {
    putChildrenOnList(nextItem);
    nextItem=(QguListItem*)nextItem->nextSibling();
  }
  
  emit listModified();
  return;  
}


/// \fn void QguListView::putChildrenOnList(QguListItem *item)
/// \brief Method that adds an item and all its descendents to the selected list (localList)
/// \param item of type QguListItem * 
/// \exception none
/// \return none
/// 
void QguListView::putChildrenOnList(QguListItem *item){

  // adds an item and all its descendents to the local list
  // this is used by setSelectedList and does not need to worry
  // about whether anything is already on the list
  // (differs in this respect from AddToSelectedList() )
  
  if (item == 0 ) return;
 
  localList.push_back(item);
  item->set_Selected(true);

  QguListItem *nextItem=dynamic_cast<QguListItem*>(item->firstChild());
  while(nextItem)
  {
    putChildrenOnList(nextItem);
    nextItem=(QguListItem*)nextItem->nextSibling();
  }
}


/// \fn void QguListView::AddToSelectedList(QguListItem *item)
/// \brief Method that inserts a given item and its descendents in the select list (localList) of the current list view
///
/// It protects against putting duplicates on the list in mulit-selection mode
/// \param item of type QguListItem * 
/// \exception none
/// \return None
/// 
void QguListView::AddToSelectedList(QguListItem *item)
{

  // this is used by the mouse click methods for adding items
  // and all descendents to the selected list. 
  // It takes care of ensuring items do not get put on twice
  // The iterator descends from the selected item and is
  // stopped when it reaches nextItem that is not a descendent of item
  if(QuSlotDebug())
  {
	printf("QguListView::AddToSelectedList %ld %ld\n", this, item);
	fflush(stdout);
  }
  
  if (item == 0 ) return;
  
  Q3ListViewItemIterator it(item);
  QguListItem *nextItem=0;
  while(it.current())
  {
     nextItem=dynamic_cast<QguListItem*>(it.current());
     if(nextItem)
     {
       if(!(item->isDescendent(nextItem)))
         break;
       if( !(isOnSelectedList(nextItem)) && !(nextItem->isLocked()))  
       {
         localList.push_back(nextItem);
         nextItem->set_Selected(true);
       }
     }
     ++it;
  }

  emit selectionAdded();
  return;  
}


/// \fn void QguListView::RemoveFromSelectedList(QguListItem *item)
/// \brief Method that removes a given item from the list of selected items (localList) of the current list view
/// \param item of type QguListItem * 
/// \exception none
/// \return None
/// 
void QguListView::RemoveFromSelectedList(QguListItem *item)
{
  // removes an item and all its children from the 
  // selected list
  
  if (item == 0 ) return;
  
  vector<QguListItem*>::iterator itr=localList.end();
  
  for(int i=localList.size()-1;i>=0; i--)
  {
    itr--;
    if (item->isDescendent(localList[i])) {
      localList[i]->set_Selected(false);
      localList.erase(itr);
    }
  }
 
  emit listModified();
     
}


/// \fn QguListItem * QguListView::root(QguListItem *item)
/// \brief Method that returns the root (highest level folder) of item on this list view
/// \param item of type QguListItem * 
/// \exception none
/// \return Returns the root of the item in the current list view
/// 
QguListItem * QguListView::root(QguListItem *item)
{
  // locates the root item on the listView that has item
  // as a descendent
  
  QguListItem *root = item;
    while(root->parent() != NULL) root = (QguListItem *) root->parent();
  return root;
}


/// \fn QguListItem * QguListView::getItemByName(QString name)
/// \brief Method that finds an item in this list view where name is its full path (e.g. WLcontext).
/// \param name of type QString 
/// \exception none
/// \return Returns the item if found or NULL if not found
/// 
QguListItem * QguListView::getItemByName(QString name)
{
  // will find an item in a listView from the full
  // path name (WLcontext) of the item
  // It strips folder names of successively until only the item name
  // remains.
  
   // copy name into two temp strings
   // The path delimiter '/' will be used to split both
   // strings, one keeping the path and the other the residual
   
   QString nameCopy(name);
   QString nameToken(name);
   
   QguListItem *foundItem = NULL;
  // will return NULL if not found
   QguListItem *nextItem = (QguListItem *)this->firstChild();

   int tokLen = nameToken.find("/");   
   if( tokLen >= 0){
     nameToken.truncate(tokLen);
     nameCopy.remove(0,tokLen+1);
   }
   else nameCopy = "";
         
   while (nameToken != ""){
     // find folder and return null if not found
     foundItem = nextItem->findName(nameToken);
     if (foundItem == NULL) return foundItem;
     
     // descend into folder and split off next folder name
     nextItem = (QguListItem *)foundItem->firstChild();
     nameToken = nameCopy;
     int tokLen = nameToken.find("/");
     if( tokLen >= 0){
       nameToken.truncate(tokLen);
       nameCopy.remove(0,tokLen+1);
     }
     else nameCopy = "";
     // continue while loop until no path remains in name
   }
   
   return foundItem;
}


/// \fn void QguListView::QguMoveItem(QguListItem *item, QguListItem * target)
/// \brief Method that moves item to become child of target in the current list view
/// \param item of type QguListItem* 
/// \param target of type QguListItem* 
/// \exception none
/// \return None
/// 
void QguListView::QguMoveItem(QguListItem *item, QguListItem * target)
{
  // moves item (and its descendents) to become child of target
  // takes care to prevent badly formed or fatal operations

  if(item == NULL) return;
  
  //cannot drop into your own descendent
  if (item->isDescendent(target)) return;
  
  // takeItem must be done from parent (or listView if root item)
  // this removes item and descendents from listView without
  // destroying them, but they are in limbo and cannot safely be accessed
  
  if(item->parent()){
    item->parent()->takeItem(item);
  }
  else{
    this->takeItem(item);   
  }
   
  // insert the item as appropriate
   
  // if dropAt is root, drag everythying to root
  if(target == NULL) {
    this->insertItem(item);
  }
    // if target is not a folder, insert item in parent
    // and move to below it in list
  else if (target->getEntryType() == ListTypeEntry)
  {
    if(target->parent())  // use parent folder
    {
      target->parent()->insertItem(item);
      item->moveItem(target);
    }
    else  // no parent folder, insert in listView
    {
      this->insertItem(item);
      item->moveItem(target);
    }
  }
    // if target is a folder insert in folder
  else if (target->getEntryType() == ListTypeGroup)
    target->insertItem(item);
  
  else if (target->getEntryType() == ListTypeNull){
      // drag over a separator, insert below folder above
      
      QguListItem * itAbove = (QguListItem *)target->itemAbove();
      if(itAbove->parent())
        if(itAbove->parent()->parent()){
          // this alows user to drag below a folder
          itAbove->parent()->parent()->insertItem(item);
	  item->moveItem(itAbove->parent());
	}
	else{
	  itAbove->parent()->insertItem(item);
	  item->moveItem(itAbove);
	}
      else
        this->insertItem(item);

  }    
  else  // default drop at listView level (defensive only)
    this->insertItem(item);
     
  emit listModified();
  return;      
}


/// \fn void QguListView::insertEntries(QguListItem *dropAt, QguListItem *item)
/// \brief Method that inserts a new item at target dropAt in the current list view. Item takes WLcontext as its name
/// \param dropAt of type QguListItem* 
/// \param item of type QguListItem* 
/// \exception none
/// \return None
/// 
void QguListView::insertEntries(QguListItem *dropAt, QguListItem *item){

  // utility call for non-WL widgets to insert WL context as item
  // when creating entry from drag or copy off WL
  // if item is a folder it is ignored - hierarchy is not maintained
  
  QguListItem * newItem;
  
  // only insert entry type items (items for data objects)
  
  if(item->getEntryType() == ListTypeEntry){
    
    // if dropAt target is null (blank list space) drop at top level
    if(dropAt == 0 ) {
       newItem = new QguListItem(this, item->getWLcontext(), 
	                                       ListTypeEntry);    
    }
    // if target is an entry type drop into its parent
    else if(dropAt->getEntryType() == ListTypeEntry){
      if( dropAt->parent() ){
         newItem = new QguListItem ((QguListItem *)dropAt->parent(), 
	                           item->getWLcontext(),  ListTypeEntry);
         if(newItem != NULL) newItem->moveItem(dropAt);
      }
      else{ // no parent means target is a root item, drop in listView
        newItem = new QguListItem (this, item->getWLcontext(), 
	                                          ListTypeEntry);          
      }
    } 
    // target is a folder, so drop into it
    else{
       newItem = new QguListItem (dropAt, item->getWLcontext(), 
	                                          ListTypeEntry);
    }
  }

}


/// \fn void QguListView::rightButtonPressEvent(Q3ListViewItem *item, const QPoint &P,int c)
/// \brief Method that invokes the context popup menu synchronously at the specified global point
/// \param item of type QguListItem* 
/// \param P    of type const QPoint 
/// \param c    of type int          
/// \exception None
/// \return    None
///  
void QguListView::rightButtonPressEvent(Q3ListViewItem *item, const QPoint &P,int c)
{    
   //  RtButton menu item gets launched
	RButtonMenu->exec(P);
}


/// \fn void QguListView::copy()
/// \brief Method that copies the WLcontext of the selected item of the current list view onto the clipboard
/// \param     None
/// \exception None
/// \return    None
///  
void QguListView::copy(){

 if(QuSlotDebug())
 {
	printf("QguListView::copy %ld\n", this);
	fflush(stdout);
 }
 
 // copy the WLcontext of selected item to clipboard
 // copy always refers to the WL data object not the local name
 
   QguListItem *selected_item = (QguListItem *)selectedItem();
   
   if(selected_item){
     QString str =  selected_item->getWLcontext() ;

     //QApplication::clipboard()->setData( new Q3TextDrag(str, 0) );
     QMimeData * clipData=new QMimeData;
     clipData->setText(str);
     QApplication::clipboard()->setMimeData(clipData);
   }
   return;
}


/// \fn void QguListView::paste()
/// \brief Method that pastes a new item into the current list view using the content of the clipboard as WLcontext. 
/// \param     None
/// \exception None
/// \return    None
///  
void QguListView::paste(){

 if(QuSlotDebug())
 {
	printf("QguListView::paste %ld\n", this);
	fflush(stdout);
 }
 
 // insert an item with WLcontext pointing to WL item
 // identified by text in copy and paste
 // NOTE: insertEntries() is overloaded for most classes based
 // on QguListView, so paste does not need overloading
 
  QString text;
  
  QguListItem *dropAtItem = (QguListItem *)selectedItem();
    
  if ( QApplication::clipboard()->mimeData()->hasText()/*Q3TextDrag::decode(QApplication::clipboard()->data(), text) */){
	  text=QApplication::clipboard()->mimeData()->text();
      	QguListItem * item = qsasWindow->WorkingList->getItemByName(text);
        if (item) insertEntries(dropAtItem, item);
	else QuMsgBar("failed to paste item for",text.toLatin1().data());
    }
    emit listModified();
}



/// \fn void QguListView::moveItemDown()
/// \brief Method that moves a selected item down within the current list view. Will move in and out of folders.
/// \param     None
/// \exception None
/// \return    None
///  
void QguListView::moveItemDown(){    

 if(QuSlotDebug())
 {
	printf("QguListView::moveItemDown %ld\n", this);
	fflush(stdout);
 }
 
   // move an item down one place in a listView
   // Note that this descends into folders so the
   // WLcontexts can change and listModified must be emitted
   
   QguListItem * itemBelow;
   QguListItem *atItem =  (QguListItem *) selectedItem();
   if( atItem == NULL || atItem->isLocked()) return;

    itemBelow = getSafeItemBelow(atItem);
    if( itemBelow ) {
       if( atItem->parent() )
         atItem->parent()->takeItem(atItem);
       else
         // take from list view root
         takeItem(atItem);

	 
       if( itemBelow->getEntryType() == ListTypeEntry ){
	 
         if (itemBelow->parent() )
	   itemBelow->parent()->insertItem(atItem);
	 else
	   // insert as root item
	   insertItem(atItem);
 
	 atItem->moveItem(itemBelow);
         
       }
       else if( itemBelow->getEntryType() == ListTypeGroup )
         itemBelow->insertItem(atItem);
    } 
    else{
      // no item below
      	QguListItem *folder = (QguListItem *)atItem->parent();
        if( folder ){
	  // in a folder, move below folder

	  if(folder->parent()){
	    folder->takeItem(atItem);
	    folder->parent()->insertItem(atItem);
            atItem->moveItem(folder);	
	  }
	  // we do not permit an item to be moved to root level - this may give join list trouble  
	}
	// else at end of list, do nothing
     }
      
     clearSelection();
     setCurrentItem(atItem);
     emit listModified();
}


/// \fn void QguListView::moveItemUp()
/// \brief Method that moves a selected item up within the current list view. Moves in and out of folders
/// \param     None
/// \exception None
/// \return    None
///  
void QguListView::moveItemUp(){     

 if(QuSlotDebug())
 {
	printf("QguListView::moveItemUp %ld\n", this);
	fflush(stdout);
 }
 
   // move an item up one place in a listView
   // Note that this descends into folders so the
   // WLcontexts can change and listModified must be emitted

   QguListItem * itemAbove;
   QguListItem *atItem =  (QguListItem *) selectedItem();
   if (atItem == NULL || atItem->isLocked()) return;
       
     itemAbove = getSafeItemAbove(atItem);

     if(itemAbove){

       if( atItem->parent() )
         atItem->parent()->takeItem(atItem);
       else
         // take from list view root
         takeItem(atItem);
	 
       if( itemAbove->getEntryType() == ListTypeEntry ){
	 
         if (itemAbove->parent() )
	   itemAbove->parent()->insertItem(atItem);
	 else
	   // insert as root item
	   insertItem(atItem);
         
	 // move down list to below itemAbove then swap
	 // done this way as safeItemAbove may be in different folder
	 atItem->moveItem(itemAbove);
         itemAbove->moveItem(atItem);
         
       }
       else if( itemAbove->getEntryType() == ListTypeGroup )
         itemAbove->insertItem(atItem);
     }
     
     clearSelection();
     setCurrentItem(atItem);
     emit listModified();
}


/// \fn QguListItem * QguListView::getSafeItemAbove(QguListItem *atItem)
/// \brief Method that retrieves an item above atItem that is a safe target to insert an item during move 
/// \param     atItem of type QguListItem * 
/// \exception None
/// \return    Returns a safe item above the specified item
///  
QguListItem * QguListView::getSafeItemAbove(QguListItem *atItem){
  QguListItem *above;
  
    // ensures that item above is valid place to put item
    // returns NULL if not possible, and if item above in list is
    // the parent folder go to its parent otherwise
    // move up drops into parent which puts it back where it was
    
    above = (QguListItem *)atItem->itemAbove();

    if(above && above == atItem->parent() ) 
      above = (QguListItem *)atItem->parent()->itemAbove();
    
    if(above && above->getEntryType() == ListTypeNull ) return NULL; 
    
    return above;
}


/// \fn QguListItem * QguListView::getSafeItemBelow(QguListItem *atItem)
/// \brief Method that retrieves an item below atItem that is a safe target to insert an item during move 
/// \param     atItem of type QguListItem * 
/// \exception None
/// \return    Returns a safe item below the specified item
///  
QguListItem * QguListView::getSafeItemBelow(QguListItem *atItem){
  QguListItem *below;
  
   // finds an item below that is not a descendent of the
   // selected item, otherwise move down will try to
   // insert an item (with children) into one of its children (ouch)
   
    below = (QguListItem *)atItem->itemBelow();
    
    // find item that is not locked or a descendent of atItem
    while ( below && (below->isLocked() ||  atItem->isDescendent(below)) ){
      // don't cross separators
      if(below->getEntryType() == ListTypeNull ) return NULL; 
      
      below = getSafeItemBelow(below);
    }
    return below;
}


/// \fn void QguListView::addItemsInclusive( QguListItem *newItem, QguListItem *last )
/// \brief Method that adds a set of items to the selected list (localList) within the current list view between last and newItem
/// \param     newItem of type QguListItem * 
/// \param     last    of type QguListItem * 
/// \exception None
/// \return    None
///  
void QguListView::addItemsInclusive(QguListItem *newItem, QguListItem *last){
    QguListItem *ancestor;
    QguListItem *nextItem;;
    
    // used for shift selection to add all items between
    // the last selection and this
    
    AddToSelectedList(newItem);
    
   // find common ancestor for selected list

   ancestor = findCommonFolder();
   
   if(ancestor == NULL || ancestor->firstChild() == NULL) {
     // no common ancestor on list - use list view
     nextItem = (QguListItem *)newItem->listView()->firstChild();
   }
   else{
     // common ancestor on list
      nextItem = (QguListItem *)ancestor->firstChild();
   }
   
   // find first of 'newItem' and 'last' - select ancestor
   
   while(nextItem){
     if( nextItem->isDescendent(newItem) || nextItem->isDescendent(last) ) {
       if ( !(isOnSelectedList(nextItem)) ) AddToSelectedList(nextItem);
       nextItem = (QguListItem *)nextItem->nextSibling();
       break;
     }
     nextItem = (QguListItem *)nextItem->nextSibling();
   }
   
   // select every sibling inclusive until item 1 or item 2 is descendent  
   while(nextItem){
     if ( !(isOnSelectedList(nextItem)) ) AddToSelectedList(nextItem);
     if( nextItem->isDescendent(newItem) || nextItem->isDescendent(last) ) break;
     nextItem = (QguListItem *)nextItem->nextSibling();
   }
}


/// \fn void QguListView::validateWLitems()
/// \brief Method that validates all items on this list view against the Working List
/// \exception None
/// \return    None
///  
void QguListView::validateWLitems(){

 if(QuSlotDebug())
 {
	printf("QguListView::validateWLitems %ld\n", this);
	fflush(stdout);
 }
 
   // calls the QguListItem validator with the first child
   // of the list with the starting assumption that all items are
   // valid - gets set to false if one is not
   // the QguListItem::validateWLitems() call will check all siblings and descendents recursively
   
   QguListItem * startItem = (QguListItem *) this->firstChild();
   if (startItem) 
      startItem->validateWLitems(true); // start by assuming all items valid

}


/// \fn void QguListView::setItemChoices()
/// \brief Method does nothing, overloaded in derived classes
/// \param     None
/// \exception None
/// \return    None
///  
void QguListView::setItemChoices(){
 
 if(QuSlotDebug())
 {
	printf("QguListView::setItemChoices %ld\n", this);
	fflush(stdout);
 }
 
 // does nothing, overloaded in derived classes
}


/// \fn void QguListView::insertListItems( QguListItem * dropAtItem )
/// \brief Method that inserts items from the dragList within the current list view at dropAt item
/// \param     dropAtItem of type QguListItem * 
/// \exception None
/// \return    None
///  
void QguListView::insertListItems(QguListItem * dropAtItem){
    int i;

    for ( i=0; i<dragList.size(); i++ ){   
      if ( dragList[i]->isLocked() ){
        QuMsgBar("A drag item is locked", false);
		return;
      }
    }
    
    for ( i=0; i<dragList.size(); i++ ){   

		QguListItem * item = dragList[i];
		if(isOnDragList((QguListItem*)item->parent())) continue; // parent will do it
		insertDragItem(dropAtItem,item);
    }

}

bool QguListView::isOnDragList(QguListItem * item)
{
    for (int i=0; i<dragList.size(); i++ ){   
		if(dragList[i] == item) return true;
    }
	return false;
}

/// \fn void QguListView::insertDragItem(QguListItem * dropAt, QguListItem *item)
/// \brief Method that inserts an item within the current list at dropAt. Used for drag and drop.
/// \param     dropAtItem of type QguListItem * 
/// \param     item       of type QguListItem * 
/// \exception None
/// \return    None
///  
void QguListView::insertDragItem(QguListItem * dropAt, QguListItem *item){
  QguListItem * folder;
  QguListItem * itemParent;
  QguListItem * sameItem;
  
 if(QuSlotDebug())
 {
	printf("QguListView::insertDragItem %ld\n", this);
	fflush(stdout);
 }

  QuMsgBar(" ", false);

  
  if( item->listView() == this) {
    // drag from this List
       
    // only move items whose parent is not on the list too
    // note localList and dragList are the same at this stage
    if( item == 0 || isOnSelectedList((QguListItem*)item->parent()) ) {
      return;
    }
    
    if(dropAt == NULL) {
      QguMoveItem(item, NULL);   
      return; 
    }
  
    itemParent = (QguListItem *)item->parent();  
   
    if (dropAt->getEntryType() == ListTypeEntry)
       folder = dropAt->safeParent();
    else if (dropAt->getEntryType() == ListTypeGroup)
       folder = dropAt;
    else return;   
  
    if(folder == NULL) return;
  
    // if name already exists in target folder reject move
    if ( (sameItem = ((QguListItem *)folder->firstChild())->findName(item->text(Obj_Name_))) != 0 ) {
      if(sameItem != item) {
        QuMsgBar(item->text(Obj_Name_).toLatin1().data(),"not transferred; name exists", true) ;
        return;
      }
    }
	 
    QguMoveItem(item, dropAt);    
    
   }
  
             // drag from another list
  else{
    
    insertEntries(dropAt, item);
    
  }
}


/// \fn void QguListView::dropEvent(QDropEvent *event)
/// \brief Event handler called when the drag is dropped on this list view. Re-implemented from QWidget. Drop uses dragList
/// \param     event of type QDropEvent * 
/// \exception None
/// \return    None
///  
void QguListView::dropEvent(QDropEvent *event)
{
  QString txt;
  QguListItem *newItem;
  
 if(QuSlotDebug())
  std::cout << "QguListView::dropEvent " << dynamic_cast<QguListItem*>(currentItem()) << endl;
 

  QguListItem *dropAtItem = (QguListItem *)currentItem() ;  

    // this method is overloaded for the WL itself

  if (dragList.size() == 0) return;

  if ( /*Q3TextDrag::decode(event, txt)*/ event->mimeData()->hasText()) {
	  txt=event->mimeData()->text();

    // special case of single click on item
    if (dropAtItem && txt == dropAtItem->getWLcontext() )    return;


     //  copy drag text to clipboard to allow repeat paste
     //QApplication::clipboard()->setData( new Q3TextDrag(txt) ); 
    QMimeData * clipData=new QMimeData;
    clipData->setText(txt);
    QApplication::clipboard()->setMimeData(clipData);
    // prevent drop onto item in dragList or item with ancestor on list
    // Note dragList and localList contain same items here
    if(dragList[0]->listView()==this)
    {
      if ( isOnSelectedList(dropAtItem) || findAncestorOnList(dropAtItem)) {
        refreshSelectedList();
        return;
      }
    }

    insertListItems(dropAtItem);
    clearSelectedList();
    
    dragList.clear();
    validateWLitems();
    emit listModified();
    
  }// end if on decode event 
}



// QguWindow class
//-----------------
DataNamesTable_var QguWindow::_saved_objects=new DataNamesTable();
DataNamesTable_var QguWindow::_restored_objects=new DataNamesTable();

/// \fn QguWindow::QguWindow( QWidget *parent, const char *name )
/// \brief Constructor of the QguWindow class
/// \param     parent of type QWidget *
/// \param     name   of type const char *
/// \exception None
/// \return    None
///
  QguWindow::QguWindow( QWidget *parent, const char *name )
          :QMainWindow(0,name) // not QMainWindow(parent,name) any more. If so, the children windows are always *over* the main window on Windows, which feels weird.
{
  msgBar=statusBar();
  _menu_bar=menuBar();

  QString userFolder = QSAS_profiles->current()->SaveRestorePath;

  if (userFolder.isEmpty()) this->_filename = QString(".");
  else                    this->_filename = userFolder;

  this->SAVE_EXTENSION = QString(".qs_qgu_default");
  fromSession=false;

}

QguWindow::~QguWindow()
{
 // QguWindow destructor
}

/// \fn void QguWindow::bringToFront()
/// \brief Method that brings the current main window to the foreground and activates it.
/// \param     None
/// \exception None
/// \return    None
///
void QguWindow::bringToFront(){
  
  this->setActiveWindow();
  this->raise();
}


/// \fn void QguWindow::save()
/// \brief Method saves the content of the window to a local file.  Last used save file name is used if it exists, or default.
/// \param     None
/// \exception None
/// \return    None
///
 void QguWindow::save()
 {

 if(QuSlotDebug())
 {
	printf("QgWindow::save %ld\n", this);
	fflush(stdout);
 }


    if(_filename.isEmpty())
    {
       saveAs();
       return;
    }

    QApplication::setOverrideCursor( Qt::waitCursor );
    qApp->processEvents(QEventLoop::AllEvents, 2000);

    try
    {
      VXDRostream xdr_os(_filename.toLatin1().data());
      save(xdr_os);
    }
    catch(Exception &e)
    {
       status(e.msg().c_str(),10000);
       return;
    }

    QApplication::restoreOverrideCursor();
    qApp->processEvents(QEventLoop::AllEvents, 2000);

    status(_filename+" saved",10000);
    return;
 }


/// \fn void QguWindow::saveAs()
/// \brief Method saves the content of the working list within a local file using the filename as provided by the user
/// \param     None
/// \exception None
/// \return    None
///
void QguWindow::saveAs()
{
   QString _filter = QString("save (*");
   _filter += SAVE_EXTENSION;
   _filter += QString(")");
   QString fn=QFileDialog::getSaveFileName(_filename, _filter, this, "save", "Select file - Save Settings");
   if(!(fn.isEmpty()) )
   {
     QFileInfo fn_fi(fn);
     QString tmp=fn_fi.fileName();
     int pos = tmp.findRev(SAVE_EXTENSION);
     if (pos < 0) {
       pos = tmp.findRev(".");
       if (pos > 0) tmp.truncate(pos);
       tmp += SAVE_EXTENSION;
     }
     _filename=fn_fi.dirPath()+"/"+tmp;
     save();
   }
   else
     status("saving aborted ",10000);
 }


  
 /// \fn void QguWindow::saveWithOut()
/// \brief Method that saves the content of the current main window without the data using the previously provided filename
/// \param     None
/// \exception None
/// \return    None
///
void QguWindow::saveWithOut()
 { with=0; save();}
 
  
/// \fn void QguWindow::saveWithOutAs()
/// \brief Method that saves the content of the current main window without the data prompting the user to enter a filename
/// \param     None
/// \exception None
/// \return    None
///
void QguWindow::saveWithOutAs()
 { with=0; saveAs();}
 

void QguWindow::include()
{
  restore(false);
}


/// \fn void QguWindow::restore(bool replace)
/// \brief Method that restores the content of the current window using the file selected by the user (or from session restore)
/// \param     bool replace
/// \exception None
/// \return    None
///
//
//
void QguWindow::restore(bool replace)
 {
 if(QuSlotDebug())
 {
	printf("QguWindow::restore %ld replace = %ld\n", this, replace);
	fflush(stdout);
 }

   QString fn;
   if(!fromSession){
      fn=QFileDialog::getOpenFileName(_filename,_saveset_filter,this);
      _filename=fn;
   }
   fromSession = false;

   if(!(_filename.isEmpty()) )
   {
      QApplication::setOverrideCursor( Qt::waitCursor );
      qApp->processEvents(QEventLoop::AllEvents, 2000);

      try
      {
        VXDRistream xdr_is(_filename.toLatin1().data());
        restore(xdr_is,replace);
	    status(_filename+" restored",10000);
      }
      catch(Exception &e)
      {
         QApplication::restoreOverrideCursor();
         qApp->processEvents(QEventLoop::AllEvents, 2000);
         status(e.msg().c_str(),10000);
         return;
      }

      QApplication::restoreOverrideCursor();
      qApp->processEvents(QEventLoop::AllEvents, 2000);

   }
   else
     QuMsgBar("Empty file name", true);
 }


/// \fn void QguWindow::saveSession(QString folder)
/// \brief Method that saves the content of this window when saving the current QSAS session
/// \param     folder of type QString
/// \exception None
/// \return    None
///
 void QguWindow::saveSession(QString folder)
 {
  if(folder.isEmpty()) return;
  // keep original name
  QString fileN = QString(_filename.toLatin1().data());

  // set session folder and save
     this->_filename = QString(folder.toLatin1().data());
     this->_filename += QString("/saveSet");
     this->_filename += this->SAVE_EXTENSION;
  saveWithOut();

  // put file name back
  _filename = QString(fileN.toLatin1().data());
}


/// \fn void QguWindow::restoreSession(QString folder)
/// \brief Method that restores the content of this window when restoring a session
/// \param     folder of type QString 
/// \exception None
/// \return    None
///
void QguWindow::restoreSession(QString folder){

   if ( QuSlotDebug() )
   {
	printf( "QguWindow::restoreSession %ld\n", this );
	fflush(stdout);
   }
    
   if(folder.isEmpty()) return;
  // keep original name
  QString fileN = QString(_filename.toLatin1().data());
  
  // insert session folder and save
  this->_filename = QString(folder.toLatin1().data());
  this->_filename += QString("/saveSet");
  this->_filename += this->SAVE_EXTENSION;
  this->fromSession = true;
  restore(true);
  // put file name back
  _filename = QString(fileN.toLatin1().data());
}

   /// \fn void status(const QString& str)
  /// \brief Function that displays the message str within the current window message bar.
  /// \param     str of const QString& type
  /// \exception None
  /// \return    None
  ///
 void QguWindow::status(const QString& str)
  {
      msgBar->message(str);
	  if(qApp) qApp->processEvents();
  }

  /// \fn void status(const QString& str,int ms)
  /// \brief Function that displays the message str within the current window message bar
  /// \param     str of const QString& type
  /// \param     ms  of int            type
  /// \exception None
  /// \return    None
  ///
  void QguWindow::status(const QString& str,int ms)
  {
    msgBar->message(str,ms);
    if(qApp) qApp->processEvents();
  }
