//  TimeUtils.cc
//
//  Time Editor Utility Functions
//
//  July 2000: David Burgess
//

#include "TimeUtils.h"

// read from duration string: eg "6d 4h 3m 27.3s"
// split by spaces
// each field
// <n>d (days)
// <n>h (hours)
// <n>m (minutes)
// <n>s (seconds)
// <n>  (seconds)


int calc_tol_spacing(QDOS::Time& spacing_t,const QDOS::TimeSeq& timeseq,
  const QDOS::Time& discard_t0,
  const double& tol, const double& fraction_reqd, const int& max_iterations)
{
// Calc  Spacing for a time series
// Finds the minimum spacing which exceeds discard_t0 and counts the number of
// spacings which are within a factor of tol of the minimum spacing found. If less
// an acceptable fraction fraction_reqd (1%) of half-spacings are included in this count, the
// algorithm is repeated until a suitable spacing is found or the number of iterations
// exceeds max_iterations. Sets spacing_t to tol*spacing found

  int n_data=timeseq.size();
  if(n_data<3) return 1;

  QDOS::Time discard_t(discard_t0);
  QDOS::TimeSeq tint_seq(n_data);
  adjacent_difference(timeseq.begin(),timeseq.end(),tint_seq.begin());
  int num_iterations=0;
  while(num_iterations<max_iterations)
  {
    QDOS::SequenceMask msk(tint_seq.get_mask(QDOS::TimeSeq::GT(discard_t)));
    msk[0]=false; // since adjacent_difference returns first elt unchanged so = timeseq[0]

    QDOS::TimeSeq spacing_seq;
    tint_seq.get_subsequence(spacing_seq,msk);

    QDOS::TimeSeq::iterator spacing_itr=min_element(
      spacing_seq.begin(),spacing_seq.end());
    spacing_t=*spacing_itr;
    QDOS::SequenceMask msk2(spacing_seq.get_mask(
			    QDOS::TimeSeq::LE(tol*spacing_t)));
    if(msk2.num_true()>=fraction_reqd*(n_data-1))
    {
     
      spacing_t*=tol;
      return 0;
    }
    discard_t=spacing_t;
    num_iterations++;
  }
  return 1;
}

std::string strict_iso_srep(const QDOS::Time& t)
{
  return t.iso_srep("T","Z");
}
	
QDOS::Time set_from_duration_str( const std::string& d_str )
{
  if( d_str.length() == 0 )
  {
    throw QDOS::BadTime();
  }

  std::string s = d_str;
  while( s[0] == ' ' && s.length()!=0 )
    s.erase(0,1);
  while( s[s.length()-1] == ' ' && s.length()!=0 )
    s.erase(s.length()-1,std::string::npos);

  if( s.length() == 0 )
  {
    throw QDOS::BadTime();
  }

  std::vector<std::string> tokens;
  int idx;
  while( (idx=s.find(' '))!=std::string::npos )
  {
    if( idx!=0 )
      tokens.push_back( s.substr(0,idx) );
    s = s.substr(idx+1,std::string::npos);
  }
  tokens.push_back( s.substr(0,std::string::npos) );

  QDOS::Time duration;

  for(int i=0;i<tokens.size();i++)
  {
    int l = tokens[i].length();
    if( l == 0 )
      continue;

    if( tokens[i][l-1] == 'd' )
      duration += QDOS::Days( atof( tokens[i].c_str() ) );
    else if( tokens[i][l-1] == 'h' )
      duration += QDOS::Hours( atof( tokens[i].c_str() ) );
    else if( tokens[i][l-1] == 'm' )
      duration += QDOS::Minutes( atof( tokens[i].c_str() ) );
    else
      duration += QDOS::Seconds( atof( tokens[i].c_str() ) );
  }

  return duration;
}

#include <stdio.h> // for sprintf

std::string int_str( int i )
{
  static char sbuff[30];
  sprintf(sbuff,"%i",i);
  return sbuff;
}

std::string double_str( double d )
{
  static char sbuff[30];
  sprintf(sbuff,"%g",d);
  return sbuff;
}

std::string duration_str( const QDOS::Time t )
{
  static char sbuff[20];
  int d, h, m;
  double s;
  QDOS::Seconds t2(fabs(t.get_time_rep()));
  t2.split_dhms( d, h, m, s );
  if(t.get_time_rep()<0.0)
  {
    d=-d;
    h=-h;
    m=-m;
    s=-s;
  }
  std::string d_str;
  if( d!=0 )
    d_str += int_str(d) + "d ";
  if( h!=0 )
    d_str += int_str(h) + "h ";
  if( m!=0 )
    d_str += int_str(m) + "m ";
  if( s!=0 )
    d_str += double_str(s) + "s";

  return d_str;
}

bool valid_iso_time_str( const std::string& s, QDOS::Time& t )
{
// set from iso string. Assume in one of following forms
// yyyy-mm-dd
// yyyy-mm-ddTHH  (or yyyy-mm-dd HH)
// yyyy-mm-ddTHH:MM (or yyyy-mm-dd HH:MM)
// yyyy-mm-ddTHH:MM:SS.SSS (or yyyy-mm-dd HH:MM:SS.SSS)

  bool valid_time=false;

  try{
   if(s=="t" || s=="T")
     t.set_today();
   else if(s == "y" || s=="Y")
   {
     t.set_today();
     t-=QDOS::Hours(24);
   }
   else if(s=="n" || s=="N")
     t.set_current_time();
   else 
     t.set_from_iso( s );
   valid_time=true;

  } catch( QDOS::BadTime& e ) { t = QDOS::Seconds(0.); }

  return valid_time;

}

